*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Klassen fr AutoQuit.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 21.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

*======================================================================================
* Detects user activity and fires an event after the specified period of inactivity.
*======================================================================================
DEFINE CLASS InactivityTimer AS TIMER

	*----------------------------------------------------------
	* API constants
	*----------------------------------------------------------
	#DEFINE WM_KEYUP								0x0101
	#DEFINE WM_SYSKEYUP								0x0105
	#DEFINE WM_MOUSEMOVE							0x0200
	#DEFINE GWL_WNDPROC								(-4)

	*----------------------------------------------------------
	* internal properties
	*----------------------------------------------------------
	nTimeOutInSeconds = 60
	tLastActivity = {/:}
	nOldProc = 0

	*----------------------------------------------------------
	* Timer configuration
	*----------------------------------------------------------
	INTERVAL = 20000
	ENABLED = .T.

	*-------------------------------------------------------------------------------------
	* Listen to API events when the form starts. You can pass the timeout as a parameter.
	*-------------------------------------------------------------------------------------
	PROCEDURE INIT(tnTimeOutInSeconds AS INTEGER)

		DECLARE INTEGER GetWindowLong IN WIN32API INTEGER HWND, INTEGER nIndex
		DECLARE INTEGER CallWindowProc IN WIN32API INTEGER lpPrevWndFunc, INTEGER HWND,INTEGER Msg,	INTEGER wParam,	INTEGER LPARAM

		THIS.nOldProc = GetWindowLong(_VFP.HWND,GWL_WNDPROC)
		IF VARTYPE(m.tnTimeOutInSeconds) == "N"
			THIS.nTimeOutInSeconds = m.tnTimeOutInSeconds
		ENDIF
		THIS.tLastActivity = DATETIME()
		THIS.StartBinding()
	ENDPROC

	*------------------------------------------------------------
	* Das Event-Binding stzen
	*------------------------------------------------------------
	PROCEDURE StartBinding

		BINDEVENT(0,WM_KEYUP, THIS, "WndProc")
		BINDEVENT(0,WM_MOUSEMOVE, THIS, "WndProc")

	ENDPROC

	*------------------------------------------------------------
	* Das Event-Binding aufheben
	*------------------------------------------------------------
	PROCEDURE StopBinding

		UNBINDEVENTS(0, WM_KEYUP)
		UNBINDEVENTS(0, WM_MOUSEMOVE)

	ENDPROC

	*------------------------------------------------------------
	* Stop listening at Form-UNLOAD
	*------------------------------------------------------------
	PROCEDURE UNLOAD

		THIS.StopBinding()

	ENDPROC

	*------------------------------------------------------------
	* Every event counts as activity
	*------------------------------------------------------------
	PROCEDURE WndProc( HWND AS LONG, Msg AS LONG, wParam AS LONG, LPARAM AS LONG )

		THIS.tLastActivity = DATETIME()
		RETURN CallWindowProc(THIS.nOldProc, HWND, Msg, wParam,LPARAM)

	ENDPROC

	*------------------------------------------------------------
	* Check last activity against time out
	*------------------------------------------------------------
	PROCEDURE TIMER

		LOCAL ltFireEvent

		*-- Die Inaktivitt nur zwischen 17:00 und 07:00 prfen
		IF TIME() < '07:00:00' .OR. TIME() > '17:00:00'
			ltFireEvent = THIS.tLastActivity + THIS.nTimeOutInSeconds
			IF DATETIME() > m.ltFireEvent

				*-- Der Timeout wurde erreicht
				THIS.StopBinding()
				THIS.EventTimeout()

				*-- Und weiter gehts
				THIS.tLastActivity = DATETIME()
				THIS.StartBinding()
			ENDIF
		ENDIF
	ENDPROC

	*------------------------------------------------------------
	* Override this event or bind to it to respond to user
	* inactivity. You can change the nTimeOutInSeconds to offer
	* multiple stages of timeouts.
	*------------------------------------------------------------
	PROCEDURE EventTimeout
		MESSAGEBOX('Timeout')
	ENDPROC

ENDDEFINE

*-- Klasse fr die Laufzeit
DEFINE CLASS Inactivity AS InactivityTimer

	*-- INIT-Methode
	PROCEDURE INIT(tnTimeOutInSeconds AS INTEGER)

		*-- Wartezeit auf 60 Sekunden setzen
		DODEFAULT( tnTimeOutInSeconds )

	ENDPROC


	*-- Wir berschreiben hier die EventTimeout-Methode
	PROCEDURE EventTimeout

		*-- HaPe fertig
		DO Quit_KDV WITH .T.
	ENDPROC

ENDDEFINE
