********************************************************************************
*-- PROCEDURE   : ErrorHandler.
*-------------------------------------------------------------------------------
*-- PARAMETER   : Siehe unten, PROCEDURE SetErrorHandler.
*-------------------------------------------------------------------------------
*-- BESCHREIBUNG: Globaler Error Handler.
*-------------------------------------------------------------------------------
*-- AUTHOR      : Markus Winhard, Lauton Software GmbH (eMail: mw@bingo-ev.de).
********************************************************************************
LPARAMETERS tnError, tcMessage, tcCode, tcProgram, tnLineNo, tcWindowName, ;
	tcMessageParameter, tcSys16, tnLastKey, tcVarRead, tcDBF, tcAlias, ;
	tcOrder, tnDataSession

*--------------------------------------------
* #DEFINEs.
*--------------------------------------------
* Diese Konstanten mssen fr jedes Programm angepat werden.
#DEFINE ccMANUFACTURER			'BKI / die FREUNDliche software'
#DEFINE ccAPPCAPTION			'BKI_KDV'
#DEFINE ccAPPUSERNAME			m.gcUserName
* #DEFINE ccAPPVERSION			m.gcVersion
#DEFINE ccAPPVERSION			exe_version(.T.)

* Konstanten fr die Logdatei.
#DEFINE MAX_ERROR_RECORDS		50
* #DEFINE ERRORLOG_DIR			ADDBS(SYS(2023))+"bki_error_kdv\"
* #DEFINE ERRORLOG_DIR			'C:\BKI\ERROR\'
#DEFINE ERRORLOG_DIR			'ERROR\'

* Eigene Konstanten fr MESSAGEBOX() Texte.
#DEFINE ERR_OCCURRED			'Es ist ein Fehler aufgetreten.'
#DEFINE ERR_MESSAGE				"Meldung: "
#DEFINE ERR_NO					"Fehler #: "
#DEFINE ERR_LINE				"Codezeile: "
#DEFINE ERR_LINENO				"Zeile: "
#DEFINE ERR_PROGRAM				"Programm: "
#DEFINE ERR_ALIAS				"Alias: "
#DEFINE ERR_MASK				"Maske: "
#DEFINE ERR_ABORT				"Abbrechen: Programm  K u n d e n v e r w a l t u n g    b e e n d e n  !!! (Alle Daten werden gespeichert!)"
#DEFINE ERR_RETRY				"Wiederholen: Versuchen aktuelle Aktion noch mal auszufhren (Immer zuerst versuchen!)."
#DEFINE ERR_IGNORE				"Ignorieren: Aktuelle Aktion berspringen (hilft manchmal, kann aber auch zu weiteren Fehlermeldungen fhren!)."
#DEFINE CR						CHR(13)
#DEFINE LF						CHR(10)

* MessageBox Parameter aus FoxPro.h
#DEFINE MB_OK                   0       && OK button only
#DEFINE MB_OKCANCEL             1       && OK and Cancel buttons
#DEFINE MB_ABORTRETRYIGNORE     2       && Abort, Retry, and Ignore buttons
#DEFINE MB_YESNOCANCEL          3       && Yes, No, and Cancel buttons
#DEFINE MB_YESNO                4       && Yes and No buttons
#DEFINE MB_RETRYCANCEL          5       && Retry and Cancel buttons

#DEFINE MB_ICONSTOP             16      && Critical message
#DEFINE MB_ICONQUESTION         32      && Warning query
#DEFINE MB_ICONEXCLAMATION      48      && Warning message
#DEFINE MB_ICONINFORMATION      64      && Information message

#DEFINE MB_APPLMODAL            0       && Application modal message box
#DEFINE MB_DEFBUTTON1           0       && First button is default
#DEFINE MB_DEFBUTTON2           256     && Second button is default
#DEFINE MB_DEFBUTTON3           512     && Third button is default
#DEFINE MB_SYSTEMMODAL          4096    && System Modal

#DEFINE IDOK            		1       && OK button pressed
#DEFINE IDCANCEL        		2       && Cancel button pressed
#DEFINE IDABORT         		3       && Abort button pressed
#DEFINE IDRETRY         		4       && Retry button pressed
#DEFINE IDIGNORE        		5       && Ignore button pressed
#DEFINE IDYES           		6       && Yes button pressed
#DEFINE IDNO            		7       && No button pressed

ON ERROR

*--------------------------------------------
* Screeen sichtbar machen
*--------------------------------------------
_SCREEN.VISIBLE = .T.

*--------------------------------------------
* "Zu wenig Speicher" Fehler zuerst behandeln.
*--------------------------------------------
DO CASE
	CASE INLIST( m.tnError, 21, 22 )
		ON ERROR *
		CLEAR WINDOWS
		CLEAR POPUPS
		CLEAR MENUS
		CLOSE ALL
		CLEAR TYPEAHEAD
		SET PRINTER OFF
		SET DEVICE TO SCREEN
		SET CURSOR ON
		WAIT WINDOW "Zu viele Variablen. Das Programm wird beendet."
		* Hier ggf. User abmelden.
		ON SHUTDOWN
		QUIT
	CASE INLIST( m.tnError, 43, 1149, 1150, 1151, 1250, 1600, 1809, 1845 )
		ON ERROR *
		CLEAR WINDOWS
		CLEAR POPUPS
		CLEAR MENUS
		CLOSE ALL
		CLEAR TYPEAHEAD
		SET PRINTER OFF
		SET DEVICE TO SCREEN
		SET CURSOR ON
		WAIT WINDOW NOWAIT "Nicht gengend Speicher [ERROR #: " + LTRIM(STR( m.tnError )) + "]. Das Programm wird beendet."
		* Hier ggf. User abmelden.
		ON SHUTDOWN
		QUIT
ENDCASE

*--------------------------------------------
* Bildschirmanzeige erlauben
*--------------------------------------------
TRY
	_SCREEN.LOCKSCREEN = .F.
CATCH
ENDTRY

*--------------------------------------------
* ErrorLogDir festlegen
*--------------------------------------------
PUBLIC gcErrorLogDir
gcErrorLogDir = ERRORLOG_DIR
IF !DIRECTORY(gcErrorLogDir)
	TRY
		MD (gcErrorLogDir)
	CATCH
		WAIT WINDOW 'Pfad fr "Error.log" kann nicht erstellt werden!'
	ENDTRY
ENDIF

*--------------------------------------------
* Setup.
*--------------------------------------------
LOCAL lnSelect, lcStatus, lcErrMsg, lcSys0, lcWinUserName, lcMachineName, ;
	lcAppUserName, lcTempFile, i, llErrorLogged, laTempFiles[1], ;
	lnTempFiles, lcTempDir, lnAction, lnPrgStack, laStack[1], lcVersion, ;
	llDebug, llDevelopment, loForm, llFormFound, lcDate, lcTime, ;
	laSessions[1], llCOMServer, laErrorInfo[1], ;
	lcKundenNr,lcAnwenderNr,lcFreischaltung,llMail,lnMsg
llMail = .T.
lnMsg  = 0
lcDate = DTOC(DATE())
lcTime = TIME()
lnSelect = SELECT()
lcErrMsg = ""
lcSys0 = SYS(0)
lcTempDir = ADDBS(GETENV( "TEMP" ))
lcWinUserName = SUBSTR( m.lcSys0, AT( "#", m.lcSys0 ) + 2 )
lcMachineName = LEFT( m.lcSys0, AT( "#", m.lcSys0 ) - 2 )
lcAppUserName = IIF( TYPE( [ccAPPUSERNAME] ) == "C", ccAPPUSERNAME, "(unbekannt)" )
lcVersion = IIF( TYPE( [ccAPPVERSION] ) == "C", ccAPPVERSION, "(unbekannt)" )

#IF VAL(SUBSTR(VERSION(),14)) < 6
	llCOMServer = .F.
#ELSE
	llCOMServer = INLIST( _VFP.STARTMODE, 2, 3, 5 )
#ENDIF
=AERROR( laErrorInfo )	&& Zustzliche Informationen bei OLE, OBDC und Trigger Fehler.
lcTempFile = m.lcTempDir + SYS(2015) + ".TMP"
lnTempFiles = ADIR( laTempFiles, m.lcTempDir + "*.*", "HS" )
RELEASE laTempFiles
llDebug = .F.
llDevelopment = NOT VERSION(2) == 0
USE IN SELECT( "ErrorLog" )
ERASE (m.lcTempFile)
FLUSH

*--------------------------------------------
* Zustzliche Umgebungsinformationen sammeln.
*--------------------------------------------
lcStatus = ;
	"*===========================================================" + CR + ;
	"* Programm: " + [ccAPPCAPTION] + CR + ;
	"* Programm Version: " + m.lcVersion + CR + ;
	"* Programm Login: " + m.lcAppUserName + CR + ;
	"* Windows Login: " + m.lcWinUserName + CR + ;
	"* Arbeitsstation: " + m.lcMachineName + CR + ;
	"* Datum: " + m.lcDate +", "+ m.lcTime + CR + ;
	"*===========================================================" +CR

*--------------------------------------------
* Fehlernummer.
*--------------------------------------------
lcStatus = m.lcStatus + CR + "FehlerNr.: " + LTRIM(STR( m.tnError ))

*--------------------------------------------
* Fehlermeldung.
*--------------------------------------------
lcStatus = m.lcStatus + CR + "Meldung:   " + m.tcMessage
IF NOT EMPTY( m.tcMessageParameter )
	lcStatus = m.lcStatus +" ("+ m.tcMessageParameter + ")"
ENDIF

*--------------------------------------------
* Ausgefhrter Code.
*--------------------------------------------
lcStatus = m.lcStatus + CR + "Code:      "+ m.tcCode

*--------------------------------------------
* Methode bzw. Prozedur.
*--------------------------------------------
lcStatus = m.lcStatus + CR + "Programm:  " + m.tcProgram + " (" + m.tcSys16 + ")"

*--------------------------------------------
* Zeilennummer.
*--------------------------------------------
lcStatus = m.lcStatus + CR + "Zeile:     " + LTRIM(STR( m.tnLineNo ))

*--------------------------------------------
* DatasessionID.
*--------------------------------------------
lcStatus = m.lcStatus + CR + "Datasession:" + LTRIM(STR( SET("Datasession") ))

*--------------------------------------------
* Call Stack.
*--------------------------------------------
lcStatus = m.lcStatus + CR +CR+ ;
	"*===========================================================" + CR + ;
	"* Call Stack:" + CR + ;
	"*===========================================================" +CR
#IF VAL(SUBSTR(VERSION(),14)) < 7
	*--------------------------------------------
	* Vfp6 Version.
	*--------------------------------------------
	FOR i = 1 TO PROGRAM( -1 )
		lcStatus = m.lcStatus + CR + ;
			SYS( 16, i )
	ENDFOR
#ELSE
	*--------------------------------------------
	* In Vfp7 zustzlich Zeilennummer und Code.
	*--------------------------------------------
	lnPrgStack = ASTACKINFO( laStack )
	lcStatus = m.lcStatus + CR + SYS(16,0)
	FOR i = 1 TO m.lnPrgStack - 1
		lcStatus = m.lcStatus + CR + UPPER( laStack[ i, 3 ] ) + " " + UPPER( laStack[ i, 4 ] ) + ", " + ;
		LTRIM(STR( laStack[ i, 5 ] )) + CR + SPACE(4) + ALLTRIM(STRTRAN( laStack[ i, 6 ], CHR(9), SPACE(4) ))
	ENDFOR
	lcStatus = m.lcStatus + CR + UPPER( laStack[ m.lnPrgStack, 3 ] ) +" "+ UPPER( laStack[ i, 4 ] )
	RELEASE laStack
#ENDIF

*--------------------------------------------
* OpenTables.
*--------------------------------------------
lcStatus = m.lcStatus + CR +CR+ ;
	"*===========================================================" + CR + ;
	"* OpenTables:" + CR + ;
	"*===========================================================" +CR
FOR i = 1 TO 100
	IF !EMPTY(ALIAS(i))
		lcStatus = m.lcStatus + CR + "Alias:      " + PADR(LOWER(ALIAS(i)),20)+TRIM(STR(RECNO(i),8))+PADR(TAG(i),20)+KEY(i)
	ENDIF
NEXT

*--------------------------------------------
* Aktive Form und aktives Steuerelement.
*--------------------------------------------
lcStatus = m.lcStatus + CR +CR+ ;
	"*===========================================================" + CR + ;
	"* Active form and control:" + CR + ;
	"*===========================================================" +CR
lcStatus = m.lcStatus + CR + "Aktiv:     "
IF TYPE( "_Screen.ActiveForm.Name" ) == "C"
	*--------------------------------------------
	* Wenn auf einer Form ActiveX Controls liegen,
	* kann es passieren, da _Screen.ActiveForm
	* eine Referenz auf ein ActiveX Control ist.
	*--------------------------------------------
	loForm = _SCREEN.ACTIVEFORM
	llFormFound = .F.
	IF TYPE( "m.loForm.BaseClass" ) == "C"
		IF UPPER( m.loForm.BASECLASS ) == "FORM"
			llFormFound = .T.
		ELSE
			IF UPPER( m.loForm.BASECLASS ) == "OLECONTROL"
				DO WHILE TYPE( "m.loForm.Parent.BaseClass" ) == "C"
					loForm = m.loForm.PARENT
					IF UPPER( m.loForm.BASECLASS ) == "FORM"
						llFormFound = .T.
						EXIT
					ENDIF
				ENDDO
			ENDIF
		ENDIF
	ENDIF
	IF m.llFormFound
		lcStatus = m.lcStatus + m.loForm.NAME
		IF TYPE( "m.loForm.ActiveControl.Name" ) == "C"
			lcStatus = m.lcStatus +" ("+ m.loForm.ACTIVECONTROL.NAME + ")"
		ENDIF
	ENDIF
	loForm = .NULL.
	RELEASE loForm
ENDIF
lcStatus = m.lcStatus + CR + "WOutput:   " + m.tcWindowName
lcStatus = m.lcStatus + CR + "VarRead:   " + m.tcVarRead
lcStatus = m.lcStatus + CR + "Session:   " + LTRIM(STR( m.tnDataSession ))

*--------------------------------------------
* Aktuelle Tabelle und Sortierung.
*--------------------------------------------
lcStatus = m.lcStatus + CR + "Tabelle:   " + m.tcAlias
IF NOT EMPTY( m.tcDBF )
	lcStatus = m.lcStatus +" ("+ m.tcDBF + ")"
ENDIF
lcStatus = m.lcStatus + CR + "Index:     " + m.tcOrder

*--------------------------------------------
* Zuletzt gedrckte Taste.
*--------------------------------------------
lcStatus = m.lcStatus + CR + "LastKey:   " + LTRIM(STR( m.tnLastKey ))

*--------------------------------------------
* Betriebssystem-Umgebung.
*--------------------------------------------
lcStatus = m.lcStatus + CR + "CurDir:    " + SYS(5) + CURDIR()
lcStatus = m.lcStatus + CR + "DiskSpace: " + LTRIM(STR( DISKSPACE()/1024/1024, 25 )) + " MB"
lcStatus = m.lcStatus + CR + "TempDir:   " + m.lcTempDir + IIF( m.lcTempDir == SYS(2023), "", " (" + SYS(2023) + ")" )
#IF VAL(SUBSTR(VERSION(),14)) >= 6
	lcStatus = m.lcStatus + CR + "TempSpace: " + LTRIM(STR( DISKSPACE(LEFT( m.lcTempDir, 1 ))/1024/1024, 25 )) + " MB"
#ENDIF
lcStatus = m.lcStatus + CR + "TempFiles: " + LTRIM(STR(m.lnTempFiles))
lcStatus = m.lcStatus + CR + "Screen:    " + LTRIM(STR(SYSMETRIC(1))) +" * "+ LTRIM(STR(SYSMETRIC(2)))
lcStatus = m.lcStatus + CR + "OS:        " + OS()

DO CASE
	CASE "5.01"$OS()
		lcStatus = m.lcStatus + " (Windows XP)"
	CASE "6.00"$OS()
		lcStatus = m.lcStatus + " (Windows VISTA)"
	CASE "6.01"$OS()
		lcStatus = m.lcStatus + " (Windows 7)"
	CASE "6.02"$OS()
		lcStatus = m.lcStatus + " (Windows 8)"
ENDCASE

#IF VAL(SUBSTR(VERSION(),14)) >= 7
	FOR i = 1 TO 11
		lcStatus = m.lcStatus + CR + PADR( "OS(" + LTRIM(STR(i)) + "):", 11 ) + OS(i)
	ENDFOR
#ENDIF
lcStatus = m.lcStatus + CR + "Vers(1):   " + VERSION(1)
lcStatus = m.lcStatus + CR + "Vers(2):   " + LTRIM(STR(VERSION(2)))
lcStatus = m.lcStatus + CR + "Vers(3):   " + VERSION(3)
#IF VAL(SUBSTR(VERSION(),14)) >= 6
	lcStatus = m.lcStatus + CR + "StartMode: " + LTRIM(STR( _VFP.STARTMODE ))
#ENDIF
lcStatus = m.lcStatus + CR + "SYS(1016): " + LTRIM(STR( VAL(SYS(1016)) / 1024 )) + " user object memory used"
lcStatus = m.lcStatus + CR + "SYS(1001): " + LTRIM(STR( VAL(SYS(1001)) / 1024 )) + " pool available memory"
lcStatus = m.lcStatus + CR + "CPU:       " + SYS(17)
lcStatus = m.lcStatus + CR + "Video:     " + SYS(2006)

*--------------------------------------------
* List Status.
*--------------------------------------------
* Option *
LIST STATUS TO (m.lcTempFile) NOCONSOLE
FLUSH
lcStatus = m.lcStatus + CR +CR+ ;
	"*===========================================================" + CR + ;
	"* Status Listing:" + CR + ;
	"*===========================================================" + CR + ;
	FILETOSTR( m.lcTempFile ) + CR
ERASE (m.lcTempFile)
FLUSH

*--------------------------------------------
* Bei Bedarf:
* Variablen die der Anwender nicht im
* Errorlog sehen soll temporr in eine Datei
* speichern und anschlieend freigeben.
* Der dazu verwendete Dateiname ist
* willkrlich gewhlt.
*--------------------------------------------
*!*	ERASE (m.lcTempDir + "OS2996PC.TMP")
*!*	SAVE TO m.lcTempDir + "OS2996PC.TMP" ALL LIKE g*
*!*	RELEASE gcUserPassword, gcMasterPassword, gcLicenceInfo, gcModules
*--------------------------------------------
* List Memory.
*--------------------------------------------

* OPTION *
LIST MEMORY TO (m.lcTempFile) NOCONSOLE
FLUSH
lcStatus = m.lcStatus + CR + ;
	"*===========================================================" + CR + ;
	"* Memory Listing:" + CR + ;
	"*===========================================================" + CR + ;
	FILETOSTR( m.lcTempFile ) + CR
ERASE (m.lcTempFile)

*--------------------------------------------
* Bei Bedarf:
* Variablen wiederherstellen.
* Die mssen hier neu als PUBLIC definiert
* werden, sonst erstellt der RESTORE FROM
* nur PRIVATE Variablen die nach Beenden
* dieser Prozedur wieder weg sind.
*--------------------------------------------
*!*	PUBLIC gcUserPassword, gcMasterPassword, gcLicenceInfo, gcModules
*!*	RESTORE FROM m.lcTempDir + "OS2996PC.TMP" ADDITIVE
*!*	ERASE (m.lcTempDir + "OS2996PC.TMP")

*--------------------------------------------
* VFPs Cache leeren.
*--------------------------------------------
=SYS(1104)

*--------------------------------------------
* Versuchen, den Fehler zu protokollieren.
*--------------------------------------------
DO Open_ErrorLog
IF USED( "ErrorLog" ) ;
		AND NOT ISREADONLY( "ErrorLog" )
	SELECT ErrorLog
	APPEND BLANK IN ErrorLog
	REPLACE ;
		cDate		WITH m.lcDate, ;
		cTime		WITH m.lcTime, ;
		cUser		WITH m.lcAppUserName, ;
		cMachine	WITH m.lcMachineName, ;
		cWinUser	WITH m.lcWinUserName, ;
		cErrMsg		WITH m.tcMessage, ;
		cErrNo		WITH LTRIM(STR( m.tnError )), ;
		cErrLine	WITH m.tcCode, ;
		cLineNo		WITH LTRIM(STR( m.tnLineNo )), ;
		cErrPrg 	WITH m.tcProgram, ;
		cAlias		WITH m.tcAlias, ;
		cWindow		WITH m.tcWindowName, ;
		mStatus		WITH m.lcStatus ;
		IN ErrorLog
	IF CURSORGETPROP( "Buffering" ) > 1
		llErrorLogged = TABLEUPDATE( .F., .T. )
	ELSE
		llErrorLogged = .T.
	ENDIF
ELSE
	llErrorLogged = .F.
ENDIF

*--------------------------------------------
* Sonderbehandlung fr bekannte Fehler die zu
* Supportanfragen fhren, wenn sie nicht
* programmatisch gelst werden.
*--------------------------------------------
DO CASE
	CASE m.tnError == 5
		SET STEP ON

	CASE m.tnError == 1

	CASE m.tnError == 41

	CASE m.tnError == 1542

	CASE m.tnError == 6
		* Zu viele Dateien offen (Fehler Nr. 6).
		lcErrMsg = "Zu viele Dateien geffnet!" + CR + ;
			"Bitte berprfen Sie Ihre 'FILES'-Einstellung in der Datei 'CONFIG.SYS'."
		lcErrMsg = ERR_OCCURRED + CR +CR+ m.lcErrMsg
		lcErrMsg = lcErrMsg + CR +CR+ ERR_ABORT + CR + ERR_RETRY + CR + ERR_IGNORE

		IF NOT m.llCOMServer
			=MESSAGEBOX( m.lcErrMsg, MB_ICONSTOP + MB_OK, ccAPPCAPTION )
		ENDIF
		lnAction = IDABORT

	CASE m.tnError == 39
		* Numerischer berlauf (Fehler Nr. 39).
		llMail = .F.
		lcErrMsg = "Werte sind zu gro zum speichern!" + CR + ;
			"Bitte prfen Sie Ihre Eingaben und " + ;
			"fhren Sie ggf. eine Neuberechung" + CR +;
			"(Men Kostenplan) durch."
		SET STEP ON
		IF NOT m.llCOMServer
			lnAction = MESSAGEBOX( m.lcErrMsg, ;
				MB_ICONEXCLAMATION + MB_OK + MB_DEFBUTTON1, ;
				ccAPPCAPTION )
		ENDIF

	CASE m.tnError == 101
		* Datei <Datei> kann nicht geffnet werden (Fehler Nr. 101).
		* VFP kann die angeforderte Datei nicht ffnen.
		* Die Datei wird mglicherweise von einem anderen Benutzer im Netzwerk gesperrt.

	CASE m.tnError == 111
		* Der Cursor kann nicht aktualisiert werden (Fehler Nr. 111).
		* Datei ist schreibgeschtzt.

	CASE m.tnError == 202
		* Ungltiger Pfad oder Dateiname (Fehler Nr. 202).
		lcErrMsg = "Pfad- oder Dateiname nicht gefunden!" + CR + ;
			ccAPPCAPTION + " hat versucht, auf eine Datei oder " + ;
			"ein Verzeichnis zuzugreifen, das nicht existiert." + CR +;
			"Bitte berprfen Sie, ob alle Verzeichnisse korrekt angelegt sind." + CR +CR+ ;
			ERR_NO + LTRIM(STR( m.tnError ))
		lcErrMsg = ERR_OCCURRED + CR +CR+ m.lcErrMsg
		lcErrMsg = lcErrMsg + CR +CR+ ERR_ABORT + CR + ERR_RETRY + CR + ERR_IGNORE

		IF NOT m.llCOMServer
			lnAction = MESSAGEBOX( m.lcErrMsg, ;
				MB_ICONSTOP + MB_ABORTRETRYIGNORE + MB_DEFBUTTON2, ;
				ccAPPCAPTION )
		ENDIF

	CASE m.tnError == 1705
		* Zugriff auf Datei wurde verweigert (Fehler Nr. 1705).
		* Es wurde versucht, in eine schreibgeschtzte Datei zu schreiben
		* bzw. Datei wird bereits von anderer Anwendung benutzt.

	CASE INLIST( m.tnError, 1420, 1426, 1427, 1428, 1440, 2003 )
		* Wahrscheinlich ist eine Library / Dll / Ocx nicht registriert.
ENDCASE

*--------------------------------------------
* Fehler melden (falls das nicht schon oben
* geschehen ist).
*--------------------------------------------
IF EMPTY( m.lcErrMsg )
	lcErrMsg = ;
		ERR_MESSAGE + m.tcMessage + IIF( EMPTY( m.tcMessageParameter ), "", " ("+ m.tcMessageParameter + ")" ) + CR + ;
		ERR_NO + LTRIM(STR( m.tnError )) + CR + ;
		ERR_LINE + m.tcCode + CR + ;
		ERR_LINENO + LTRIM(STR( m.tnLineNo )) + CR + ;
		ERR_PROGRAM + m.tcProgram + CR + ;
		ERR_ALIAS + m.tcAlias + CR + ;
		ERR_MASK + m.tcWindowName
	lcErrMsg = ERR_OCCURRED + CR +CR+ m.lcErrMsg
	lcErrMsg = lcErrMsg + CR +CR+ ERR_ABORT + CR + ERR_RETRY + CR + ERR_IGNORE

	IF NOT m.llCOMServer
		IF m.llDebug
			*-- Entwicklungsumgebung.
			lcErrMsg = m.lcErrMsg + CR + "(Wiederholen = Debug)"
		ENDIF
		m.lnAction = MESSAGEBOX( m.lcErrMsg, MB_ICONSTOP + MB_ABORTRETRYIGNORE + MB_DEFBUTTON2, ccAPPCAPTION )
	ENDIF
ENDIF

*--------------------------------------------
* Anwender informieren, wenn der Fehler nicht
* protokolliert werden konnte.
*--------------------------------------------
IF NOT m.llErrorLogged
	IF m.llCOMServer
		lcErrMsg = m.lcErrMsg + CR + "Der Fehler konnte nicht protokolliert werden."
	ELSE
		=MESSAGEBOX( "Der Fehler konnte nicht protokolliert werden.", ;
			MB_ICONINFORMATION + MB_OK, ccAPPCAPTION )
	ENDIF
ENDIF

*--------------------------------------------
* Optional Versand des Fehlerberichts an
* uns selbst.
*--------------------------------------------
IF m.llMail
	IF MESSAGEBOX( "Wollen Sie einen Problembericht senden?", MB_ICONQUESTION + MB_YESNO + MB_DEFBUTTON1, ccAPPCAPTION ) == IDYES
		*-- 30.09.21 HPG: Umgestellt auf EMail von HPG
		*MESSAGEBOX( SendMail( "freund@bki.de", ErrorLog.mStatus, DATETIME(), m.lcVersion, m.tnError, m.tcMessage ))
		MESSAGEBOX( SendMail( "hanspeter@schwaben-team.de", ErrorLog.mStatus, DATETIME(), m.lcVersion, m.tnError, m.tcMessage ))
	ENDIF
ENDIF
IF DIRECTORY(gcErrorLogDir)
	SET SAFETY OFF
	STRTOFILE( ErrorLog.mStatus, gcErrorLogDir + "Error_" + CHRTRAN( TTOC( DATETIME(), 1 ), ":." ,"--" ) + ".log" )
	SET SAFETY ON
ENDIF

*--------------------------------------------
* Cleanup.
*--------------------------------------------
USE IN SELECT( "ErrorLog" )
SELECT (m.lnSelect)

*--------------------------------------------
* Abbruch, Wiederholen, Ignorieren?
*--------------------------------------------
DO CASE
	CASE m.llCOMServer
		=COMRETURNERROR( ccAPPCAPTION, m.lcErrMsg )
		* COMRETURNERROR() wirkt wie CANCEL.
		* Der COM Server wird aber nicht entladen.
	CASE m.lnAction == IDRETRY
		IF m.llDevelopment
			* Entwicklungsumgebung.
			llDebug = .T.
		ELSE
			* Runtime.
			RETRY
		ENDIF
	CASE m.lnAction == IDIGNORE
		* Nix tun, einfach weiter im Programm.
	OTHERWISE
		* Programm abbrechen.
		ON ERROR *
		ON SHUTDOWN
		IF VERSION(2)=0
			ON ERROR *
			CLEAR WINDOWS
			CLEAR POPUPS
			CLEAR MENUS
			CLOSE ALL
			CLEAR TYPEAHEAD
			SET PRINTER OFF
			SET DEVICE TO SCREEN
			SET CURSOR ON
			ON SHUTDOWN

			CLEAR ALL
			RELEASE ALL

			QUIT
		ELSE
			SET SYSMENU TO
			*CANCEL
			SET STEP ON
		ENDIF
ENDCASE

*--------------------------------------------
* Soll der Debugger gestartet werden?
*--------------------------------------------
IF m.llDebug
	SET VIEW ON
ENDIF
ON ERROR DO SetErrorHandler
RETURN


*===========================================================
* Hilfsfunktionen.
*===========================================================
*--------------------------------------------
* Protokoll DBF ffnen.
*--------------------------------------------
PROCEDURE Open_ErrorLog
	IF FILE( gcErrorLogDir+[ErrorLog.Dbf] )
		USE IN SELECT( "ErrorLog" )
		USE ( gcErrorLogDir+[ErrorLog] ) IN 0 SHARED
	ENDIF
	IF NOT USED( "ErrorLog" ) ;
			OR RECCOUNT( "ErrorLog" ) > MAX_ERROR_RECORDS	&& DBF nicht vorhanden oder defekt. && Zuviele Stze im ErrorLog.
		USE IN SELECT( "ErrorLog" )
		DO Backup_ErrorLog
		DO Create_ErrorLog
		USE ( gcErrorLogDir+[ErrorLog] ) IN 0 SHARED
	ENDIF

ENDPROC

*--------------------------------------------
* ErrorLog DBF erzeugen.
*--------------------------------------------
PROCEDURE Create_ErrorLog

	*--------------------------------------------
	* Setup.
	*--------------------------------------------
	LOCAL lnSelect,lcFile
	lnSelect = SELECT(0)
	SELECT 0
	lcFile = '"'+gcErrorLogDir+'errorlog.dbf"'
	ERASE &lcFile
	lcFile = '"'+gcErrorLogDir+'errorlog.fpt"'
	ERASE &lcFile
	lcFile = '"'+gcErrorLogDir+'errorlog.cdx"'
	ERASE &lcFile
	*--------------------------------------------
	* Tabelle erzeugen.
	*--------------------------------------------
	LOCAL ARRAY laFields[ 13, 4 ]
	laFields[1,1] = "CDATE"
	laFields[1,2] = "C"
	laFields[1,3] = 10
	laFields[1,4] = 0
	laFields[2,1] = "CTIME"
	laFields[2,2] = "C"
	laFields[2,3] = 8
	laFields[2,4] = 0
	laFields[3,1] = "CUSER"
	laFields[3,2] = "C"
	laFields[3,3] = 50
	laFields[3,4] = 0
	laFields[4,1] = "CMACHINE"
	laFields[4,2] = "C"
	laFields[4,3] = 50
	laFields[4,4] = 0
	laFields[5,1] = "CWINUSER"
	laFields[5,2] = "C"
	laFields[5,3] = 50
	laFields[5,4] = 0
	laFields[6,1] = "CERRMSG"
	laFields[6,2] = "C"
	laFields[6,3] = 100
	laFields[6,4] = 0
	laFields[7,1] = "CERRNO"
	laFields[7,2] = "C"
	laFields[7,3] = 5
	laFields[7,4] = 0
	laFields[8,1] = "CERRLINE"
	laFields[8,2] = "C"
	laFields[8,3] = 100
	laFields[8,4] = 0
	laFields[9,1] = "CLINENO"
	laFields[9,2] = "C"
	laFields[9,3] = 5
	laFields[9,4] = 0
	laFields[10,1] = "CERRPRG"
	laFields[10,2] = "C"
	laFields[10,3] = 50
	laFields[10,4] = 0
	laFields[11,1] = "CALIAS"
	laFields[11,2] = "C"
	laFields[11,3] = 20
	laFields[11,4] = 0
	laFields[12,1] = "CWINDOW"
	laFields[12,2] = "C"
	laFields[12,3] = 50
	laFields[12,4] = 0
	laFields[13,1] = "MSTATUS"
	laFields[13,2] = "M"
	laFields[13,3] = 4
	laFields[13,4] = 0

	lcFile = '"'+gcErrorLogDir+'errorlog"'
	CREATE TABLE &lcFile FREE FROM ARRAY laFields
	*--------------------------------------------
	* Cleanup.
	*--------------------------------------------
	USE
	RELEASE laFields
	SELECT (m.lnSelect)

ENDPROC

*--------------------------------------------
* Altes ErrorLog sichern.
*--------------------------------------------
PROCEDURE Backup_ErrorLog
	IF FILE( gcErrorLogDir+[ErrorLog.Dbf] )
		=Backup( gcErrorLogDir+[ErrorLog.Dbf] )
	ENDIF
	IF FILE( gcErrorLogDir+[ErrorLog.Fpt] )
		=Backup( gcErrorLogDir+[ErrorLog.Fpt] )
	ENDIF
ENDPROC

*--------------------------------------------
* Sicherungsdatei der uebergebenen Datei mit
* umgekehrter Endung ( z.B. DBF -> FBD )
* erzeugen (durch Umbenennen).
*--------------------------------------------
PROCEDURE Backup
	LPARAMETERS lcFileName

	*--------------------------------------------
	* Setup.
	*--------------------------------------------
	LOCAL lcExtension, lcName, lcReverseExtension, i
	lcExtension = SUBSTR( m.lcFileName, RAT( ".", m.lcFileName ) + 1 )
	lcName = SUBSTR( m.lcFileName, 1, RAT( ".", m.lcFileName ) )

	*--------------------------------------------
	* Endung umdrehen.
	*--------------------------------------------
	lcReverseExtension = ""
	FOR i = 1 TO LEN( m.lcExtension )
		lcReverseExtension = m.lcReverseExtension + SUBSTR( m.lcExtension, LEN( m.lcExtension ) + 1 - i, 1 )
	ENDFOR

	*--------------------------------------------
	* Alte Backup Datei ggf. loeschen.
	*--------------------------------------------
	IF FILE( m.lcName + m.lcReverseExtension )
		ERASE ( m.lcName + m.lcReverseExtension )
	ENDIF

	*--------------------------------------------
	* Backup Datei erzeugen.
	*--------------------------------------------
	TRY
		RENAME (m.lcName + m.lcExtension) TO (m.lcName + m.lcReverseExtension)
	CATCH
	ENDTRY

ENDPROC

*--------------------------------------------
* ON ERROR Routine aktivieren.
*--------------------------------------------
PROCEDURE SetErrorHandler
	ON ERROR DO ErrorHandler WITH ERROR(), MESSAGE(), MESSAGE(1), PROGRAM(), LINENO(1), WOUTPUT(), ;
		SYS(2018), SYS(16), LASTKEY(), VARREAD(), DBF(), ALIAS(), SET("ORDER"), SET("DATASESSION")
ENDPROC

PROCEDURE debug_error
	SET STEP ON
ENDPROC
**************************************************************************************************************************

PROCEDURE SendMail
	LPARAMETERS tcMailTo, tcErrorlog, ttDatetime, tcVersion, tnError, tcMessage

	#DEFINE MAPI_ORIG        0
	#DEFINE MAPI_TO          1
	#DEFINE MAPI_DIALOG      8
	#DEFINE SUCCESS_SUCCESS  0

	DO DECL

	LOCAL hSession, lcName
	hSession = getNewSession()

	IF hSession = 0
		RETURN "MAPI Anmeldung erfolglos. Vermutlich haben Sie kein eMail Konto eingerichtet."
	ENDIF

	LOCAL loRcpEmail, loSndBuf, lcRcpBuf, loSubject, loNoteText,;
		loRcpBuf, lcMapiMessage, lnResult

	* populating message recipient, subject and body
	loRcpEmail = CREATEOBJECT ("PChar", m.tcMailTo)
	lcName = ccAPPCAPTION
	loSubject  = CREATEOBJECT ("PChar", lcName + " (" + m.tcVersion + ") Errorlog vom " + TTOC(m.ttDatetime) + " [" + LTRIM(STR( m.tnError )) + " " + m.tcMessage + "]")
	loNoteText = CREATEOBJECT ("PChar", m.tcErrorlog)

	* initializing buffer with single recipient data
	lcRcpBuf = num2dword(0) +;
		num2dword(MAPI_TO) +;
		num2dword(loRcpEmail.getAddr()) +;
		REPLI(CHR(0), 12)
	loRcpBuf = CREATEOBJECT ("PChar", lcRcpBuf)

	* initializing buffer with sender data -- practically empty
	loSndBuf = CREATEOBJECT ("PChar", REPLI(CHR(0), 24))

	* merging all parts to a message buffer -- no file attachments
	lcMapiMessage = num2dword(0) +;
		num2dword(loSubject.getAddr()) +;
		num2dword(loNoteText.getAddr()) +;
		num2dword(0) + num2dword(0) + num2dword(0) + num2dword(0) +;
		num2dword(loSndBuf.getAddr()) +;
		num2dword(1) +;
		num2dword(loRcpBuf.getAddr()) +;
		num2dword(0) + num2dword(0)

	* sending the message with or without a confirmation dialog
	lnResult = MAPISendMail(hSession, 0, @lcMapiMessage, 0, 0)

	* closing current MAPI session
	= MAPILogoff (hSession, 0, 0, 0)

	IF lnResult <> SUCCESS_SUCCESS
		* 1 MAPI_E_USER_ABORT
		* 2 MAPI_E_FAILURE
		* 3 MAPI_E_LOGIN_FAILURE
		* 5 MAPI_E_INSUFFICIENT_MEMORY
		* 6 MAPI_E_ACCESS_DENIED
		* 9 MAPI_E_TOO_MANY_FILES
		*10 MAPI_E_TOO_MANY_RECIPIENTS
		*15 MAPI_E_BAD_RECIPTYPE
		*18 MAPI_E_TEXT_TOO_LARGE
		*14 MAPI_E_UNKNOWN_RECIPIENT
		* ...
		RETURN "Es ist ein Fehler aufgetreten: " + TRANSFORM( lnResult, "999,999,999" )
	ELSE
		* Achtung: Auch eine Mail mit einem ungltigen
		* Empfnger kann erfolgreich verschickt werden!
		RETURN "Errorlog erfolgreich verschickt"
	ENDIF

	*--------------------------------------------------------------------
PROCEDURE  getNewSession()
	* creates a new MAPI session and returns its handle
	#DEFINE MAPI_LOGON_UI           1
	#DEFINE MAPI_NEW_SESSION        2
	#DEFINE MAPI_USE_DEFAULT       64
	#DEFINE MAPI_FORCE_DOWNLOAD  4096 && 0x1000
	#DEFINE MAPI_PASSWORD_UI   131072 && 0x20000

	LOCAL lnResult, lnSession, lcStoredPath
	lcStoredPath = SYS(5) + SYS(2003)
	lnSession = 0

	TRY
		lnResult = MAPILogon (0, "", "",;
			MAPI_USE_DEFAULT+MAPI_NEW_SESSION, 0, @lnSession)
	CATCH
		lnResult = -1
	ENDTRY

	RETURN IIF(lnResult=SUCCESS_SUCCESS, lnSession, 0)

PROCEDURE  num2dword (lnValue)
	#DEFINE m0       256
	#DEFINE m1     65536
	#DEFINE m2  16777216
	LOCAL b0, b1, b2, b3
	b3 = INT(lnValue/m2)
	b2 = INT((lnValue - b3*m2)/m1)
	b1 = INT((lnValue - b3*m2 - b2*m1)/m0)
	b0 = MOD(lnValue, m0)
	RETURN CHR(b0)+CHR(b1)+CHR(b2)+CHR(b3)

	*-- for some structures you need not just strings but pointers to strings
	*-- to be assigned to structure fields;
	*-- this class implements such "dual" strings
DEFINE CLASS PChar AS CUSTOM
	PROTECTED HMEM

	PROCEDURE  INIT (lcString)
		THIS.HMEM = 0
		THIS.setValue (lcString)

	PROCEDURE  DESTROY
		THIS.ReleaseString

	PROCEDURE getAddr  && returns a pointer to the string
		RETURN THIS.HMEM

	PROCEDURE getValue && returns string value
		LOCAL lnSize, lcBuffer
		lnSize = THIS.getAllocSize()
		lcBuffer = SPACE(lnSize)

		IF THIS.HMEM <> 0
			DECLARE RtlMoveMemory IN kernel32 AS Heap2Str STRING @, INTEGER, INTEGER
			Heap2Str (@lcBuffer, THIS.HMEM, lnSize)
		ENDIF
		RETURN lcBuffer

	PROCEDURE getAllocSize  && returns allocated memory size (string length)
		DECLARE INTEGER GlobalSize IN kernel32 INTEGER HMEM
		RETURN IIF(THIS.HMEM=0, 0, GlobalSize(THIS.HMEM))

	PROCEDURE setValue (lcString) && assigns new string value
		#DEFINE GMEM_FIXED   0
		THIS.ReleaseString

		DECLARE INTEGER GlobalAlloc IN kernel32 INTEGER, INTEGER
		DECLARE RtlMoveMemory IN kernel32 AS Str2Heap INTEGER, STRING @, INTEGER

		LOCAL lnSize
		lcString = lcString + CHR(0)
		lnSize = LEN(lcString)
		THIS.HMEM = GlobalAlloc (GMEM_FIXED, lnSize)
		IF THIS.HMEM <> 0
			Str2Heap (THIS.HMEM, @lcString, lnSize)
		ENDIF

	PROCEDURE ReleaseString  && releases allocated memory
		IF THIS.HMEM <> 0
			DECLARE INTEGER GlobalFree IN kernel32 INTEGER
			GlobalFree (THIS.HMEM)
			THIS.HMEM = 0
		ENDIF
ENDDEFINE

PROCEDURE  DECL
	DECLARE INTEGER MAPILogon IN mapi32 INTEGER ulUIParam, STRING lpszProfileName,;
		STRING lpszPassword, INTEGER flFlags, INTEGER ulReserved, INTEGER @lplhSession
	DECLARE INTEGER MAPILogoff IN mapi32 INTEGER lhSession, INTEGER ulUIParam,;
		INTEGER flFlags, INTEGER ulReserved
	DECLARE INTEGER MAPISendMail IN mapi32 INTEGER lhSession, INTEGER ulUIParam, STRING @lpMessage,;
		INTEGER flFlags, INTEGER ulReserved
ENDPROC


PROCEDURE exe_version
	LPARAMETERS llLong
	LOCAL ARRAY laInfo[1],laDat[1]
	LOCAL lnDat,lcVersion,lcEXE,N
	lcEXE = GetEXEFile()
	N=ADIR(laDat,lcEXE)
	IF N>0
		IF laDat[1,2]<100000
			lcEXE = FULLPATH( BKI_KP_EXE )
		ENDIF
	ENDIF
	AGETFILEVERSION(laInfo,lcEXE)

	AGETFILEVERSION(laInfo,SYS(16))
	lcVersion = ""
	TRY
		IF llLong
			lcVersion = laInfo[10]+" ("+laInfo[4]+") "+laDat[i,1]
		ELSE
			lcVersion = (ALLTRIM(laInfo[4]))
		ENDIF
	CATCH
	ENDTRY
	RETURN (lcVersion)
ENDPROC
