*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Startprogramm der Kundenverwaltung.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 12.02.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 13.04.21 HPG: Wir mssen hier erkennen, welchen Fall wir zur Berechnung der Versand-Gewichtes haben.
*				 - Es ist eine Versandkosten-Position vorhanden:
*				   Artikel-Nummern: 9982, 9992, 9993, 9994, 9995 (in gcShippingCostPosition)
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 12.02.20 HPG: Das Einlesen der Ausschluliste ins Kunden-Formular verschoben.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

#INCLUDE	..\INCLUDE\stinclude.h

LOCAL lcSystem AS STRING, lcValue AS STRING

*-- Parameter fr die RestAPI: Sind die RestAPI-Funktionen aktiviert
PUBLIC glUseRestAPI AS BOOLEAN

*-- Globales Applikations-Objekt
PUBLIC oApp AS OBJECT

*-- Globale Zeichenkette fr CheckName
PUBLIC gcAusschluss AS STRING

*-- DPD-Import-Verzeichnis
PUBLIC gcDPDAbo AS STRING
PUBLIC gcDPDLog AS STRING
PUBLIC gcDPDImport AS STRING
PUBLIC gcDPDPrinter AS STRING
PUBLIC glDPDMessageBox AS BOOLEAN

*-- INI-Datei fr die Formular-Position
PUBLIC gcINIFileName AS STRING

*-- Stammdaten
PUBLIC goStamm AS BKIStamm

*-- Weitere globale Parameter
PUBLIC gu_user AS STRING, gu_trace_mode AS STRING, gu_trace_alias AS STRING, gu_user_name AS STRING, gu_user_email AS STRING
PUBLIC gcMachineName AS STRING, gcWinUserName AS STRING, gcParcelService AS STRING
PUBLIC gu_trace_record AS INTEGER, gu_trace_id AS INTEGER, gnInactivityTime AS INTEGER
PUBLIC gu_trace AS BOOLEAN
PUBLIC goTimer AS OBJECT
PUBLIC glCheckPassword AS BOOLEAN

*-- Artikel-Nummern der Versandkosten-Position mit hinterlegtem
*-- Gewicht sowie die Liste fr die Verpackungsgewicht-ComboBox
PUBLIC gcShippingCostPosition AS STRING
PUBLIC gcShippingCostList AS STRING

*-- Fr Rechnungen per EMail
PUBLIC gcSubjectInvoice AS STRING
PUBLIC gcSubjectWarning AS STRING

*-- Datentypen fr die Werbestop-Tabelle
PUBLIC gcPromotionStopText AS STRING

*-- 18.11.24 HPG: Artikel-Nummern fr KA2-Artikel
PUBLIC gcKA2Article AS STRING

*-- Die minimale Verzgerung beim Suchen, um einen Log-Eintrag zu erstellen
PUBLIC gnMinSeekDelay AS Integer

*-- Fr die Rechnungsarten:
*-- Konfiguration des Berichtes Rechnung.frx
*-- 1 = Rechnung
*-- 2 = Rechnung Kreditkarte
*-- 3 = Rechnung ohne Rckgabe
*-- 4 = Rechnung zur Ansicht
*-- 5 = Rechnung Abo - ohne Bestellzeichen
*-- 6 = Rechnungs-Korrektur - Storno
PUBLIC gnPrintMode AS INTEGER
gnPrintMode = 1

*-- Parameter fr Druck-Art: 1 = Rechnung, 2 = Rechnungskopie, 3 = Rechnungskorrektur
PUBLIC gnPrintLabel AS INTEGER
gnPrintLabel = 1

*-- Parameter fr Lieferungen ins Ausland
PUBLIC gcTaxInfo AS STRING

*-- Parameter fr Flextext
PUBLIC gcFlexText AS STRING

*-- Windows-User und Rechner-Name
lcSystem = SYS( 0 )
gcMachineName = ALLTRIM( GETWORDNUM( lcSystem, 1 , '#' ) )
gcWinUserName = ALLTRIM( GETWORDNUM( lcSystem, 2 , '#' ) )

*-- INI-Datei fr die Formular-Position setzen
gcINIFileName = FULLPATH( 'KDV.ini' )

*-- Pfad setzen
SET PATH TO BERICHTE, BMPS, CLASS, ERROR, FORM, INCLUDE, LIB, MENU, PROGRAM, TOOLS ADDITIVE

*-- Error-Handler erzeugen
ON ERROR DO ErrorHandler WITH ERROR(), MESSAGE(), MESSAGE( 1 ), PROGRAM(), LINENO( 1 ), WOUTPUT(), ;
	SYS( 2018 ), SYS( 16 ), LASTKEY(), VARREAD(), DBF(), ALIAS(), SET( 'ORDER' ), SET( 'DATASESSION' )

*-- VFP-Berichte verwenden
SET REPORTBEHAVIOR 80

*-- SET PROCEDURE setzen
Init_Proc()

*-- Applikations-Klasse 
oApp = CREATEOBJECT( 'bki_kdv_app' )

*-- Parameter setzen
gnInactivityTime = 3600
gcAusschluss	= 1
gcDPDAbo		= ''
gcDPDImport		= ''
gu_trace_mode	= ''
gu_trace_record = 0
gu_trace_id		= 0
gu_trace_alias	= ''
gu_trace		= .T.				&& Logfunktion freischalten
gu_user			= 'BKI'				&& bki_user.krzel: Username (Krzel)
gu_user_email	= ''				&& EMail des Anwenders
glCheckPassword = .T.

*-- 256 MB als Vordergrundspeicher und Hintergrundspeicher verwenden
lnMaxRam = 256000000
SYS( 3050, 1, lnMaxRam )
SYS( 3050, 2, lnMaxRam )

*-- Parameter fr die RestAPI: Sind die RestAPI-Funktionen aktiviert
*-- Parameter in der RestAPI-Funktion
lcValue = getKDVini( 'RestAPI', 'UseRestAPI' )
IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
	glUseRestAPI = IIF( lcValue = '1', .T., .F. )
ELSE
	glUseRestAPI = .F.
ENDIF

*-- Den Login-Modus holen
lcValue = getKDVini( 'Login', 'LoginMode' )
IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
	glCheckPassword = IIF( lcValue == 'WriteBack', .F., .T. )
ELSE
	glCheckPassword = .T.
ENDIF

*-- Folgendes nur in der IDE
IF VERSION( 2 ) != 0

	*-- Caption setzen
	_SCREEN.CAPTION = 'BKI - Kundenverwaltung'

*-- Folgendes nur in der EXE
ELSE

	LOCAL ARRAY laEXE1[1], laEXE2[1]
	AGETFILEVERSION( laEXE1, SYS( 16 ) )
	ADIR( laEXE2, SYS( 16 ) )
	SET DATE GERMAN
	_SCREEN.CAPTION = 'BKI - Kundenverwaltung (' + laEXE1[4] + ' vom ' + DTOC( laEXE2[1, 3] ) + ') - VFP-RT: ' + ALLTRIM( VERSION( 4 ) )
ENDIF
_SCREEN.WINDOWSTATE = 2

*-- Das Hauptmen starten
DO Kdv.mpr

*-- SHUTDOWN-Prozedure setzen
ON SHUTDOWN DO Quit_Kdv

*-- Ereignis-Verarbeitung starten
READ EVENTS

RETURN


*----------------------------------------------------
PROCEDURE Error_FreiSchaltCode
*----------------------------------------------------

	LPARAMETERS lcExpr AS STRING

	MESSAGEBOX( 'Fehler in Formel Freischaltung:' + CHR( 13 ) + lcExpr, 32, 'Fehler' )
ENDPROC


*----------------------------------------------------
DEFINE CLASS bki_kdv_app AS CUSTOM
*----------------------------------------------------

	lImport			= .F.
	lLiveSystem		= .T.
	lEXEVersion		= .T.
	cEXEPath		= ''
	cURLTestSystem	= 'HTTPS://BKI.FUF.DE'
	cURLLiveSystem	= 'HTTPS://WWW.BKI.DE'

	PROCEDURE INIT()

		LOCAL lcValue AS STRING, lcError AS STRING, lcINIFileName AS STRING, lcURL4Customer AS STRING, lcURLINI AS STRING
		LOCAL lnError AS INTEGER, lnURLLen AS INTEGER
		LOCAL llCheckDB AS BOOLEAN

		*-- Folgendes nur in der IDE
		IF VERSION( 2 ) != 0

			*-- Wir laufen in der Runtime
			THIS.lEXEVersion = .F.
			THIS.cEXEPath = 'RUNTIME'

		*-- Folgendes nur in der EXE
		ELSE

			*-- Wir laufen in der EXE
			THIS.lEXEVersion = .T.
			THIS.cEXEPath = SYS( 16 )
		ENDIF

		*-- Eintrag CheckDBOn auslesen
		lcValue = getKDVini( 'RestAPI', 'CheckDBOn' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			llCheckDB = ( lcValue == '1' )
		ELSE
			llCheckDB = .F.
		ENDIF

		*-- In welcher Umgebung laufen wir und
		*-- ist die Datenbankprfung aktiviert?
		IF THIS.lEXEVersion .AND. llCheckDB

			*-- Anhand des Pfades zur KDV.ini bestimmen wir,
			*-- ob wir auf dem BKI-System arbeiten und ob
			*-- wir im LiveSystem oder im TestSystem sind.
			*-- Bei HPG luft beides unter D:\ und somit
			*-- trifft der Vergleich nicht zu und es wird
			*-- nicht auf die korrekte Datenbank geprft.
			lcINIFileName = UPPER( gcINIFileName )
			IF LEFT( lcINIFileName, 3 ) == 'V:\'

				*-- Wir arbeiten auf dem BKI-System
				*-- Arbeiten wir im LiveSystem?
				IF OCCURS( '\', gcINIFileName ) = 2

					*-- Ja
					THIS.lLiveSystem = .T.

				*-- Andernfalls auf dem TestSystem
				ELSE
					THIS.lLiveSystem = .F.
				ENDIF

				*-- Parameter setzen
				lcError = ''
				lnError = 0

				*-- Wir laufen in der EXE-Version
				*-- Eintrge aus der INI-Datei auslesen
				*-- Fr Aufruf der RestAPI-Funktionen
				*-- Eintrag URLTestSystem auslesen
				lcValue = getKDVini( 'RestAPI', 'URLTestSystem' )
				IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
					IF UPPER( lcValue ) <> UPPER( THIS.cURLTestSystem )
						lcError = lcError + 'Eintrag >URLTestSystem< aus der KDV.ini stimmt nicht mit dem EXE-Eintrag berein' + CR
						lnError = lnError + 1
					ENDIF
				ELSE
					lcError = lcError + 'Eintrag >URLTestSystem< wurde in der KDV.ini nicht gefunden' + CR
					lnError = lnError + 1
				ENDIF

				*-- Eintrag URLLiveSystem auslesen
				lcValue = getKDVini( 'RestAPI', 'URLLiveSystem' )
				IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
					IF UPPER( lcValue ) <> UPPER( THIS.cURLLiveSystem )
						lcError = lcError + 'Eintrag >URLLiveSystem< aus der KDV.ini stimmt nicht mit dem EXE-Eintrag berein' + CR
						lnError = lnError + 1
					ENDIF
				ELSE
					lcError = lcError + 'Eintrag >URLLiveSystem< wurde in der KDV.ini nicht gefunden' + CR
					lnError = lnError + 1
				ENDIF

				*-- Wenn bisher alles OK ist
				IF lnError = 0

					*-- Eintrag URL4Customer auslesen
					lcValue = getKDVini( 'RestAPI', 'URL4Customer' )
					IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
						lcURL4Customer = UPPER( lcValue )

						*-- Arbeiten wir im LiveSystem?
						IF THIS.lLiveSystem

							*-- Ja
							lnURLLen = LEN( THIS.cURLLiveSystem )
							lcURLINI = LEFT( lcURL4Customer, lnURLLen )

							*-- Jetzt den Eintrag in der INI-Datei gegenber
							*-- dem Soll-Eintrag prfen
							IF lcURLINI <> THIS.cURLLiveSystem
								lcError = lcError + 'Die EXE luft im LiveSystem, der RestAPI-Link zeigt aber auf das TestSystem' + CR
								lnError = lnError + 1
							ENDIF

						*-- Nein, im TestSystem
						ELSE
							lnURLLen = LEN( THIS.cURLTestSystem )
							lcURLINI = LEFT( lcURL4Customer, lnURLLen )

							*-- Jetzt den Eintrag in der INI-Datei gegenber
							*-- dem Soll-Eintrag prfen
							IF lcURLINI <> THIS.cURLTestSystem
								lcError = lcError + 'Die EXE luft im TestSystem, der RestAPI-Link zeigt aber auf das LiveSystem' + CR
								lnError = lnError + 1
							ENDIF
						ENDIF
					ELSE
						lcError = lcError + 'Eintrag >URL4Customer< wurde in der KDV.ini nicht gefunden' + CR
						lnError = lnError + 1
					ENDIF
				ENDIF

				*-- Auf Fehler prfen und anzeigen
				IF lnError > 0
					Fehler( 'Es gibt ein Problem bei der Konfiguration der RestAPI-Funktionen' + CR + CR + StrDelR( lcError ) )
				ENDIF
			ENDIF

		*-- Andernfalls in der IDE
		ELSE

		ENDIF
	ENDPROC

ENDDEFINE


*----------------------------------------------------
PROCEDURE Init_Proc
*----------------------------------------------------

	*-- 21.10.20 HPG: AutoQuit bei Inaktivitt
	SET PROCEDURE TO PROGRAM\AUTOQUIT.PRG ADDITIVE

	*-- PROZEDUREn
	SET PROCEDURE TO PROGRAM\COMMON.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\ERRORHANDLER.PRG ADDITIVE

	*-- 09.01.20 HPG: Routing-Label-Optimierer hinzugefgt
	SET PROCEDURE TO PROGRAM\QUIT_KDV.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\REGLIB.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WINAPI.PRG ADDITIVE

	*-- 28.04.23 HPG: Fr die West-Wind-Tools zur SFTP-bertragung bei der Freischaltung
	SET PROCEDURE TO PROGRAM\WWAPI.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WWDOTNETBRIDGE.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WWFTP.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WWFTPCLIENT.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WWSFTP.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WWSFTPCLIENT.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WWUTILS.PRG ADDITIVE

	*-- 05.07.23 Fr die West-Wind-Tools zur RestAPI
	SET PROCEDURE TO PROGRAM\WWCOLLECTIONS.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WWHTTP.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WWJSONSERIALIZER.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WWJSONSERVICE.PRG ADDITIVE
	SET PROCEDURE TO PROGRAM\WWJSONSERVICECLIENT.PRG ADDITIVE

	*-- Ends the calling process and all its threads.
	DECLARE ExitProcess IN KERNEL32.DLL INTEGER uExitCode

	*-- To enter an alertable wait state, use the SleepEx function.
	DECLARE INTEGER Sleep IN WIN32API INTEGER dwMilliseconds

	*-- Name der EXE
	DECLARE INTEGER GetModuleFileName IN kernel32 INTEGER hModule, STRING @lpFilename, INTEGER nSize

	*-- ShellExecute
	DECLARE INTEGER ShellExecute IN shell32.DLL INTEGER hndWin, STRING cAction, STRING cFileName, STRING cParams, STRING cDir, ;
		INTEGER nShowWin
	DECLARE LONG FindWindowA IN USER32 AS FindA STRING, STRING
	DECLARE LONG SendMessageA IN USER32 AS SendA LONG, LONG, LONG, LONG

	*-- Retrieves the path of the directory designated for temporary files.
	DECLARE INTEGER GetTempPath IN KERNEL32.DLL INTEGER nBufferSize, STRING @lpBuffer

	*-- Gibt den letzten Fehlercode eines Threads zurck
	DECLARE INTEGER GetLastError IN KERNEL32.DLL

	*-- The CreateProcess function creates a new process and its primary thread.
	DECLARE INTEGER CreateProcess IN KERNEL32.DLL STRING lpApplicationName, STRING lpCommandLine, INTEGER lpProcessAttributes, ;
		INTEGER lpThreadAttributes, INTEGER bInheritHandles, INTEGER dwCreationFlags, INTEGER lpEnvironment, ;
		STRING lpCurrentDirectory, STRING @lpStartupInfo, STRING @lpProcessInformation

	*-- The WaitForSingleObject function returns when one of the following occurs: The specified object is in the signaled
	*-- state. The time-out interval elapses.
	DECLARE INTEGER WaitForSingleObject IN KERNEL32.DLL INTEGER hHandle, INTEGER dwMilliseconds

	*-- Schliet ein offenes Objekt- bzw. Datei-Handle
	DECLARE INTEGER CloseHandle IN KERNEL32.DLL INTEGER hObject

	*-- Name der EXE
	DECLARE INTEGER GetModuleFileName IN kernel32 INTEGER hModule, STRING @lpFilename, INTEGER nSize

	*-- 22.10.24 HPG: API-Funktion fr den QRCode
	DECLARE FastQRCode IN CLASS\QRCodeLib.DLL STRING cText, STRING cFileName

	*-- 19.11.24 HPG: Fr Funktion CreateBKIGUID
	DECLARE INTEGER CoCreateGuid IN Ole32.DLL STRING @lcGUIDStruc
	DECLARE INTEGER StringFromGUID2 IN Ole32.DLL STRING cGUIDStruc, STRING @cGUID, LONG nSize

ENDPROC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Error-Handler-Light der beim Start der Applikation gesetzt ist bis das globale Applikations-Objekt voll-
*				initialisiert ist.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 18.09.15
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 19.09.15 HPG: In der Entwicklungs-Umgebung bieten wir das Starten des Debuggers an.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

PROCEDURE stErrorLight
**********************

	LPARAMETERS tnErrorNumber AS INTEGER, tcMessage AS STRING, tcErrorLine AS STRING, tnLineNumber AS INTEGER, tcProgramm AS INTEGER

	LOCAL lcErrorMessage AS STRING, lcErrorFile AS STRING, lcPCName AS STRING
	LOCAL lnOption AS INTEGER, lnChoice AS INTEGER

	*-- Namen des PC's ermitteln
	lcPCName = ALLTRIM( LEFT( SYS( 0 ), AT( '#', SYS( 0 ) ) - 1 ) )

	*-- Fehlermeldung aufbauen
	lcErrorMessage = 'Fehler-Nummer: ' + TRANSFORM( tnErrorNumber ) + CR + 'Fehler-Text: ' + tcMessage + CR + 'Fehler-Zeile: ' + ;
		tcErrorLine + CR + 'Zeilen-Nummer: ' + TRANSFORM( tnLineNumber ) + CR + 'Programm-Name: ' + tcProgramm + CR + CR + 'PC: ' + ;
		lcPCName

	*-- Fehlermeldungs-Text in Log-Datei schreiben
	*-- Den Pfad dabei aus dem Parameter tcErrorLine
	*-- holen und die CRs und LFs mit einem ' # '
	*-- ersetzen fr einen einzeiligen Eintrag.
	lcErrorFile = ADDBS( JUSTPATH( GETWORDNUM( tcErrorLine, 2, ' ' ) ) ) + 'FPR_FEHLERPROTOKOLL.TXT'
	STRTOFILE( TTOC( DATETIME() ) + ': ' + STRTRAN( STRTRAN( lcErrorMessage, CR, ' # ' ), LF, '' ) + CR, lcErrorFile, 1 )

	*-- Sind wir in der Entwicklungs-Umgebung?
	IF VERSION( 2 ) <> 0

		*-- Ja, dann bieten dem Entwickler den Debugger an
		lcErrorMessage = lcErrorMessage + CR + CR + 'Wollen Sie den Debugger starten oder wollen Sie abbrechen?'
		lnOption = 19

	*-- Andernfalls ...
	ELSE

		*-- ... kann der Anwender nur Abbrechen (das Programm Beenden)
		lnOption = 16
	ENDIF

	*-- Fehlermeldung ausgeben
	lnChoice = MESSAGEBOX( lcErrorMessage, lnOption, 'Fehler' )

	*-- Was hat der Anwender ausgewhlt?
	DO CASE

		*-- Ja ausgewhlt
		CASE lnChoice = 6

			*-- Das Debugger-Fenster ffnen und den Debugger starten
			ACTIVATE WINDOW DEBUG
			SET STEP ON					&& OK

		*-- Nein ausgewhlt
		CASE lnChoice = 7

			*-- Wir beenden wir das Programm
			*-- Wiedereinsprung verhindern
			ON SHUTDOWN QUIT

			*-- Alles Schlieen und das Programm beenden
			CLOSE ALL
			QUIT

		*-- Abbrechen ausgewhlt
		CASE lnChoice = 2

			*-- Versuchen weiter zu machen
	ENDCASE

	RETURN

ENDPROC
