SET PROCEDURE TO ..\PROGRAM\wwFtpClient ADDITIVE
DO wwDotnetBridge

#IF .F.

	CLEAR
	loFtp  = CREATEOBJECT("wwFtpClient")
	loFtp.lUseTls = .T.
	loFtp.cLogFile = "c:\temp\ftp.log" && verbose log - leave empty
	loFtp.lIgnoreCertificateErrors = .F.   && self-signed cert not installed


	SET PROCEDURE TO wwutils ADDITIVE  && for UI helpers

	*** cServer can be "someserver.com", "someserver.com:22", "123.213.222.111"
	lcServer =  INPUTBOX("Server Domain/IP")
	IF EMPTY(lcServer)
		RETURN
	ENDIF

	lcUsername = INPUTBOX("User name")
	IF EMPTY(lcUsername)
		RETURN
	ENDIF

	lcPassword = GetPassword("Password")
	IF EMPTY(lcPassword)
		RETURN
	ENDIF

	loFtp.oProgressEventObject = CREATEOBJECT("FtpClientProgressEvents")

	loFtp.cServer = lcServer
	loFtp.cUsername = lcUsername
	loFtp.cPassword = lcPassword
	*loFtp.nPort = 21 && only needed if custom port is required

	IF !loFtp.CONNECT()
		? loFtp.cErrorMsg
		RETURN
	ENDIF
	? "Connected to " + lcServer

	loFtp.EXISTS("Tools/jsMinifier1.zip")


	IF !loFtp.DownloadFile("Tools/jsMinifier.zip", "c:\temp\jsMinifier.zip")
		? loFtp.cErrorMsg
		RETURN
	ENDIF
	? "Downloaded " + "Tools/jsMinifier.zip"

	lcUploadFile = "Tools/jsMinifier" + SYS(2015) + ".zip"
	IF !loFtp.UploadFile("c:\temp\jsMinifier.zip", lcUploadFile)
		? loFtp.cErrorMsg
		RETURN
	ENDIF
	? "Uploaded " + lcUploadFile

	*** provide a folder name (no wildcards)
	loCol = loFtp.ListFiles("/Tools")
	IF ISNULL(loCol)
		? "Error: " + loFtp.cErrorMsg
		RETURN
	ENDIF
	? TRANSFORM(loCol.COUNT ) + " matching file(s)"
	? loFtp.cErrorMsg
	FOR EACH loFile IN loCol FOXOBJECT
		IF ( AT("jsMinifier_",loFile.NAME) = 1)
			? loFtp.oBridge.ToJson(loFile)  && for kicks print out as json
			IF loFtp.DeleteFile(loFile.FULLNAME)
				? "Deleted " + loFile.FULLNAME
			ENDIF

		ENDIF
	ENDFOR

	loFiles = loFtp.ListFiles("/Tools")
	FOR EACH loFile IN loFiles
		? loFile.NAME + " " + TRANSFORM(loFile.LastWriteTime)
	ENDFOR


	* loFtp.Close()  && automatic when released

	RETURN


DEFINE CLASS FtpClientProgressEvents AS CUSTOM

	FUNCTION OnFtpBufferUpdate(lnPercent, lnDownloadedBytes, lcRemotePath, lcMode)
		lcMsg = lcMode + ": " + TRANSFORM(lnPercent) + "% complete. " + lcRemotePath + " - " + TRANSFORM(lnDownloadedBytes) + " bytes"
		? "*** " + lcMsg
	ENDFUNC

ENDDEFINE
#ENDIF



#INCLUDE ..\INCLUDE\WCONNECT.H

*************************************************************
DEFINE CLASS wwFtpClient AS CUSTOM
	*************************************************************
	*: Author: Rick Strahl
	*:         (c) West Wind Technologies, 2011
	*:Contact: http://www.west-wind.com
	*:Created: 08/30/2011
	*************************************************************
	#IF .F.
		*:Help Documentation
		*:Topic:
		CLASS wwFtpClient

		*:Description:
		Ftp CLASS that uses the .NET Fluent Ftp LIBRARY TO provide
		Ftp AND FTPS (Ftp OVER TLS). THIS CLASS **does NOT SUPPORT SFTPS**.
		*:ENDHELP
	#ENDIF

	*** Domain name or IP address for the server
	cServer = ""

	*** The port to run the FTP connection on
	nPort = 21


	*** Tsl used for this request (FTPS protocol: FTP over TLS)
	lUseTls = .F.

	*** Allows you to ignore certificate errors **NOT RECOMMENDED: Fix your Certs!**
	lIgnoreCertificateErrors = .F.

	*** Username for the connection
	cUsername = ""

	*** Password for the connection
	cPassword = ""

	*** Connection timeout in seconds
	nTimeout = 10

	*** The FTP server to connect to
	cftpserver = ""

	*** The Source file to upload from the client or download from the server.
	cftpsource = ""

	*** The file to download to from the server or upload to on the server.
	cftptarget = ""

	*** Flag that sets whether to use passive FTP.
	lpassiveftp = .T.

	lEnableProgressEvents = .F.

	*** Optional log file that receives ftp log output
	cLogFile = ""

	cHttpProxyName = ""
	cHttpProxyPort = 8888
	cHttpProxyPassword = ""
	cHttpProxyUsername = ""
	cHttpProxyByPass = ""

	*** Cancel flag that can be set in OnFTPBufferUpdate to let wwFTP know to stop downloading a file from the Internet.
	lcanceldownload = .F.

	*** Error message set after a method call that fails
	cErrorMsg = ""

	oProgressEventObject = NULL

	*** wwDotnetBridge instance
	oBridge = NULL

	*** .NET wwFtpClient instance
	oFtpClient = NULL

	************************************************************************
	*  Init
	****************************************
	***  Function:
	***    Assume:
	***      Pass:
	***    Return:
	************************************************************************
	PROCEDURE INIT()

		#IF SHAREWARE
			WAIT WINDOW NOWAIT ;
				"Welcome to West Wind Client Tools..." + CHR(13) + ;
				"This is a shareware copy of wwFtpClient." + CHR(13)+ CHR(13) +;
				"Please register your copy."
		#ENDIF

		IF ISNULL(THIS.oBridge)
			THIS.oBridge = GetwwDotnetBridge()
		ENDIF

		THIS.oFtpClient = THIS.oBridge.CreateInstance("Westwind.WebConnection.wwFtpClient")
		IF ISNULL(THIS.oFtpClient)
			*** Throw an error with the actual error message
			ERROR THIS.oBridge.cErrorMsg
		ENDIF

	ENDFUNC


	************************************************************************
	*  Destroy
	****************************************
	***  Function: Closes down connection if still open
	***    Assume:
	***      Pass:
	***    Return:
	************************************************************************
	PROCEDURE DESTROY
		THIS.CLOSE()
	ENDFUNC




	************************************************************************
	*  OnFtpBufferUpdate
	****************************************
	***  Function: Event that fires whenever the buffer is updated by
	***            downloadfile or uploadfile.
	***      Pass:
	***    Return: lcMode: upload download
	************************************************************************
	FUNCTION OnFtpBufferUpdate(lnPercent, lnDownloadedBytes, lcRemotePath, lcMode)

	ENDFUNC


	************************************************************************
	*  FtpConnect
	****************************************
	***  Function:
	***    Assume:
	***      Pass:
	***    Return:
	************************************************************************
	FUNCTION CONNECT(lcServer, lcUsername, lcPassword)
		LOCAL lhIP, lhHTTP, lnError, lnHTTPPort


		IF ISNULL(THIS.oFtpClient)
			THIS.errorMsg = "No active FtpClient instance available."
			RETURN .F.
		ENDIF

		lcServer= LOWER(TRIM(IIF(!EMPTY(lcServer),lcServer,THIS.cServer)))
		lcUsername= TRIM(IIF(!EMPTY(lcUsername),lcUsername,THIS.cUsername))
		lcPassword= TRIM(IIF(!EMPTY(lcPassword),lcPassword,THIS.cPassword))

		IF lcServer = "ftp://"
			lcServer = STRTRAN(lcServer,"ftp://","")
		ENDIF

		*** Assign Default Ports
		IF THIS.nPort = 0
			lnHTTPPort = INTERNET_DEFAULT_FTP_PORT
		ELSE
			lnHTTPPort = THIS.nPort
		ENDIF

		THIS.oFtpClient.HOSTNAME = lcServer
		THIS.oFtpClient.Port = THIS.nPort

		IF !EMPTY(lcUsername)
			THIS.oFtpClient.Username = lcUsername
			THIS.oFtpClient.PASSWORD = lcPassword
		ENDIF

		THIS.oFtpClient.UseTls = THIS.lUseTls

		IF !EMPTY(THIS.cHttpProxyName)
			THIS.oFtpClient.ProxyHostname = THIS.cHttpProxyName
			THIS.oFtpClient.ProxyPort = THIS.nHttpProxyPort
			THIS.oFtpClient.ProxyUsername = THIS.cHttpProxyUsername
			THIS.oFtpClient.ProxyPassport = THIS.cHttpProxyPassport
		ENDIF

		THIS.oFtpClient.IgnoreCertificateErrors = THIS.lIgnoreCertificateErrors
		THIS.oFtpClient.LogFile = THIS.cLogFile
		THIS.oFtpClient.TIMEOUT = THIS.nTimeout * 1000

		IF !ISNULL(THIS.oProgressEventObject)
			THIS.oFtpClient.FoxProwwFtpClientForProgress = THIS.oProgressEventObject
		ENDIF

		IF ISNULL(THIS.oFtpClient.CONNECT())
			*** Error
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF

		RETURN .T.
	ENDFUNC


	************************************************************************
	*  FtpClose
	****************************************
	***  Function: Closes an open Ftp Connection
	***    Assume:
	***      Pass:
	***    Return:
	************************************************************************
	FUNCTION CLOSE()

		IF !ISNULL(THIS.oFtpClient)
			THIS.oFtpClient.CLOSE()
		ENDIF
	ENDFUNC


	************************************************************************
	*  AddClientCertificate
	****************************************
	***  Function: Adds a client certificate
	***    Assume: Call before call to Connect()
	***      Pass: Certificate file (.cer)
	***    Return: .T. or .F.
	************************************************************************
	FUNCTION AddCertificate(lcCertFile)

		IF !THIS.oFtpClient.AddCertificate(FULLPATH(lcCertFile))
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF

		RETURN .T.
	ENDFUNC
	*   AddClientCertificate


	************************************************************************
	*  DownloadFile
	****************************************
	***  Function: Copies a file from the server to the client
	***    Assume: Connection is open already with FtpOpen
	***      Pass: lcSourceFile - Server relative path to a file to download
	***            lcTargetFile - Full path to a local file to download to
	***    Return: 0 on success - error number of failure
	************************************************************************
	FUNCTION DownloadFile(lcSourceFile, lcTargetFile)

		IF !THIS.oFtpClient.DownloadFile(lcSourceFile, lcTargetFile)
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF

		RETURN .T.
	ENDFUNC

	************************************************************************
	*  UploadFile
	****************************************
	***  Function: Sends a file to the server
	***    Assume: Connection to the server is already open
	***      Pass: lcSourceFile  - full path to file to upload
	***            lcTargetFile  - Server relative path to file on server
	***    Return: 0 or error code
	************************************************************************
	FUNCTION UploadFile(lcSourceFile, lcTargetFile)

		IF !THIS.oFtpClient.UploadFile(lcSourceFile, lcTargetFile)
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF

		RETURN .T.
	ENDFUNC



	************************************************************************
	*  FtpDeleteFile
	****************************************
	***  Function: Deletes a file from the server
	***    Assume: Connection must be open
	***      Pass: lcFile - server relative file to delete
	***    Return:
	************************************************************************
	FUNCTION DeleteFile(lcFile)

		IF !THIS.oFtpClient.DeleteFile(lcFile)
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF

		RETURN .T.
	ENDFUNC

	************************************************************************
	*  Exists
	****************************************
	***  Function:
	***    Assume:
	***      Pass:
	***    Return:
	************************************************************************
	FUNCTION EXISTS(lcFile)

		IF !THIS.oFtpClient.EXISTS(lcFile)
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF

		RETURN .T.
	ENDFUNC
	*   Exists


	************************************************************************
	*  FtpRenameFile
	****************************************
	***  Function: Renames a file on the FTP Server
	***    Assume:
	***      Pass:
	***    Return:
	************************************************************************
	FUNCTION RenameFile(lcOldName, lcNewName)

		IF !THIS.oFtpClient.RenameFile(lcOldName, lcNewName)
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF

		RETURN .T.
	ENDFUNC



	************************************************************************
	*  FtpListFiles
	****************************************
	***  Function: Retrieves an FTP directory file listing into a Collection
	***    Assume: connection must be open with FtpOpen() first
	***      Pass: lcFileSpec - file spec (can include wildcards)
	***    Return: Collection or null Collection may be empty
	***            if no files are found
	************************************************************************
	FUNCTION ListFiles(lcfilespec)
		LOCAL loCol, loFile, loFileItems, lnCount, lnX

		loFileItems = THIS.oBridge.InvokeMethod(THIS.oFtpClient,"ListFiles",lcfilespec)

		IF ISNULL(loFileItems)
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN NULL
		ENDIF

		lnCount = loFileItems.COUNT

		loCol = CREATEOBJECT("Collection")
		FOR lnX = 0 TO lnCount -1
			loFile = loFileItems.ITEM(lnX)
			loCol.ADD(loFile)
		ENDFOR

		RETURN loCol
	ENDFUNC
	* EOF ListFiles()

	************************************************************************
	*  CreateDirectory
	****************************************
	***  Function:
	***    Assume:
	***      Pass:
	***    Return:
	************************************************************************
	FUNCTION CreateDirectory()
		LPARAMETER lcPath

		IF EMPTY(lcPath)
			RETURN .F.
		ENDIF

		DECLARE INTEGER FtpCreateDirectory ;
			IN WININET.DLL ;
			INTEGER hConnect,;
			STRING lpszDirectory

		IF FtpCreateDirectory(THIS.hftpsession,lcPath)=0
			THIS.nError = GetLastError()
			THIS.cErrorMsg = THIS.GetSystemErrorMsg()
			RETURN .F.
		ENDIF

		RETURN .T.
	ENDFUNC

	************************************************************************
	*  DeleteDirectory
	****************************************
	***  Function: Removes a path from the server. Path is server relative
	***    Assume: Connection must be open with FtpOpen
	***      Pass: lcPath - Server relative path to remove
	***    Return: .T. or .F.
	************************************************************************
	FUNCTION DeleteDirectory(lcPath)

		IF !THIS.oFtpClient.DeleteDirectory(lcPath)
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF


		RETURN .T.
	ENDFUNC


	************************************************************************
	*  ChangeDirectory
	****************************************
	***  Function: Changes the directory to a new base location
	***      Pass: lcRemotePath - server relative path
	***    Return: .t. or .f.
	************************************************************************
	FUNCTION ChangeDirectory()

		IF !THIS.oFtpClient.DeleteDirectory(lcPath)
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF

		RETURN .T.
	ENDFUNC
	*   ChangeDirectory


	************************************************************************
	*  FtpCommand
	****************************************
	***  Function: Issue an arbitrary FTP Command
	***    Assume: Connection must exist with FtpOpen
	***      Pass:
	***    Return:
	************************************************************************
	FUNCTION ExecuteCommand(lcCommand)

		IF !THIS.oFtpClient.ExecuteCommand(lcCommand)
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF

	ENDFUNC


	************************************************************************
	*  ExecuteDownloadCommand
	****************************************
	***  Function: Issue an arbitrary FTP Command
	***    Assume: Connection must exist with FtpOpen
	***      Pass:
	***    Return:
	************************************************************************
	FUNCTION ExecuteDownloadCommand(lcCommand)

		IF !THIS.oFtpClient.ExecuteDownloadCommand(lcCommand)
			THIS.cErrorMsg = THIS.oFtpClient.ERRORMESSAGE
			RETURN .F.
		ENDIF

	ENDFUNC


ENDDEFINE
*
*** EndDefine: wwftp
**************************************************
