SET PROCEDURE TO ..\Program\wwSFtpClient ADDITIVE
DO wwDotnetBridge 

#IF .F.   && EXAMPLE Code
 
*** Using Rebex Tiny SFTP Server for local testing
lcServer = "127.0.0.1"
lnPort = 23
lcUsername = "tester"
lcPassword = "password"

CLEAR

? "***  wwSftpClient Test***"
loFtp  = CREATEOBJECT("wwsFtpClient")
loFtp.cLogFile = "c:\temp\ftp.log" && verbose log - leave empty
loFtp.lIgnoreCertificateErrors = .T.   && self-signed cert not installed

loFtp.cServer = lcServer
loFtp.nPort = lnPort
loFtp.cUsername = lcUsername
loFtp.cPassword = lcPassword


*** Optional Events
loFtp.oProgressEventObject = CREATEOBJECT("SFtpClientProgressEvents")

SET PROCEDURE TO wwutils ADDITIVE  && for UI helpers

IF !loFtp.Connect()
	? loFtp.cErrorMsg
	RETURN
ENDIF
? "Connected to " + lcServer	


IF !loFtp.DownloadFile("SubFolder/Sail-Big.jpg", "c:\temp\Sail-Big.jpg")
	? loFtp.cErrorMsg
	RETURN
ENDIF	
? "Downloaded " + "SubFolder/Sail-Big.jpg"

lcUploadFile = "SubFolder/Sail-Big-UPLOADED" + SYS(2015) + ".jpg"
IF !loFtp.UploadFile("c:\temp\Sail-Big.jpg", lcUploadFile)
	? loFtp.cErrorMsg
	RETURN
ENDIF
? "Uploaded " + lcuploadFile


*** provide a folder name (no wildcards)
loCol = loFtp.ListFiles("/SubFolder")
IF ISNULL(loCol)
   ? loFtp.cErrorMsg
   RETURN
ENDIF

? TRANSFORM(loCol.Count ) + " matching file(s)"
? loFtp.cErrorMsg
FOR EACH loFile IN loCol FOXOBJECT
   ? loFile.Name
   
   IF ( AT("Sail-Big-UPLOADED",loFile.Name) = 1)
	   *? loFtp.oBridge.ToJson(loFile)  && for kicks print out as json
	   IF loFtp.DeleteFile(loFile.FullName)
	      ? "*** Deleted " + loFile.FullName
	   ENDIF
	   
   ENDIF
ENDFOR

loFtp.Close()
RETURN


DEFINE class SFtpClientProgressEvents as Custom

*** You can set this to .T. to abort an upload or download
lCancelDownload = .F.

FUNCTION OnFtpBufferUpdate(lnPercent, lnDownloadedBytes, lcRemotePath, lcMode)
  lcMsg = lcMode + ": " + TRANSFORM(lnPercent) + "% complete. " + lcRemotePath + " - " + TRANSFORM(lnDownloadedBytes) + " bytes"
  ? "*** " + lcMsg
ENDFUNC

ENDDEFINE 
#ENDIF 


#INCLUDE ..\INCLUDE\WCONNECT.H

*************************************************************
DEFINE CLASS wwSFtpClient AS Custom
*************************************************************
*: Author: Rick Strahl
*:         (c) West Wind Technologies, 2011
*:Contact: http://www.west-wind.com
*:Created: 08/30/2011
*************************************************************
#IF .F.
*:Help Documentation
*:Topic:
Class wwFtpClient

*:Description:
Ftp class that uses the .NET Fluent FTP library to provide
FTP and FTPS (FTP over TLS). This class **does not support SFTPS**.
*:ENDHELP
#ENDIF

*** Domain name or IP address for the server
cserver = ""

*** The port to run the FTP connection on
nPort = 21


*** Tsl used for this request (FTPS protocol: FTP over TLS)
lUseTls = .F.

*** Allows you to ignore certificate errors **NOT RECOMMENDED: Fix your Certs!**
lIgnoreCertificateErrors = .F.

*** Username for the connection
cUsername = ""

*** Password for the connection
cPassword = ""

*** Optional Private key file for sign on
cPrivateKeyFile = ""

*** Optional Private key file password for sign on
cPrivateKeyFilePassword =""

*** Connection timeout in seconds
nTimeout = 10

*** The FTP server to connect to
cftpserver = ""

*** The Source file to upload from the client or download from the server.
cftpsource = ""

*** The file to download to from the server or upload to on the server.
cftptarget = ""

*** Flag that sets whether to use passive FTP.
lpassiveftp = .T.

lEnableProgressEvents = .F.

*** Logging is not supported for SFTP - property provided only for 
*** compatibility so code can be used interchangeably
cLogFile = ""

cHttpProxyName = ""
cHttpProxyPort = 8888
cHttpProxyPassword = ""
cHttpProxyUsername = ""
cHttpProxyByPass = ""

*** Cancel flag that can be set in OnFTPBufferUpdate to let wwFTP know to stop downloading a file from the Internet.
lcanceldownload = .F.

*** Error message set after a method call that fails
cErrorMsg = ""

*** Optional Log file name. If not empty logging occurs otherwise no logging
*** Note: SFTP has no internal logging support, so logging simply tracks connects
***       and individual operations.
cLogFile = ""

oProgressEventObject = null 

*** wwDotnetBridge instance
oBridge = null

*** .NET wwFtpClient instance
oFtpClient = null


nftpworkbuffersize = 24576

************************************************************************
*  Init
****************************************
***  Function:
***    Assume:
***      Pass:
***    Return:
************************************************************************
PROCEDURE Init()

#IF SHAREWARE
WAIT WINDOW NOWAIT ;
  "Welcome to West Wind Client Tools..." + CHR(13) + ;
  "This is a shareware copy of wwFtp." + CHR(13)+ CHR(13) +;
  "Please register your copy."
#ENDIF


IF ISNULL(this.oBridge)
   this.oBridge = GetwwDotnetBridge()
ENDIF

THIS.oFTPClient = this.oBridge.CreateInstance("Westwind.WebConnection.wwSftpFtpClient")

IF ISNULL(THIS.oFTPClient )
   ERROR "Unable to load SFTP Client. Make sure Renci.SshNet.dll is available." + ;
         CRLF + this.oBridge.cErrorMsg
ENDIF

ENDFUNC



************************************************************************
*  Destroy
****************************************
***  Function: Closes down connection if still open
***    Assume:
***      Pass:
***    Return:
************************************************************************
PROCEDURE Destroy
THIS.Close()
ENDFUNC




************************************************************************
*  OnFtpBufferUpdate
****************************************
***  Function: Event that fires whenever the buffer is updated by
***            downloadfile or uploadfile.
***      Pass:
***    Return: lcMode: upload download
************************************************************************
FUNCTION OnFtpBufferUpdate(lnbytesdownloaded,lnbufferreads,lccurrentchunk, lnTotalBytes, loFtp)

ENDFUNC


************************************************************************
*  Connect
****************************************
***  Function:
***    Assume:
***      Pass:
***    Return:
************************************************************************
FUNCTION Connect(lcServer, lcUsername, lcPassword)
LOCAL loSftpClient, loSftp

IF EMPTY(lcServer)
   lcServer = this.cServer
ENDIF

IF ISNULLOREMPTY(lcUserName)
   lcUsername = this.cUsername
ENDIF
IF ISNULLOREMPTY(lcPassword)
   lcPassword = this.cPassword
ENDIF


IF ISNULLOREMPTY(lcUserName)
   lcUsername = ""
ENDIF
IF ISNULLOREMPTY(lcPassword)
   lcPassword = ""
ENDIF   

loSftp = this.oFtpClient
loSftp.Timeout = this.nTimeout

*** Messing with this buffer causes file corruption
loSftp.BufferSize = this.nFtpWorkBuffersize

loSftp.ProxyName = this.cHttpProxyName
loSftp.ProxyUsername = this.cHttpProxyUsername
loSftp.ProxyPassword = this.cHttpProxyPassword
loSftp.LogFile = GetFullPath(this.cLogFile)

IF !ISNULLOREMPTY(this.cPrivateKeyFile)
   loSftp.PrivateKeyFile = this.cPrivateKeyFile
   loSftp.PrivateKeyFilePassword = this.cPrivateKeyFilePassword
ENDIF

loSftp.FoxProwwFtpClientForProgress = THIS.oProgressEventObject

loSftpClient = loSftp.Connect(lcServer, INT(this.nPort),lcUsername,lcPassword)
IF ISNULL(loSftpClient)
   THIS.cErrorMsg = "Unable to connect to SFTP server: " + this.oFtpClient.ErrorMessage
   RETURN  .F.
ENDIF

RETURN .T.
ENDFUNC


************************************************************************
*  FtpClose
****************************************
***  Function: Closes an open Ftp Connection
***    Assume:
***      Pass:
***    Return:
************************************************************************
FUNCTION Close()

IF !ISNULL(this.oFtpClient)
    this.oFtpClient.Close()        
    THIS.oFtpClient = null
ENDIF
ENDFUNC



************************************************************************
*  DownloadFile
****************************************
***  Function: Copies a file from the server to the client
***    Assume: Connection is open already with FtpOpen
***      Pass: lcSourceFile - Server relative path to a file to download
***            lcTargetFile - Full path to a local file to download to
***    Return: 0 on success - error number of failure
************************************************************************
FUNCTION DownloadFile(lcSourceFile, lcTargetFile)
LOCAL llResult 

llResult = this.oFtpClient.DownloadFile(lcSourceFile,FULLPATH(lcTargetFile))
IF (!llResult)
   this.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN .F.
ENDIF   

RETURN .T.
ENDFUNC

************************************************************************
*  UploadFile
****************************************
***  Function: Sends a file to the server 
***    Assume: Connection to the server is already open
***      Pass: lcSourceFile  - full path to file to upload
***            lcTargetFile  - Server relative path to file on server
***    Return: 0 or error code
************************************************************************
FUNCTION UploadFile(lcSourceFile, lcTargetFile)
LOCAL lcFile, llResult

lcFile = LOWER(FULLPATH(lcSourceFile))
IF !FILE(lcFile)
   this.cErrorMsg = "Invalid file passed for uploading."
   RETURN .F.
ENDIF
   
llResult = this.oFtpClient.UploadFile(lcFile,lcTargetFile)
IF (!llResult)
   this.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN .F.
ENDIF   

ENDFUNC



************************************************************************
*  FtpDeleteFile
****************************************
***  Function: Deletes a file from the server
***    Assume: Connection must be open
***      Pass: lcFile - server relative file to delete
***    Return:
************************************************************************
FUNCTION DeleteFile(lcFile)

IF EMPTY(lcFile)
   this.cErrorMsg = "No file specified to delete on server."
   RETURN .F.
ENDIF  

IF !this.oFtpClient.DeleteFile(lcFile)
   THIS.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN .F.
ENDIF

RETURN .T.
ENDFUNC


************************************************************************
*  Exists
****************************************
***  Function: Checks to see if a file exists on the server
***    Assume:
***      Pass:
***    Return:
************************************************************************
FUNCTION Exists(lcFile)

IF !this.oFtpClient.Exists(lcFile)
   THIS.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN .F.
ENDIF

RETURN .T.
ENDFUNC
*   Exists

************************************************************************
*  FtpRenameFile
****************************************
***  Function: Renames a file on the FTP Server
***    Assume: 
***      Pass:
***    Return:
************************************************************************
FUNCTION RenameFile(lcOldName, lcNewName)

IF !this.oFtpClient.RenameFile(lcOldName, lcNewName)
   THIS.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN .F.
ENDIF

RETURN .T.
ENDFUNC



************************************************************************
*  ListFiles
****************************************
***  Function: Retrieves an FTP directory file listing into a Collection
***    Assume: connection must be open with FtpOpen() first
***      Pass: lcFileSpec - file spec (can include wildcards)
***    Return: Collection or null Collection may be empty 
***            if no files are found
************************************************************************
FUNCTION ListFiles(lcFolderName)
LOCAL loCol, loFile, loFileItems, lnCount, lnX, loFileList

loFileItems = this.oBridge.InvokeMethod(this.oFtpClient,"ListFiles",lcFolderName)

IF ISNULL(loFileItems)
   this.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN NULL
ENDIF   

*** Convert to a FoxPro collection
loFileList = CREATEOBJECT("Collection")
lnCount = loFileItems.Count
FOR lnX = 0 TO lnCount - 1
   loFile = loFileItems.Item(lnX)
   IF loFile.Name == "." OR loFile.Name = ".."
      LOOP
   ENDIF
   loFileList.Add(loFileItems.Item(lnX))
ENDFOR
 
RETURN loFileList
ENDFUNC
* EOF ListFiles()

************************************************************************
*  CreateDirectory
****************************************
***  Function: Creates a new Server directory
***    Assume:
***      Pass:
***    Return:
************************************************************************
FUNCTION CreateDirectory(lcPath)

IF EMPTY(lcPath) 
   RETURN .F.
ENDIF

if !this.oFtpClient.CreateDirectory(lcPath)
   THIS.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN .F.
ENDIF

RETURN .T.
ENDFUNC

************************************************************************
*  DeleteDirectory
****************************************
***  Function: Removes a path from the server. Path is server relative
***    Assume: Connection must be open with FtpOpen
***      Pass: lcPath - Server relative path to remove
***    Return: .T. or .F.
************************************************************************
FUNCTION DeleteDirectory(lcPath)

IF !this.oFtpClient.RemoveDirectory(lcPath)
   THIS.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN .F.
ENDIF

RETURN .T.
ENDFUNC

************************************************************************
*  ChangeDirectory
****************************************
***  Function: Changes the active directory on the FTP Serve3r
***    Assume:
***      Pass: Remote path
***    Return:
************************************************************************
FUNCTION ChangeDirectory(lcPath)

IF !this.oFtpClient.ChangeDirectory(lcPath)
   THIS.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN .F.
ENDIF

RETURN .T.
ENDFUNC
*   ChangeDirectory



************************************************************************
*  FtpCommand
****************************************
***  Function: Issue an arbitrary FTP Command 
***    Assume: Connection must exist with FtpOpen
***      Pass:
***    Return:
************************************************************************
FUNCTION ExecuteCommand(lcCommand)

IF !this.oFtpClient.ExecuteCommand(lcCommand)
   THIS.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN .F.
ENDIF

ENDFUNC


************************************************************************
*  ExecuteDownloadCommand
****************************************
***  Function: Issue an arbitrary FTP Command 
***    Assume: Connection must exist with FtpOpen
***      Pass:
***    Return:
************************************************************************
FUNCTION ExecuteDownloadCommand(lcCommand)

IF !this.oFtpClient.ExecuteDownloadCommand(lcCommand)
   THIS.cErrorMsg = this.oFtpClient.ErrorMessage
   RETURN .F.
ENDIF

ENDFUNC


ENDDEFINE
*
*** EndDefine: wwftp
**************************************************
