*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Ntzliche Routinen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

#INCLUDE ..\Include\stInclude.h

*############################################################################################################################
* Allgemeine Funktionen
*############################################################################################################################

*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Memento zum Selectieren eines neuen Aliases und gleichzeitges Merken des aktuellen Aliases.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 27.02.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

DEFINE CLASS StSelect AS CUSTOM

	cAlias = ''
	cOrder = ''
	nRecno = .NULL.


	PROCEDURE INIT
	**************

		LPARAMETERS tcNewAlias AS STRING, tlNoSaveRecno AS BOOLEAN

		*-- Den aktuellen Alias in den Eigenschaften merken
		IF !EMPTY( ALIAS() )

			*-- Aktuellen Alias merken
			THIS.cAlias = ALIAS()

			*-- Wenn Sortierung gesetzt, ebenfalls merken
			IF !EMPTY( ORDER() )
				THIS.cOrder = ORDER()
			ENDIF

			*-- Wenn wir mindestens einen Datensatz haben, wir uns den
			*-- Satz merken wollen und wir nicht bei einer Neuanlage sind.
			IF RECNO() >= 1 .AND. !tlNoSaveRecno

				*-- Recno merken
				THIS.nRecno = RECNO()
			ENDIF
		ENDIF

		*-- Es soll ein neuer Alias gesetzt werden
		IF !EMPTY( tcNewAlias )

			*-- neuer Alias bergeben, also auch selectieren
			SELECT ( tcNewAlias )
		ENDIF

		*-- Alten Alias zurckgeben
		RETURN( THIS.cAlias )

	ENDPROC


	PROCEDURE DESTROY
	*****************

		*-- Den alten Alias wiederherstellen falls einer eingetragen
		*-- ist und auch noch geffnet ist
		IF !EMPTY( THIS.cAlias ) .AND. USED( THIS.cAlias )

			*-- Wenn wir einen haben, dann stellen wir Ihn wieder her
			SELECT ( THIS.cAlias )

			*-- Wenn Sortierung vorhanden =>
			IF !EMPTY( THIS.cOrder )

				*-- Exception abfangen
				TRY

					*-- wieder herstellen
					SET ORDER TO ( THIS.cOrder )
				CATCH
				ENDTRY

			*-- Oder wieder zurcksetzen, denn es knnte
			*-- ja eine Sortierung gesetzt worden sein
			ELSE
				*-- Exception abfangen
				TRY
					SET ORDER TO
				CATCH
				ENDTRY
			ENDIF

			*-- Wenn der Satz berhaupt gespeichert wurde
			IF !ISNULL( THIS.nRecno )

				*-- Wenn die Satznummer grer oder gleich 1 und kleiner gleich RECCOUNT
				IF BETWEEN( THIS.nRecno, 1, RECCOUNT() ) .AND. THIS.nRecno != RECNO()
					GO THIS.nRecno
				ENDIF
			ENDIF
		ENDIF

	ENDPROC

ENDDEFINE


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Kopiert den Inhalt eines Arrays in eine CSV-Liste.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 15.01.16
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.: Das Array MUSS per Referenz bergeben werden.
*				Zahlen sind auch erlaubt.
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
* 26.01.16 HPG: Wir mssen das bergebene Array als EXTERNAL ARRAY dekalieren, sonst meckert der VFP-Compiler dass er taArray
*				nicht finden kann.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION ArrayToCSV
*******************

	LPARAMETERS taArray, tnColumn AS INTEGER, tcSeparator AS STRING

	LOCAL lcCSV AS STRING, lcSeparator AS STRING
	LOCAL lnRow AS INTEGER, lnRowCount AS INTEGER, lnColumn AS INTEGER
	EXTERNAL ARRAY taArray

	*-- Parameter prfen
	IF VARTYPE( tnColumn ) == 'N'
		lnColumn = tnColumn
	ELSE
		lnColumn = 1
	ENDIF
	IF VARTYPE( tcSeparator ) == 'C'
		lcSeparator = tcSeparator
	ELSE
		lcSeparator = ','
	ENDIF

	*-- Parameter setzen
	lcCSV = ''
	lnRowCount = ALEN( taArray, 1 )

	*-- Schleife ber das Array
	FOR lnRow = 1 TO lnRowCount

		*-- Aktuelles Element an CSV-Liste anhngen
		lcCSV = lcCSV + ALLTRIM( TRANSFORM( taArray( lnRow, lnColumn ) ) ) + lcSeparator
	ENDFOR

	*-- Angehngte Trennzeichen lschen
	lcCSV = StrDelR( lcCSV, LEN( lcSeparator ) )

	*-- CSV-Liste zurckgeben
	RETURN( lcCSV )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Aus einer 4-Byte Zeichenkette ein DWORD bestimmen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Buf2DWord
******************

	LPARAMETERS tcBuffer AS STRING

	RETURN( ASC( SUBSTR( tcBuffer, 1, 1 ) ) + BITLSHIFT( ASC( SUBSTR( tcBuffer, 2, 1 ) ), 8 ) + ;
		BITLSHIFT( ASC( SUBSTR( tcBuffer, 3, 1 ) ), 16 ) + BITLSHIFT( ASC( SUBSTR( tcBuffer, 4, 1 ) ), 24 ) )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Erstellt eine Vergleichsbedingung als Zeichenkette.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 20.10.2017
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

PROCEDURE BuildEquation
***********************

	*-- Parameter:
	*-- tuValue:		Umzuwandelnder Wert mit beliebigem Datentyp
	*-- tcFieldName:	Feldename fr den Vergleich
	*-- tcOperator:		Operator fr den Vergleich, Default ist '=='
	LPARAMETERS tuValue AS VARIANT, tcFieldName AS STRING, tcOperator AS STRING

	LOCAL luValue AS VARIANT
	LOCAL lcEquation AS STRING, lcValue AS STRING, lcDataType AS STRING, lcOperator AS STRING

	*-- Parameter-Prfung und damit .NULL.-Werte
	*-- und den fehlenden Feldnamen abfangen
	IF .NOT. ISNULL( tuValue ) .AND. VARTYPE( tcFieldName ) == 'C'

		*-- Parameter prfen und setzen
		lcOperator = EVL( tcOperator, '==' )
		luValue = tuValue

		*-- Datentyp bestimmen und Vergleichsbedingung aufbauen
		lcDataType = VARTYPE( luValue )
		lcEquation = ' ' + tcFieldName + ' ' + lcOperator + ' '

		*-- Datentyp prfen => wir haben ein Text- oder Memofeld
		IF lcDataType == 'C'

			*-- Suchausdruck aufbauen
			lcEquation = lcEquation + "'" + luValue + "'"

		*-- Andernfalls ...
		ELSE

			*-- ... mssen wir den Wert zuerst in eine Zeichenkette umwandeln
			lcValue = TRIM( CAST( luValue AS CHARACTER( 200 ) ) )

			*-- Je nach Datentyp verzweigen
			*-- D, L, N, Q, T, Y
			DO CASE

				*-- Wir haben ein Datums- oder ein Zeitfeld
				CASE lcDataType $ 'DT'

					*-- Suchausdruck aufbauen
					lcEquation = lcEquation + '{' + lcValue + '}'

				*-- Wir haben ein logisches Feld
				CASE lcDataType == 'L'

					*-- Suchausdruck aufbauen, wobei hier der Operator immer auf '==' gesetzt wird
					lcEquation = tcFieldName + "=='" + lcValue + "'"

				*-- Wir haben ein numerisches oder ein Whrungsfeld
				CASE lcDataType $ 'NY'

					*-- Suchausdruck aufbauen
					lcEquation = lcEquation + lcValue
			ENDCASE
		ENDIF
	ENDIF

	RETURN( lcEquation )

ENDPROC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Erstellt die Wert-Zuweisung fr einen SQL-SELECT.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 24.10.2017
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

PROCEDURE BuildSQLSet
*********************

	*-- Parameter:
	*-- tuValue:		Wert der zu setzen ist
	*-- tcFieldName:	Feldename dessen Wert zu setzen ist
	LPARAMETERS tuValue AS VARIANT, tcFieldName AS STRING

	LOCAL luValue AS VARIANT
	LOCAL lcSetValueString AS STRING, lcValue AS STRING, lcDataType AS STRING, lcOperator AS STRING

	*-- Parameter-Prfung und damit .NULL.-Werte
	*-- und den fehlenden Feldnamen abfangen
	IF .NOT. ISNULL( tuValue ) .AND. VARTYPE( tcFieldName ) == 'C'

		*-- Parameter setzen
		luValue = tuValue

		*-- Datentyp bestimmen und Wert-Zuweisung aufbauen
		lcDataType = VARTYPE( luValue )
		lcSetValueString = ' ' + tcFieldName + ' = '

		*-- Datentyp prfen => wir haben ein Text- oder Memofeld
		IF lcDataType $ 'CL'

			*-- Wert-Zuweisung aufbauen
			lcSetValueString = lcSetValueString + "'" + luValue + "'"

		*-- Andernfalls ...
		ELSE

			*-- ... mssen wir den Wert zuerst in eine Zeichenkette umwandeln
			lcValue = TRIM( CAST( luValue AS CHARACTER( 200 ) ) )

			*-- Je nach Datentyp verzweigen
			*-- D, L, N, Q, T, Y
			DO CASE

				*-- Wir haben ein Datumsfeld
				CASE lcDataType == 'D'

					*-- Wert-Zuweisung aufbauen
					lcSetValueString = lcSetValueString + "'" + lcValue + "'"

				*-- Wir haben ein Zeitfeld
				CASE lcDataType == 'T'

					*-- Wert-Zuweisung aufbauen
					lcSetValueString = lcSetValueString + ConvertDateTime2Str( tuValue )

				*-- Wir haben ein numerisches oder ein Whrungsfeld
				CASE lcDataType $ 'NY'

					*-- Wert-Zuweisung aufbauen
					lcSetValueString = lcSetValueString + lcValue
			ENDCASE
		ENDIF
	ENDIF

	RETURN( lcSetValueString )

ENDPROC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Funktion zum Umwandeln eines binr codierten Zahlenwertes aus einem String-Puffer in einen Integer-Wert.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION ConvertcBuffer2Int
***************************

	LPARAMETERS tcBuffer AS STRING

	LOCAL lnInteger AS INTEGER, lnLength AS INTEGER

	*-- Pufferlnge bestimmen
	lnLength = LEN( tcBuffer )
	lnInteger = 0

	*-- Je nach Lnge des Puffers verzweigen
	DO CASE

		*-- Wenn der Puffer 2 Bytes lang ist
		CASE lnLength = 2

			*-- Integerwert aus String-Puffer bestimmen
			lnInteger = ASC( SUBSTR( tcBuffer, 1, 1 ) ) + ;
				BITLSHIFT( ASC( SUBSTR( tcBuffer, 2, 1 ) ),  8 )

		*-- Wenn der Puffer 4 Bytes lang ist
		CASE lnLength = 4

			*-- Integerwert aus String-Puffer bestimmen
			lnInteger = ASC( SUBSTR( tcBuffer, 1, 1 ) ) + ;
				BITLSHIFT( ASC( SUBSTR( tcBuffer, 2, 1 ) ),  8 ) + ;
				BITLSHIFT( ASC( SUBSTR( tcBuffer, 3, 1 ) ), 16 ) + ;
				BITLSHIFT( ASC( SUBSTR( tcBuffer, 4, 1 ) ), 24 )

		*-- Wenn der Puffer 8 Bytes lang ist
		CASE lnLength = 8

			*-- Integerwert aus String-Puffer bestimmen
			lnInteger = ( ( ( ( ( ( ( ASC( SUBSTR( tcBuffer, 8, 1 ) ) ) * 256 + ;
				ASC( SUBSTR( tcBuffer, 7, 1 ) ) ) * 256 + ;
				ASC( SUBSTR( tcBuffer, 6, 1 ) ) ) * 256 + ;
				ASC( SUBSTR( tcBuffer, 5, 1 ) ) ) * 256 + ;
				ASC( SUBSTR( tcBuffer, 4, 1 ) ) ) * 256 + ;
				ASC( SUBSTR( tcBuffer, 3, 1 ) ) ) * 256 + ;
				ASC( SUBSTR( tcBuffer, 2, 1 ) ) ) * 256 + ;
				ASC( SUBSTR( tcBuffer, 1, 1 ) )
	ENDCASE

	RETURN( lnInteger )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Funktion zum Umwandeln eines DatumZeit-Ausdruckes in eine eindeutige DatumZeit-Zeichenkette (die auch fr
*               SQL-Server geeignet ist).
* Autor.......: Hans-Peter Grzinger
* Datum.......: 21.02.2005
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

PROCEDURE ConvertDateTime2Str
*****************************

	LPARAMETERS ttDateTime AS DATETIME

	LOCAL lcDateTime AS STRING

	*-- DatumZeit-Ausdruck in Zeichenkette umwandeln
	lcDateTime = '{^' + LEFT( DTOS( ttDateTime ), 4 ) + '-' + SUBSTR( DTOS( ttDateTime ), 5, 2 ) + '-' + ;
		RIGHT( DTOS( ttDateTime ), 2 ) + ' ' + TTOC( ttDateTime, 2 ) + '}'

	RETURN( lcDateTime )

ENDPROC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Wandelt Daten verschiedenen Datentyps in eine Zeichenkette.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 29.09.14
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

PROCEDURE Convert2Str
*********************

	LPARAMETERS tuValue AS VARIANT

	LOCAL lcDataType AS STRING, lcValue AS STRING

	*-- Datentyp bestimmen
	lcDataType = VARTYPE( tuValue, .T. )

	*-- Je nach Datentyp verzweigen
	DO CASE

		*-- Datentyp Zeichen
		CASE lcDataType = 'C'
			lcValue = '"' + STRTRAN( tuValue, '"', ["+'"'+"] ) + '"'

		*-- Datentyp numerisch
		CASE lcDataType = 'N'
			lcValue = TRANSFORM( tuValue )

		*-- Datentyp logisch
		CASE lcDataType = 'L'
			lcValue = IIF( tuValue, '.T.', '.F.' )

		*-- Datentyp Datum/DatumZeit
		CASE lcDataType $ 'DT'
			lcValue = ' {^' + YEAR( tuValue ) + '-' + MONTH( tuValue ) + '-' + DAY( tuValue )

			*-- Beim Typ DatumZeit ...
			IF lcDataType = 'T'

				*-- ... noch den Zeitanteil anhngen
				lcValue = lcValue + ',' + HOUR( tuValue ) + ':' + MINUTE( tuValue ) + ':' + SEC( tuValue )
			ENDIF
			lcValue = lcValue + "}"

		*-- Datentyp NULL/Objekt
		CASE lcDataType $ 'XO'
			lcValue = 'NULL'
	ENDCASE

	RETURN( lcValue )

ENDPROC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Kopiert ausgewhlte Spalten der aktuellen Tabelle/Cursor/View in die Zwischenablage.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.03.2004
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.: Ursprungsfunktion Copy2Clip kommt aus der Projektverfolgung und wurde von HPG entwickelt. Die aktuelle
*				Funktion wurde aus der Adressverwaltung bernommen und nur leicht modifiziert fr das ST-Frame.
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 22.02.09 HPG: Anstatt des Arrays mit der Feldliste bergeben wir jetzt eine Komma separierte Liste der zu kopierenden
*				Felder
* 24.09.12 HPG: Wir verwenden bei Vorhandensein einer Feld-berschrift diese anstatt den Feldnamen.
* 08.12.12 HPG: Neues Parameter-Verhalten: Wenn 'NO_HEADER' im Parameter tcFieldList bergeben wird, lassen wir die ber-
*				schrift weg.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
* 03.12.06 HPG: Alias immer auf den Master-BO setzen ...
* 18.12.06 HPG: ... aber nur wenn ein BO vorhanden ist.
* 07.01.07 HPG: Keine Tabelle im aktuellen Arbeitsbereich geffnet => Fehlermeldung und wieder zurck.
*----------------------------------------------------------------------------------------------------------------------------

PROCEDURE Copy2Clip
*******************

	*#DEFINE DEBUG_C2C

	LPARAMETERS toForm AS OBJECT, tlDontShowInfo AS BOOLEAN, tcFieldList AS STRING

	LOCAL lcCommand AS STRING, lcFieldName AS STRING, lcFieldType AS STRING, lcField AS STRING, lcDec AS STRING
	LOCAL lcBuffer AS STRING, lcOldSet AS STRING, lcAlias AS STRING, lcFieldList AS STRING, lcCaption AS STRING
	LOCAL llCopyFields AS BOOLEAN, llNoHeader AS BOOLEAN
	LOCAL lnFieldCount AS INTEGER, lnOldDataSession AS INTEGER, lnField AS INTEGER
	LOCAL loOldAlias AS 'stSelect'

	*-- Richtige DataSession setzen, zuvor die aktuelle merken
	SET DATASESSION TO toForm.DATASESSIONID

	*-- Alias immer auf den Master-BO setzen, falls ein BO vorhanden ist
	IF TYPE( 'toForm.oMasterBusinessObject' ) == 'O'

		*-- Aktuellen Alias in einem Mementoobjekt merken, dabei
		loOldAlias = CREATEOBJECT( 'stSelect', toForm.oMasterBusinessObject.GetAlias() )
	ENDIF

	*-- Wenn keine Tabelle im aktuellen Arbeitsbereich geffnet ist ...
	IF !USED()

		*-- ... Fehlermeldung und wieder zurck.
		MESSAGEBOX( 'Keine Tabelle geffnet!', 'Schlieen Sie das Formular und ffnen es nochmals.' + CR + ;
			'Kopieren Sie dann die Tabelle erneut.', 64+0+0 )
		RETURN( .F. )
	ENDIF

	*-- Wenn Feldliste der zu kopierenden Felder bergeben wurde
	IF VARTYPE( tcFieldList ) == 'C'

		*-- Nur bestimmte Felder kopieren
		llCopyFields = .T.

		*-- Die Feldliste "uppern" da in der Funktion AFIELDS
		*-- die Feldnamen komplett gro geschrieben werden.
		lcFieldList = UPPER( tcFieldList )

		*-- Prfen auf spezielle Einstellung
		IF lcFieldList = 'NO_HEADER'

			*-- wenn ja, alle Felder kopieren aber ohne Header
			llCopyFields = .F.
			llNoHeader = .T.
		ENDIF

	*-- Andernfalls ...
	ELSE

		*-- ... alle Felder kopieren
		llCopyFields = .F.
	ENDIF

	*-- 10.03.2004 Hans-Peter Grzinger [START]
	*-- damit es mal funktioniert
	*-- Alle Felder kopieren
	*llCopyFields = .F.
	lnStartField = 1

	*------------------------------------------------------------------
	* Nach der Vorbereitung kopieren wir den Inhalt des aktuellen
	* Alias bzw. Cursor ins Clipboard
	*------------------------------------------------------------------

	*-- Mauszeiger auf Sanduhr
	_SCREEN.MOUSEPOINTER = 11

	*-- Array mit Datenfeldinformation fllen und Anzahl der Felder bestimmen
	AFIELDS( laFieldArray )
	lnFieldCount = FCOUNT() - 1

	*-- Noch ein paar Parameter initialisieren
	*-- Variable fr den Kopf, Puffer und erster Teil der Befehlszeichenkette
	lcHeader = ''
	lcBuffer = ''

#IFDEF DEBUG_C2C
	lcCommand = 'lcBuffer1 = lcBuffer1'
#ELSE
	lcCommand = 'lcBuffer = lcBuffer'
#ENDIF

	*-- Schleife ber alle Felder
	FOR lnField = lnStartField TO lnFieldCount

		*-- Feldname und Feldtyp bestimmen
		lcFieldName = laFieldArray[lnField, 1]
		lcFieldType = laFieldArray[lnField, 2]

		*-- berschrift holen
		*-- Kurzen Tabellen-Namen holen
		lcTableName = LEFT( ALIAS(), 3 )

		*-- Falls etwas in die Hose geht, zb. bei einer freien Tabellen
		*-- verwenden wir hier vorsorglich gleich TRY/CATCH
		TRY
			lcCaption = DBGETPROP( lcTableName + '.' + lcFieldName, 'FIELD', 'Caption' )
			IF EMPTY( lcCaption )

				*-- Keine Caption, dann nehmen wir den Feldnamen
				lcCaption = lcFieldName
			ENDIF
		CATCH

			*-- Keine Caption, dann nehmen wir den Feldnamen
			lcCaption = lcFieldName
		ENDTRY

		*-- ToDo: ID-Felder und die vier Datums/User-Felder nicht kopieren.
		IF RIGHT( lcFieldName, 2 ) == '_I'
			LOOP
		ENDIF

		*-- Felder die nicht auszugeben sind
		*IF lcFieldName == 'INTERN'
		IF lcFieldName == 'VERWSTATUS'
			LOOP
		ENDIF

		*-- Nur ausgewhlte Felder kopieren
		IF llCopyFields

			*-- SET Einstellung merken und EXACT auf ON, damit keine Teilstrings gesucht werden
			lcOldSet = SET( 'EXACT' )
			SET EXACT ON

			*-- Soll das aktuelle Feld kopiert werden?
			*-- Wenn dieses Feld nicht in der Feldliste enthalten ist =>
			IF AT( lcFieldName, lcFieldList ) == 0

				*-- Alte SET-Einstellung wieder herstellen
				SET EXACT &lcOldSet

				*-- dann machen wir nix
				LOOP
			ENDIF

			*-- Alte SET-Einstellung wieder herstellen
			SET EXACT &lcOldSet
		ENDIF

		*-- Je nach Datentyp verzweigen
		DO CASE

			*-- Typ Zeichen
			CASE lcFieldType = 'C'

				*-- Leerzeichen abschneiden und TAB anhngen
				lcField = ' + ALLTRIM( ' + lcFieldName + ' ) + CHR( 9 )'

			*-- Typ Memo
			CASE lcFieldType = 'M'

				*-- Memofeld auf 250 Zeichen begrenzen und TAB + CR + LF entfernen
				*-- und 'am Ende' ein TAB anhngen
				lcField = " + LEFT( STRTRAN( STRTRAN( STRTRAN(" + lcFieldName + ", CHR( 9 ), ' ' ), CHR( 13 ), ' ' ), CHR( 10 ), ' ' ), 250 ) + CHR( 9 )"

			*-- Typ Datum
			CASE lcFieldType = 'D'

				*-- In Zeichenkette umwandeln und TAB anhngen
				lcField = ' + DTOC( ' + lcFieldName + ' ) + CHR( 9 )'

			*-- Typ DatumZeit
			CASE lcFieldType = 'T'

				*-- In Zeichenkette umwandeln und TAB anhngen
				lcField = ' + TTOC( ' + lcFieldName + ' ) + CHR( 9 )'

			*-- Typ Numerisch
			CASE lcFieldType = 'N'

				*-- Anzahl der Dezimalstellen
				lcDec = ALLTRIM( STR( laFieldArray[lnField, 4] ) )

				*-- Nummerisches Feld in String mit der angegebenen Anzahl Dezimalstellen
				*-- und Ersetzen eventueller Dezimalpunkte mit Dezimalkomma (fr Wixcel)
				lcField = " + STRTRAN( ALLTRIM( STR( " + lcFieldName + ", 20, " + lcDec + " ) ), '.', ',' ) + CHR( 9 )"

			*-- Typ Integer
			CASE lcFieldType = 'I'

				*-- In Zeichenkette umwandeln und TAB anhngen
				lcField = ' + ALLTRIM( STR( ' + lcFieldName + ' ) ) + CHR( 9 )'

			*-- Typ Logisch
			CASE lcFieldType = 'L'

				*-- In Zeichenkette umwandeln und TAB anhngen
				lcField = " + IIF( " + lcFieldName + ", 'Ja  ', 'Nein' ) + CHR( 9 )"

			*-- Unbekannte Feldtypen (Objekt oder so was) ignorieren wir
			OTHERWISE
		ENDCASE

		*-- Datenkopf aufbauen und Feldbefehl anhngen
		lcHeader = lcHeader + lcCaption + CHR( 9 )
		lcCommand = lcCommand + lcField
	ENDFOR

	*-- CR anhngen
	lcCommand = lcCommand + ' + CHR( 13 )'

	*-- berschriften in Puffer schreiben
	lcBuffer = lcBuffer + IIF( llNoHeader, '', lcHeader + CHR( 13 ) )

	*-- Nicht abbrechen
	llExit = .F.

	*-- ab erstem Datensatz
	GO TOP

	*-- Schleife ber alle Datenstze
	SCAN WHILE !llExit

#IFDEF DEBUG_C2C
		*-- Aktuellen Satz in Puffer kopieren
		lcBuffer1 = ''
		&lcCommand

		*-- Wenn der aktuelle Datensatz ein .NULL. erzeugt
		IF ISNULL( lcBuffer1 )

			*-- Fehlermeldung in Debugdatei schreiben
			STRTOFILE( 'Satz ergibt ein .NULL. lcCommand = ' + lcCommand + CRLF, 'C:\TMP\SATZ.TXT', 1 )
			lcBuffer = lcBuffer + lcBuffer1

		*-- Andernfalls alles OK
		ELSE

			*-- Puffer in Debugdatei schreiben
			STRTOFILE( lcBuffer1, 'C:\TMP\SATZ.TXT', 1 )
			lcBuffer = lcBuffer + lcBuffer1
		ENDIF
#ELSE
		*-- Aktuellen Satz in Puffer kopieren
		&lcCommand
#ENDIF

	ENDSCAN

	*-- Wenn nicht abgebrochen wurde und der Type von lcBuffer 'Zeichen' ist
	IF !llExit .AND. VARTYPE( lcBuffer ) == 'C'

		*-- Komplette Daten in Clipboard kopieren
		_CLIPTEXT = lcBuffer
	ENDIF

	*-- Mauszeiger wieder 'normal'
	_SCREEN.MOUSEPOINTER = 0

	*-- Wenn eine Information anzuzeigen ist
	IF !tlDontShowInfo

		*-- Meldung ausgeben
		WAIT WINDOW 'Daten wurden in die Zwischenablage kopiert!'
	ENDIF

	RETURN( .T. )

ENDPROC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Prft ob der bergebene Verzeichnisname vorhanden ist.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION DirExist
*****************

LPARAMETERS tcDirName AS STRING

LOCAL llExist AS BOOLEAN

*-- Prfen ...
TRY

	*-- Anhngenden Backslash falls vorhanden lschen
	tcDirName = IIF( RIGHT( tcDirName, 1 ) == '\', LEFT( tcDirName, LEN( tcDirName ) - 1 ), tcDirName )

	*-- ... ob Verzeichnis vorhanden ist
	llExist = ADIR( laFiles, tcDirName, 'DHS' ) == 1

	*-- Falls ein ungltiger bzw. unzulssiger Verzeichnisnamen
	*-- zum Prfen bergeben wurde ...
CATCH

	*-- ... geben wir .F. zurck
	llExist = .F.
ENDTRY

RETURN( llExist )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Prft ob der bergebene Dateiname vorhanden ist.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION FileExist
******************

LPARAMETERS tcFileName AS STRING

LOCAL llExist AS BOOLEAN

*-- Prfen ob Datei vorhanden ist
llExist = ADIR( laFiles, tcFileName, 'HS' ) == 1

RETURN( llExist )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Wandelt in einer HTML-Zeichenkette die Sonderzeichen wieder in ANSI-Umlaute/Sonderzeichen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 19.08.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION HTMLString2String( tcString AS STRING )
************************************************

	LOCAL lcString AS STRING

	lcString = STRTRAN( ALLTRIM( tcString ), '&#40;', '(' )		&& (
	lcString = STRTRAN( lcString, '&#41;', ')' )				&& )
	lcString = STRTRAN( lcString, '&#62;', '>' )				&& >
	lcString = STRTRAN( lcString, '&#92;', '\' )				&& \
	lcString = STRTRAN( lcString, '&#34;', '"' )				&& "
	lcString = STRTRAN( lcString, '&#35;', '#' )				&& #
	lcString = STRTRAN( lcString, '&#39;', "'" )				&& '
	lcString = STRTRAN( lcString, '&#61;', '=' )				&& =

RETURN( lcString )


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Prft ob eine Anwendung schon luft.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 27.04.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION IsExeRunning
*********************

	LPARAMETERS tcName AS STRING, tlTerminate AS BOOLEAN

	LOCAL loLocator AS OBJECT, loWMI AS OBJECT, loProcesses AS OBJECT, loProcess, loProcesses AS OBJECT
	LOCAL llIsRunning AS BOOLEAN

	*-- Locater-Objekt erstellen und mit dem Server verbinden
	loLocator 	= CREATEOBJECT( 'WBEMScripting.SWBEMLocator' )
	loWMI		= loLocator.ConnectServer()
	loWMI.Security_.ImpersonationLevel = 3  		&& Impersonate

	*-- Die laufenden Prozesse abfragen ob der gesuchte darunter ist
	loProcesses	= loWMI.ExecQuery([SELECT * FROM Win32_Process WHERE Name = '] + tcName + ['])
	llIsRunning = .F.

	*-- Wurden Prozesse mit den Suchkriterien gefunden
	IF loProcesses.COUNT > 0

		*-- Schleife ber alle Prozesse
		FOR EACH loProcess IN loProcesses

			*-- Kennzeichen setzen
			llIsRunning = .T.

			*-- Wenn der Prozess beendet werden soll ...
			IF tlTerminate

				*-- ... tun wir das
				loProcess.TERMINATE( 0 )
			ENDIF
		ENDFOR
	ENDIF

	RETURN( llIsRunning )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Wandelt eine Zahl in eine Zeichenkette ohne fhrende bzw. angehngte Leerzeichen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Num2C
**************

	LPARAMETERS tnValue AS INTEGER, tnDecimals AS INTEGER

	LOCAL lcValue AS STRING
	LOCAL lnDecimals AS INTEGER

	*-- Parameter prfen
	IF VARTYPE( tnDecimals ) = 'N'
		lnDecimals = tnDecimals
	ELSE
		lnDecimals = 0
	ENDIF

	*-- Umwandeln in Zeichenkette mit gewnschter
	*-- Anzahl der Nachkomma-Stellen
	lcValue = LTRIM( STR( tnValue, 12, lnDecimals ) )

	*-- Rckgabe
	RETURN( lcValue )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: DWORD in eine 4-Byte Zeichenkette schreiben.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Num2DWord
******************

	LPARAMETERS tnValue AS INTEGER

	#DEFINE m0		256
	#DEFINE m1		65536
	#DEFINE m2		16777216

	LOCAL b0 AS STRING, b1 AS STRING, b2 AS STRING, b3 AS STRING

	*-- Wenn der Wert kleiner als 0 ist ...
	IF tnValue < 0

		*-- ... mssen wir das Vorzeichen beachten
		tnValue = 0x100000000 + tnValue
	ENDIF
	b3 = INT( tnValue / m2 )
	b2 = INT( ( tnValue - b3 * m2 ) / m1 )
	b1 = INT( ( tnValue - b3 * m2 - b2 * m1 ) / m0 )
	b0 = MOD( tnValue, m0 )

	RETURN( CHR( b0 ) + CHR( b1 ) + CHR( b2 ) + CHR( b3 ) )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: WORD in eine 2-Byte Zeichenkette schreiben.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Num2Word
*****************

	LPARAMETERS tnValue AS INTEGER

	RETURN( CHR( MOD( tnValue, 256 ) ) + CHR( INT( tnValue / 256 ) ) )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: ffnet ein Dokument mit der Standard-Anwendung.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 24.02.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION OpenDocument
*********************

	LPARAMETERS tcDocument AS STRING

	LOCAL llReturn AS BOOLEAN

	*-- Zuerst prfen, ob das Dokument berhaupt vorhanden ist
	IF FileExist( tcDocument )

		*-- Zuerst die Parameter setzen
		lcOperation = 'Open'
		lcFileName = tcDocument
		lcParameter = ''
		lcWorkDir = ''

		*-- Dokument mit der Standard-Anwendung ffnen.
		ShellExecute( 0, lcOperation, lcFileName, lcParameter, lcWorkDir, 0 )

		*-- Flag auf alles OK setzen
		llReturn = .T.

	*-- Andernfalls
	ELSE

		*-- Fehlermeldung ausgeben
		MESSAGEBOX( 'Dokument >' + ALLTRIM( tcDocument ) + '< nicht vorhanden!' + CHR( 13 ) + CHR( 13 ) + ;
			'Kein ffnen mglich.', 48+0+0, 'Fehler' )

		*-- Flag auf Fehler setzen
		llReturn = .F.
	ENDIF

	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Artikel-Liste aus der INI-Daten WebView_LizenzService.ini lesen und in den Cursor cursWebViewArticle
*				schreiben.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 14.10.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 22.10.24 HPG: Format gendert, statt einem Tabulator soll der ; die Spalten trennen:
*				Artikelnummer;Verffentlichungsjahr vierstellig;Laufzeit in Tagen;Beschreibung
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION ReadWebViewArticleList()
*********************************

	LOCAL lcLicenseFile AS STRING
	LOCAL llReturn AS BOOLEAN

	*-- Parameter setzen
	lcLicenseFile = FULLPATH( 'WebView_LizenzService.ini' )

	*-- Ist die Datei WebView_LizenzService.ini vorhanden
	IF FileExist( lcLicenseFile )

		*-- Tabelle schlieen
		IF USED( 'cursWebViewArticle' )
			USE IN cursWebViewArticle
		ENDIF

		*-- Tabelle neu erstellen und die Artikel-Nummern
		*-- aus der Datei WebView_LizenzService.ini lesen
		CREATE CURSOR cursWebViewArticle ( ArtNr C (4), Jahr I (4), Tage I (4), Info C(200) )
		APPEND FROM ( lcLicenseFile ) TYPE DELIMITED WITH CHARACTER ';'

		*-- Index erstellen
		SELECT cursWebViewArticle
		INDEX ON ArtNr TAG ArtNr
		SET ORDER TO ArtNr

		*-- Alles OK
		llReturn = .T.

		*-- 14.10.24 HPG: Zum Debuggen
		*COPY TO C:\TMP\cursWebViewArticle

	*-- Andernfalls ...
	ELSE

		*-- ... den Anwender informieren
		MESSAGEBOX( 'Die Datei >' + lcLicenseFile + '< mit der Artikle-Liste ist nicht vorhanden!', 16+0+0, 'Fehler' )

		*-- Alles OK
		llReturn = .F.
	ENDIF

	*-- Erfolg/Miserfolg zurckgeben
	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Funktion zum Lesen der INI - Datei
* Autor.......: Hans-Peter Grzinger
* Datum.......: 02.12.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION ReadINI
*****************

	LPARAMETERS tcSection AS STRING, tcKey AS STRING, tnLength AS STRING, tcDefault AS STRING, tcINIFile AS STRING

	LOCAL lcReturn AS STRING, lcINIFile AS STRING

	*-- Parameter prfen
	IF VARTYPE( tcINIFile ) == 'C'
		lcINIFile = tcINIFile
	ELSE
		lcINIFile = gcINIFileName
	ENDIF

	*-- Parameter vorbesetzten
	lcReturn = SPACE( tnLength ) + CHR( 0 )

	*-- DECLARE DLL-Anweisungen zum Lesen/Schreiben von privaten INI-Dateien
	DECLARE INTEGER GetPrivateProfileString IN Win32API AS GetPrivStr STRING cSection, STRING cKey, STRING cDefault, ;
		STRING @cBuffer, INTEGER nBufferSize, STRING cINIFile

	*-- INI-Datei auslesen
	IF GetPrivStr( tcSection, tcKey, '', @lcReturn, tnLength, FULLPATH( lcINIFile ) ) = 0
		*MESSAGEBOX( 'Fehler beim Lesen der INI-Datei!', 16+0+0, 'Fehler' )
		lcReturn = tcDefault
	ELSE
		lcReturn = LEFT( lcReturn, AT( CHR( 0 ), lcReturn ) - 1 )
	ENDIF

	RETURN( lcReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Das externe Programm ReadWebViewExport ausfhren.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 15.10.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION RunWebViewExport
**************************

	LOCAL lcValue AS STRING, lcFileName AS STRING

	*-- Parameter fr FTP-Upload aus INI-Datei lesen
	lcValue = getKDVini( 'WebViewExport', 'WebViewExport' )
	IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
		lcFileName = lcValue
	ELSE
		lcFileName = 'c:\TMP\WebViewExport\WEBVIEWEXPORT.EXE'
	ENDIF

	*-- Das Arbeitsverzeichnis setzen
	lcRunDir = JUSTPATH( lcFileName )

	*-- Das externe Programm ReadWebViewExport ausfhren
	RunProcess( lcFileName, .F., '', 1, lcRunDir )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Startet einen neuen Prozess mittels W32API-Funktion CreateProcess und wartet auf dessen Beendigung, falls
*				das Wait-Flag gesetzt wurde.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.07.2017
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION RunProcess
*******************

	*-- Parameter:
	*-- tcEXEFileName:		Pfad und Name der auszufhrenden Datei.
	*-- tlWait:				Soll auf das Prozess-Ende gewartet werden.
	*-- tcParameter:		Parameter fr die auszufhrende Datei.
	*-- tnShowWindow:		Anzeigemodus des zu startenden Programmes.
	*--						 - SW_HIDE				0	Hides the window .AND. activates another window.
	*--						 - SW_SHOWNORMAL		1	Activates .AND. displays a window.
	*--						 - SW_SHOWMINIMIZED		2	Activates the window .AND. displays it as a minimized window.
	*--						 - SW_SHOWMAXIMIZED		3	Activates the window .AND. displays it as a maximized window.
	*-- tcWorkingDirectory:	Arbeitsverzeichnis in welchem gearbeitet werden soll. Default ist SYS( 5 ) + SYS( 2003 ).
	*-- Rckgabe:			 - -1						Ungltige Parameter
	*--						 - -2						Datei ist nicht vorhanden
	*--						 - -3						Prozess konnte nicht gestartet werden
	*--						 - > 0						Prozess konnte gestartet werden
	LPARAMETERS tcEXEFileName AS STRING, tlWait AS BOOLEAN, tcParameter AS STRING, tnShowWindow AS INTEGER, ;
		tcWorkingDirectory AS STRING

	LOCAL lcEXEFileName AS STRING, lcParameter AS STRING, lcWorkingDirectory AS STRING, lcStartInfo AS STRING
	LOCAL lcProcessInfo AS STRING, lcError AS STRING
	LOCAL lnShowWindow AS INTEGER, lnProcessID AS INTEGER, lnReturn AS INTEGER

	*-- Parameter prfen
	IF !EMPTY( tcEXEFileName )

		*-- Ist die auszufhrende Datei bzw. die zu bearbeitende,
		*-- zu druckende usw. Datei vorhanden
		IF FileExist( tcEXEFileName )

			*-- Parameter bernehmen bzw. Default-Wert setzen
			lcEXEFileName		= ALLTRIM( tcEXEFileName )
			lcParameter			= ALLTRIM( EVL( tcParameter, ' ' ) )
			lcWorkingDirectory	= ALLTRIM( EVL( tcWorkingDirectory, SYS( 5 ) + SYS( 2003 ) ) )
			lnShowWindow		= EVL( tnShowWindow, SW_SHOWNORMAL )

			*-- Parameter setzen fr CreateProcess
			lcStartInfo = CHR( 68 ) + REPLICATE( NUL, 43 ) + CHR( 1 ) + REPLICATE( NUL, 3 ) + CHR( lnShowWindow ) +  REPLICATE( NUL, 19 )
			lcProcessInfo = REPLICATE( NUL, 16 )

			*-- Prozess starten
			lnReturn = CreateProcess( lcEXEFileName, lcParameter, 0, 0, 0, 0, 0, lcWorkingDirectory, @lcStartInfo, @lcProcessInfo )

			*-- Wenn der Prozess gestartet werden konnte
			IF lnReturn # 0

				*-- Ja, dann holen wir das Prozess-Handle
				lnProcessID = CTOBIN( SUBSTR( lcProcessInfo, 1, 4 ), 'RS' )
				lnThreadHandle = CTOBIN( SUBSTR( lcProcessInfo, 5, 4 ), 'RS' )

				*-- Soll auf das Prozess-Ende gewartet werden?
				IF tlWait

					*-- Ist der gestartete Prozess noch vorhanden?
					DO WHILE WaitForSingleObject( lnProcessID, WAIT_INTERVAL ) == WAIT_TIMEOUT

						*-- Ja, dann Events ermglichen und WAIT_INTERVAL
						*-- (zur Zeit 250 ms) Millisekunden schlafen
						DOEVENTS
						Sleep( WAIT_INTERVAL )
					ENDDO

					*-- Prozess-Handle wieder schlieen
					CloseHandle( lnProcessID )
				ENDIF

			*-- Andernfalls ...
			ELSE

				*-- ... geben wir -3 zurck
				lnProcessID = -3
			ENDIF

		*-- Andernfalls ...
		ELSE

			*-- ... geben wir -2 zurck
			lnProcessID = -2
		ENDIF

	*-- Andernfalls ...
	ELSE

		*-- ... geben wir -1 zurck
		lnProcessID = -1
	ENDIF

	*-- Rckgabe prfen
	IF lnProcessID <= 0

		*-- Fehlernummer holen und dann Fehlermeldung ausgeben
		lcError = TRANSFORM( GetLastError() )
		MESSAGEBOX( 'Prozess nicht startbar', 'Der Prozess ' + lcEXEFileName + ' konnte wegen >' + ;
			lcError + '< nicht gestartet werden!', 16+0+0, 'Fehler' )
	ENDIF

	RETURN( lnProcessID )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Zeit in Zeichenkette umwandeln.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 27.04.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Sec2Time
*****************

	LPARAMETERS lnSeconds AS INTEGER, lnDecimals AS INTEGER

	LOCAL lnHour AS INTEGER, lnMin AS INTEGER
	LOCAL lcTime AS STRING

	IF lnSeconds < 0
		RETURN ( '' )
	ENDIF
	lnDecimals	= IIF( EMPTY( lnDecimals ), 0, lnDecimals )
	lnHour		= INT( lnSeconds / 3600 )
	lnSeconds	= lnSeconds - lnHour * 3600
	lnMin		= INT( lnSeconds / 60 )
	lnSeconds	= lnSeconds - lnMin * 60
	lcTime		= ( STRTRAN( STR( lnHour, 2 ) + ':' + STR( lnMin, 2 ) + ':' + STR( lnSeconds, ;
		IIF( lnDecimals = 0, 2, lnDecimals + 3 ), lnDecimals ), ' ', '0' ) )

	RETURN( lcTime )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Setzt den Steuer-Text fr Auslands-Rechnungen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 01.10.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION SetTaxText( tnTaxIndex AS String )
*******************************************

	*-- Werte zwischen 1 und 4 sind zulssig
	IF BETWEEN( tnTaxIndex, 1, 4 )
		gcTaxInfo = goStamm.cTaxInfo[tnTaxIndex]
	ELSE
		gcTaxInfo = goStamm.cTaxInfo[1]
	ENDIF

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Zeigt einen Meldung in einem WAIT-WINDOW an.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 27.07.2016
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 27.07.16 HPG: Kennzeichen tlNoNoWait und tlNotCentered hinzugefgt.
*               Ist tlNoNoWait gesetzt, wird die Information nur bis zum nchsten MausClick bzw. Tastendruck angezeigt.
*               Ist tlNotCentered gesetzt, zeigen wir das WAIT WINDOW nicht zentriert an.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION ShowWindow
*******************

	*-- Parameter
	*-- tcMessageText:		Anzuzeigender Infotext im WAIT WINDOW.
	*-- tlNoNoWait:			Kennzeichen fr NOWAIT-Option von WAIT WINDOW.
	*-- tlNotCentered:		Kennzeichen frs Zentrieren des WAIT WINDOW.
	LPARAMETERS tcMessageText AS STRING, tlNoNoWait AS BOOLEAN, tlNotCentered AS BOOLEAN

	LOCAL lcMessageText AS STRING, lcMessageExtensionText AS STRING

	*-- Wenn eine Information bergeben wurde ...
	IF PCOUNT() >= 1

		*-- ... prfen wir die weiteren Parameter
		IF PCOUNT() = 1

			*-- Nur ein Parameter bergeben, dann ein zentriertes WAIT WINDOW anzeigen
			WAIT WINDOW tcMessageText AT ( WROWS( 'Screen' ) / 2 ), ( ( WCOLS( 'Screen' ) - LEN( tcMessageText ) ) / 2 ) NOWAIT

		*-- Andernfalls ...
		ELSE

			*-- ... mssen wir die beiden weiteren Kennzeichen auswerten
			DO CASE

				*-- Nicht Warten und nicht zentriert
				CASE tlNoNoWait .AND. tlNotCentered
					WAIT WINDOW tcMessageText

				*-- Warten und nicht zentriert
				CASE !tlNoNoWait .AND. tlNotCentered
					WAIT WINDOW tcMessageText NOWAIT

				*-- Warten und zentriert
				CASE !tlNoNoWait .AND. !tlNotCentered
					WAIT WINDOW tcMessageText AT ( WROWS( 'Screen' ) / 2 ), ( ( WCOLS( 'Screen' ) - LEN( tcMessageText ) ) / 2 ) NOWAIT

				*-- Nicht Warten und zentriert
				CASE tlNoNoWait .AND. !tlNotCentered
					WAIT WINDOW tcMessageText AT ( WROWS( 'Screen' ) / 2 ), ( ( WCOLS( 'Screen' ) - LEN( tcMessageText ) ) / 2 )
			ENDCASE
		ENDIF

	*-- Andernfalls ...
	ELSE

		*-- ... WAIT WINDOW wieder lschen
		WAIT CLEAR
	ENDIF

	RETURN( .T. )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Lscht eine Anzahl Zeichen am Beginn der Zeichenkette.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION StrDelL
****************

	LPARAMETERS tcString AS STRING, tnChars2Delete AS INTEGER

	LOCAL lnLen AS INTEGER, lnChars2Delete AS INTEGER

	*-- Parameter prfen
	IF VARTYPE( tnChars2Delete ) == 'N'
		lnChars2Delete = tnChars2Delete
	ELSE
		lnChars2Delete = 1
	ENDIF

	*-- Lnge bestimmen
	lnLen = LEN( tcString )

	*-- Gewnschte Anzahl Zeichen lschen
	lcReturn = RIGHT( tcString, lnLen - lnChars2Delete )

	*-- Rckgabe
	RETURN( lcReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Lscht eine Anzahl Zeichen am Ende der Zeichenkette.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 06.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION StrDelR
****************

	LPARAMETERS tcString AS STRING, tnChars2Delete AS INTEGER

	LOCAL lnLen AS INTEGER, lnChars2Delete AS INTEGER

	*-- Parameter prfen
	IF VARTYPE( tnChars2Delete ) == 'N'
		lnChars2Delete = tnChars2Delete
	ELSE
		lnChars2Delete = 1
	ENDIF

	*-- Lnge bestimmen
	lnLen = LEN( tcString )

	*-- Gewnschte Anzahl Zeichen lschen
	lcReturn = LEFT( tcString, lnLen - lnChars2Delete )

	*-- Rckgabe
	RETURN( lcReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Wandelt in einer Zeichenkette die Sonderzeichen in eine HTML-gltige Schreibweise.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 19.08.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION String2HTMLString( tcString AS STRING )
************************************************

	LOCAL lcString AS STRING

	lcString = STRTRAN( ALLTRIM( tcString ), '(', '&#40;' )		&& (
	lcString = STRTRAN( lcString, ')', '&#41;' )				&& )
	lcString = STRTRAN( lcString, '>', '&#62;' )				&& >
	lcString = STRTRAN( lcString, '\', '&#92;' )				&& \
	lcString = STRTRAN( lcString, '"', '&#34;' )				&& "
	lcString = STRTRAN( lcString, '#', '&#35;' )				&& #
	lcString = STRTRAN( lcString, "'", '&#39;' )				&& '
	lcString = STRTRAN( lcString, '=', '&#61;' )				&& =

RETURN( lcString )


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Funktion zum Schreiben der INI - Datei
* Autor.......: Hans-Peter Grzinger
* Datum.......: 02.12.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION WriteINI
*****************

	LPARAMETERS tcSection AS STRING, tcKey AS STRING, tcValue AS STRING, tcINIFile AS STRING

	LOCAL lcReturn AS STRING

	*-- Parameter prfen
	IF VARTYPE( tcINIFile ) == 'C'
		lcINIFile = tcINIFile
	ELSE
		lcINIFile = gcINIFileName
	ENDIF

	*-- DECLARE DLL-Anweisungen zum Lesen/Schreiben von privaten INI-Dateien
	DECLARE INTEGER WritePrivateProfileString IN Win32API AS WritePrivStr STRING cSection, STRING cKey, STRING cValue, ;
		STRING cINIFile

	*-- INI-Datei schreiben
	IF WritePrivStr( tcSection, tcKey, tcValue, FULLPATH( lcINIFile ) ) == 0
		MESSAGEBOX( 'Fehler beim Schreiben der INI-Datei!', 16+0+0, 'Fehler' )
	ENDIF

	RETURN( .T. )

ENDFUNC


*############################################################################################################################
* KDV spezifische Funktionen
*############################################################################################################################

*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Klasse fr den Ansprechpartner.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 20.04.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

DEFINE CLASS Contact AS CUSTOM
******************************

	cQuelle		= ''
	cKdNr		= ''
	nID_Kund	= 0
	nASP_ASPNr	= 0
	nAnrede		= 0
	cTitel		= ''
	cFunktion	= ''
	cVorname	= ''
	cName		= ''
	cTel		= ''
	cEMail		= ''

ENDDEFINE


*----------------------------------------------------------------------------------------------------------------------------
*-- Den Anwender etwas fragen.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Ask
************

	LPARAMETERS tcMessage AS STRING, tlDefaultYes AS BOOLEAN

	LOCAL llReturn AS BOOLEAN

	*-- Den Anwender fragen
	IF tlDefaultYes
		llReturn = ( MESSAGEBOX( tcMessage, 4+32+256, 'Frage ...' ) = 6 )
	ELSE
		llReturn = ( MESSAGEBOX( tcMessage, 4+32, 'Frage ...' ) = 6 )
	ENDIF

	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Anrede mit Briefanrede, eventuellem Titel und Name erstellen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 11.05.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 08.07.24 HPG: Neuer Parameter tlOnly4Customer: Rckgabe nur fr den Kunden, Ansprechpartner werden nicht geprft.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

PROCEDURE BuildSalutationEx
***************************

	LPARAMETERS tlWithoutTitle AS BOOLEAN, tlOnly4Customer AS BOOLEAN

	LOCAL lcAnrede AS STRING, lcAnredeTitel AS STRING, lcAnredeTitelName AS STRING, lcName AS STRING, lcTitel AS STRING
	LOCAL lnASPID AS INTEGER

	*-- Die Anrede aufbauen
	lcAnredeTitel = ''
	lcAnredeTitelName = ''

	*-- Haben wir einen Ansprechpartner?
	IF tlOnly4Customer

		*-- Nur fr einen Kunden die Anrede bauen,
		*-- deshalb die ASP-ID auf 0 setzen
		lnASPID = 0

	*-- Ja, die Anrede auch fr einen Ansprechpartner
	*-- bauen
	ELSE
		lnASPID = Bki_Rech.ASP_ID
	ENDIF

	*-- Wir haben einen Ansprechpartner
	IF lnASPID > 0

		*-- Anrede, Titel und Name
		*lcAnrede = GetSalutation4ID( Bki_Ansprech.Anrede, 'B' )
		lcTitel = ALLTRIM( Bki_Ansprech.Titel )
		lcName = ALLTRIM( Bki_Ansprech.NAME )

	*-- Andernfalls Daten von Adresse holen
	ELSE

		*-- Anrede, Titel und Name
		*lcAnrede = GetSalutation4ID( Bkb_Kund.Anrede, 'B' )
		lcTitel = ALLTRIM( Bkb_Kund.Titel )
		lcName = ALLTRIM( Bkb_Kund.Name1 )

		*-- Wenn es ein Leerzeichen im Namen gibt und die Anrede
		*-- ist nicht gleich 'Damen und Herren' bzw. ID = 9,
		*-- holen wir nur das letzte Wort.
		IF Bkb_Kund.Anrede <> 9 .AND. ' ' $ lcName

			*-- Offset des letzten Leerzeichens
			lnOffset = RAT( ' ', lcName )
			lcName = SUBSTR( lcName, lnOffset + 1 )
		ENDIF
	ENDIF

	*-- Wenn die Anrede ohne Titel kommen soll ...
	IF tlWithoutTitle

		*-- ... setzen wir den Inhalt zurck
		lcTitel = ''
	ENDIF

	*-- Bei einem Fehler beim Holen der Anrede
	IF lcAnrede == 'ERROR'
		lcAnredeTitelName = 'Sehr geehrte Damen und Herren,'
	ELSE

		*-- Wenn wir die Briefanrede 'Sehr geehrte
		*-- Damen und Herren' haben
		IF lcAnredeTitel = 'Sehr geehrte Damen und Herren'

			*-- Briefanrede ohne folgende Namen
			lcAnredeTitelName = 'Sehr geehrte Damen und Herren,'

		*-- Andernfalls mit folgendem Namen
		ELSE

			*-- Es ist nur der Kunde zu prfen
			IF tlOnly4Customer
				lcAnredeTitelName = lcAnredeTitel + ';' + lcName
			ELSE
				lcAnredeTitelName = lcAnredeTitel + ' ' + lcName + ','
			ENDIF
		ENDIF
	ENDIF

	RETURN( lcAnredeTitelName )

ENDPROC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Funktion zur Bestimmung des Versand-Gewichtes.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 13.04.2021
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CalcWeight
*******************

	LPARAMETERS tcRechNr AS STRING, tnVerpackung AS STRING, tnBeilage AS STRING, tnWeightMode AS INTEGER, ;
		tlCalcPackaging AS BOOLEAN, tnRoutingLabel AS BOOLEAN

	LOCAL lcMessage AS STRING
	LOCAL lnWeight AS INTEGER, lnVerpackung AS INTEGER, lnBeilage AS INTEGER, lnVerpackungProdukt AS INTEGER, lnCount AS INTEGER
	LOCAL llRoutingLabel AS BOOLEAN
	LOCAL laGewicht[1], laVerpack[1]

	*-- Prfen ob die Tabellen auch offen sind
	IF NOT USED( 'Bki_Beilage' )
		USE ..\dat_kund.bki\Bki_Beilage IN 0
	ENDIF
	IF NOT USED( 'Bki_Verpack' )
		USE ..\dat_kund.bki\Bki_Verpack IN 0
	ENDIF
	IF NOT USED( 'Bki_Prod' )
		USE ..\dat_kund.bki\Bki_Prod IN 0
	ENDIF
	IF NOT USED( 'Bki_Art_Prod' )
		USE ..\dat_kund.bki\Bki_Art_Prod IN 0
	ENDIF
	IF NOT USED( 'Bkb_Art' )
		USE ..\dat_kund.bki\Bkb_Art IN 0
	ENDIF
	IF NOT USED( 'Bkb_Auft' )
		USE ..\dat_kund.bki\Bkb_Auft IN 0
	ENDIF

	*-- Gewichte je Artikel bestimmen
	SELECT Auf.ArtNr, Auf.Menge, Produkt.Gewicht, Verp.Gewicht AS Verpackung, Beil.Gewicht AS Beilage, ;
		( Produkt.Gewicht * Auf.Menge ) AS ArtGesamt, Art.ID AS Art_ID, AProd.ID_Art, Produkt.ID AS Prod_ID ;
		FROM Bkb_Art Art LEFT JOIN Bkb_Auft Auf ON Art.ArtNr = Auf.ArtNr ;
		LEFT JOIN Bki_Art_Prod AProd ON Art.ID = AProd.ID_Art ;
		LEFT JOIN Bki_Prod Produkt ON AProd.ID_Prod = Produkt.ID ;
		LEFT JOIN Bki_Verpack Verp ON Produkt.ID_Verpack = Verp.ID ;
		LEFT JOIN Bki_Beilage Beil ON Produkt.ID_Beilage = Beil.ID ;
		WHERE Auf.RechNr = tcRechNr INTO CURSOR curGewichtjeArt

	*-- Haben wir berhaupt ein Artikel-Gewicht?
	IF _TALLY > 0

		*-- Gesamt-Gewicht, Maximal-Verpackung und Maximal-Beilage bestimmen
		SELECT SUM( ArtGesamt ), MAX( Verpackung ), MAX( Beilage ) FROM curGewichtjeArt INTO ARRAY laGewicht

		*-- 17.11.22 HPG: Wunsch von Tobias Wll: Das Gewicht fr die ComboBox
		*-- bestimmen und dort eintragen, damit der Anwender dieses Gewicht
		*-- gegebenenfalls noch ndern kann.
		IF tlCalcPackaging

			*-- Verpackungs-Gewicht des Versand-Produktes des Artikels bestimmen
			*-- Wie bestimmen wir von einem Artikel das Versand-Produkt: Das
			*-- sollte der Datensatz sein, dessen Beilage 0 betrgt.
			SELECT Verpackung FROM curGewichtjeArt WHERE Beilage = 0 INTO ARRAY laVerpack
			IF _TALLY > 0
				lnVerpackungProdukt = laVerpack[1]
			ELSE
				lnVerpackungProdukt = 0
			ENDIF

			*-- Verpackungs-Gewicht aus dem Formular oder dem Produkt
			*lnVerpackung = IIF( tnVerpackung > 0, tnVerpackung, laGewicht(2) )
			DO CASE

				*-- Mit Versandkosten-Position im Auftrag: 1 - 3
				*-- Wenn dem Artikel nur 1 Produkt zugeordnet ist
				CASE tnWeightMode = 1

					*-- Grtes Verpackungs-Gewicht aller Artikel/Versand-Artikel
					lnVerpackung = laGewicht[2]

				*-- Wenn dem Artikel mehr als 1 Produkt zugeordnet ist
				CASE tnWeightMode = 2

					*-- Grtes Verpackungs-Gewicht aller Artikel/Versand-Artikel
					lnVerpackung = laGewicht[2]

				*-- Verpackungs-Gewicht von der Versandkosten-Position
				CASE tnWeightMode = 3

					*-- Grtes Verpackungs-Gewicht aller Artikel/Versand-Artikel
					lnVerpackung = laGewicht[2]

				*-- Ohne Versandkosten-Position im Auftrag: 4 - 6
				*-- Verpackungs-Gewicht vom einzigen Produkt
				CASE tnWeightMode = 4
					lnVerpackung = laGewicht[2]

				*-- Verpackungs-Gewicht vom Versand-Produkt
				CASE tnWeightMode = 5

					*-- 20.04.21 HPG: Wie bestimmen wir von einem Artikel
					*-- das Versand-Produkt: Das sollte der Datensatz
					*-- sein, dessen Beilage 0 betrgt.
					lnVerpackung = lnVerpackungProdukt

				*-- Verpackungs-Gewicht aus dem Formular
				CASE tnWeightMode = 6
					lnVerpackung = tnVerpackung

				*-- Andernfalls stimmt etwas nicht
				OTHERWISE

					*-- Info an den Anwender
					Fehler( 'Der Modus zur Bestimmung des Verpackungs-Gewichtes hat den ungltigen Wert von: ' + Num2C( tnWeightMode ) + ;
						CHR( 13 ) + 'Das Gewicht wird auf 510 g gesetzt' )

					*-- Dummy-Gewicht
					lnVerpackung = 510
			ENDCASE

		*-- Andernfalls ...
		ELSE

			*-- ... das bergebene Verpackungsgewicht aus der
			*-- ComboBox bernehmen. Das empfohlene Verpack-
			*-- ungs-Gewicht wurde im vorherigen Durchlauf
			*-- schon bestimmt und in die ComboBox eingetragen.
			*-- In diesem Durchlauf bernehmen wir den Eintrag
			*-- direkt aus der ComboBox; der Anwender kann noch
			*-- etwas gendert haben.
			lnVerpackung = tnVerpackung
		ENDIF

		*-- Beilage aus dem Formular oder dem Produkt
		lnBeilage = IIF( tnBeilage > 0, tnBeilage, laGewicht(3) )

		*-- Gesamt-Gewicht bestimmen
		lnWeight = ( laGewicht(1) + lnVerpackung + lnBeilage ) / 1000

	*-- Andernfalls setzen wir ein Dummy-Gewicht von 1kg
	ELSE
		lnWeight = 1.0
	ENDIF

	*-- Wenn wir nur das Verpackungsgewicht bestimmen wollen ...
	IF tlCalcPackaging

		*-- ... setzen wir hier diesen Wert im Rckgabe-Parameter
		lnWeight = lnVerpackung
	ENDIF

	*-- 07.12.20 HPG: Zum Debuggen
	IF .T.
		TRY
			SET SAFETY OFF
			SELECT curGewichtjeArt
			COPY TO C:\TMP\curGewichtjeArt
			STRTOFILE( 'Modus: ' + Num2C( tnWeightMode ) + ' # ArtGesamt: ' + Num2C( laGewicht(1) ) + ' # Verpackung: ' + ;
				Num2C( lnVerpackung ) + ' # Beilage: ' + Num2C( lnBeilage ) + ' # Gesamt: ' + Num2C( lnWeight, 3 ), ;
				'C:\TMP\Gewicht.txt', 0 )
		CATCH
		ENDTRY
		SET SAFETY ON

*!*			TRY
*!*				*-- 12.01.21 HPG: Zum Debuggen
*!*				lcMessage = 'Gewicht: ' + Num2C( lnWeight, 3 ) + ' # RechNr: ' + tcRechNr + ' # Verpackung: ' + ;
*!*					Num2C( lnVerpackung ) + ' # Modus: ' + Num2C( tnWeightMode ) + ;
*!*					' # Beilage: ' + Num2C( tnBeilage ) + CHR( 13 )
*!*				STRTOFILE( lcMessage, 'C:\Tmp\DPD_CalcWeight.log' )
*!*			CATCH
*!*			ENDTRY
	ENDIF

	*-- Aufrumen
	USE IN curGewichtjeArt

	*-- 31.03.23 HPG: Fehlerprfung: wenn lnWeight = .NULL. ist,
	*-- wurde keiner Position des Auftrags ein Produkt zugeordnet.
	*-- Das ist notwendig fr die Gewichtsbestimmung.
	IF ISNULL( lnWeight )

		*-- Haben wir Zugriff auf die Einstellung des
		*-- RoutingLabels?
		IF VARTYPE( tnRoutingLabel ) == 'N'
			llRoutingLabel = IIF( tnRoutingLabel = 1, .T., .F. )
		ELSE
			llRoutingLabel = .F.
		ENDIF

		*-- 17.05.23 HPG: Ist das RoutingLabel gesetzt?
		IF llRoutingLabel

			*-- Ja, dann Fehlermeldung ausgeben
			MESSAGEBOX( 'Keinem Artikel der Rechnung >' + ALLTRIM( tcRechNr ) + '< wurde ein Versandgewicht ber ein Produkt zugeordent!' + CHR( 13 ) + CHR( 13 ) + ;
				'Das Gewicht wird auf 1.0 kg gesetzt.' + CHR( 13 ) + CHR( 13 ) + ;
				'Bitte Information an Jeannette Sturm, Tobias Wll oder Hans-Peter Grzinger', 48+0+0, 'Fehler' )
			lnWeight = 1.0

		*-- Andernfalls ...
		ELSE

			*-- ... setzen wir das Gewicht auf 0.0
			lnWeight = 0.0
		ENDIF
	ENDIF

	RETURN( lnWeight )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
*-- Prfen ob es schon Auftrge mit der bergebenen Abo-ID gibt.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckAboID
*******************

	LPARAMETERS tcAboID AS STRING

	LOCAL lcAboID AS STRING
	LOCAL llReturn AS BOOLEAN
	LOCAL aRechNr
	DECLARE aRechNr[1]

	*-- Parameter formatieren
	lcAboID = ALLTRIM( tcAboID )

	*-- Gibt es schon Auftrge mit dieser Abo-ID?
	SELECT RechNr FROM Bkb_Auft WHERE ALLTRIM( Text ) == lcAboID INTO ARRAY aRechNr
	llReturn = ( _TALLY > 0 )

	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
*-- Hat ein Kunde Ansprechpartner?
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckASPs
******************

	LPARAMETERS tcKdNr AS STRING, tlWithInactive AS BOOLEAN

	LOCAL lnReturn AS INTEGER
	LOCAL aID
	DECLARE aID[1]

	*-- Haben wir berhaupt Ansprechpartner?
	*-- Alle Ansprechpartner zaehlen
	IF tlWithInactive
		SELECT ID FROM Bki_Ansprech WHERE KdNr = tcKdNr INTO ARRAY aID
	ELSE

		*-- 30.06.22 HPG: Inaktive Ansprechpartner werden nicht bercksichtigt
		SELECT ID FROM Bki_Ansprech WHERE KdNr = tcKdNr .AND. ASPInaktiv = .F. INTO ARRAY aID
	ENDIF
	lnReturn = _TALLY

	RETURN( lnReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Kunde bezglich DSVGO-Lschung prfen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 03.11.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.: Ist die Lschung zulssig?
*				Voraussetzung:
*				 - Der Kunde wurde schon gelscht => 'SG'
*				 - Kein Auftrag vorhanden => 'KA'
*				 - Der letzte Auftrag ist lter als 11 Jahre => 'AA'
*				 - Der letzte Auftrag ist jnger als 11 Jahre => 'AN'
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckCustomer
**********************

	LPARAMETERS tcKdNr AS STRING

	LOCAL lcKdNr AS STRING, lcDeletable AS STRING
	DECLARE aDatum[1], aName[1]

	*-- Parameter prfen
	lcKdNr = ALLTRIM( tcKdNr )

	*-- Wurde der Kunde schon DSVGO-gelscht?
	SELECT Name1 FROM Bkb_Kund WHERE KdNr = lcKdNr INTO ARRAY aName
	IF 'DSVGO' $ UPPER( aName[1] )
		lcDeletable = 'SG'

	*-- Andernfalls
	ELSE

		*-- Die Auftrags-Tabelle nach dem neuesten Auftrag durchsuchen
		SELECT MAX( Datum ) FROM Bkb_Auft WHERE KdNr = lcKdNr INTO ARRAY aDatum
		IF _TALLY > 0
			ldToday = DATE()
			IF aDatum[1] < ( ldToday - ( 11 * 366 ) )
				lcDeletable = 'AA'
			ELSE
				lcDeletable = 'AN'
			ENDIF
		ELSE
			lcDeletable = 'KA'
		ENDIF
	ENDIF

	RETURN( lcDeletable )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Routine zum Prfen auf eine gltige EMail-Adresse.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 20.07.2021
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 20.07.21 HPG: Mehrere EMails in einer Zeichenkette sind mittels ';' abzutrennen. Bei der berprfung der Gltigkeit, zer-
*				legen wir hier die Zeichenkette, falls ein ';' gefunden wird.
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckEMailAddress
***************************

	LPARAMETERS tcEMail AS STRING, tlShowResult AS BOOLEAN

	LOCAL lcEMail AS STRING
	LOCAL llReturn AS BOOLEAN
	LOCAL loRegEx AS OBJECT

	*-- Parameter formatieren
	lcEMail = ALLTRIM( tcEMail )
	*tlShowResult = .T.

	*-- RegEx-Objekt erstellen
	loRegEx = CREATEOBJECT( 'VBScript.RegExp' )
	loRegEx.PATTERN = '^([a-zA-Z0-9_\-\.]+)@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.)|(([a-zA-Z0-9\-]+\.)+))([a-zA-Z]{2,4}|[0-9]{1,3})(\]?)$'

	*-- Anzahl der EMails in der Zeichenkette bestimmen
	lnEMailCount = OCCURS( ';', lcEMail ) + 1

	*-- Nur eine EMail
	IF lnEMailCount = 1

		*-- EMail auf Gltigkeit prfen
		llReturn = loRegEx.Test( lcEMail )

	*-- Andernfalls mehr als eine EMail
	ELSE

		*-- Parameter setzen
		llReturn = .T.

		*-- Schleife ber alle EMail-Adressen
		FOR lnEMail = 1 TO lnEMailCount

			*-- EMail-Adresse holen und prfen
			lcOneEMail = GETWORDNUM( lcEMail, lnEMail, ';' )
			llReturn = llReturn .AND. loRegEx.Test( lcOneEMail )

			*-- Info anzeigen
			IF tlShowResult
				WAIT WINDOW 'OneEMail: ' + lcOneEMail + ' # Return: ' + IIF( loRegEx.Test( lcOneEMail ), 'OK', 'Nicht OK' )
			ENDIF
		NEXT
	ENDIF

	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Auf spezielle Zeichen testen und diese bei Vorhandensein ersetzen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 25.01.2021
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

PROCEDURE CheckSpecialCharacter
*******************************

	LPARAMETERS tcString AS STRING, tcRechNr AS STRING

	LOCAL lcString AS STRING, lcCharacter AS STRING, lcTargetFile AS STRING, lnASCII AS STRING
	LOCAL lnLength AS INTEGER, lnOffset AS INTEGER
	LOCAL llSpecialCharacterFound AS BOOLEAN

	*-- Parameter setzen
	llSpecialCharacterFound = .F.

	*-- Zeichenkette bernehmen und die Lnge bestimmen
	lcString = tcString
	lnLength = LEN( lcString )

	*-- Schleife ber die gesamte Zeichenkette
	FOR lnOffset = 1 TO lnLength

		*-- Aktuelles Zeichen holen
		lcCharacter = SUBSTR( lcString, lnOffset, 1 )

		*-- ASCII-Wert bestimmen
		lnASCII = ASC( lcCharacter )

		*-- Wenn kleiner als 32 ...
		IF lnASCII < 32

			*-- ... ersetzen wir dieses bzw. alle im String mit SPACE
			lcString = STRTRAN( lcString, CHR( lnASCII ), ' ' )

			*-- Kennzeichen setzen
			llSpecialCharacterFound = .T.
		ENDIF
	NEXT

	*-- Wenn ein Sonderzeichen gefunden wurde, schreiben wir
	*-- den ursprnglichen String in die Log-Datei.
	IF llSpecialCharacterFound

		*-- 22.12.22 HPG: Erweiterung auf GLS-CSV-Export
		lcTargetFile = gcDPDImport + gcParcelService + 'Export_Error.log'
		STRTOFILE( 'RNr.: ' + ALLTRIM( tcRechNr ) + ' - ' + tcString + CHR( 13 ), lcTargetFile, 1 )
	ENDIF

	RETURN( lcString )

ENDPROC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Auf spezielle Zeichen in einem Tabellen-Feld testen und diese bei Vorhandensein ersetzen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 25.01.2021
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

PROCEDURE CheckSpecialCharacterinField
**************************************

	LPARAMETERS tcRechNr AS STRING

	LOCAL lcString AS STRING, lcCharacter AS STRING, lcTargetFile AS STRING, lnASCII AS STRING, lcDataType AS STRING
	LOCAL lcFieldName AS STRING, lcNewString AS STRING, lcMessage AS STRING
	LOCAL lnLength AS INTEGER, lnOffset AS INTEGER, lnField AS INTEGER, lnFieldCount AS INTEGER
	LOCAL laFields[1]

	*-- Name der Log-Datei setzen
	*-- 22.12.22 HPG: Erweiterung auf GLS-CSV-Export
	lcTargetFile = ADDBS( gcDPDLog ) + gcParcelService + 'Export_Error.log'

	*-- Struktur der aktuell selektierten Tabelle/CURSOR auslesen
	lnFieldCount = AFIELDS( laFields )

	*-- Schleife ber alle Felder
	FOR lnField = 1 TO lnFieldCount

		*-- Feldname und Datentyp des Feldes
		lcFieldName = laFields[lnField, 1]
		lcDataType = laFields[lnField, 2]

		*-- Bei einem Zeichen- oder Memo-Feld
		IF lcDataType == 'C' .OR. lcDataType == 'M'

			*-- Zeichenkette holen
			lcString = EVALUATE( lcFieldName )

			*-- Steht berhaupt etwas im Feld?
			IF .NOT. EMPTY( lcString )

				*-- Die Lnge bestimmen
				lnLength = LEN( lcString )

				*-- Schleife ber die gesamte Zeichenkette
				FOR lnOffset = 1 TO lnLength

					*-- Aktuelles Zeichen holen
					lcCharacter = SUBSTR( lcString, lnOffset, 1 )

					*-- ASCII-Wert bestimmen
					lnASCII = ASC( lcCharacter )

					*-- Wenn kleiner als 32 ...
					IF lnASCII < 32

						*-- ... ersetzen wir dieses bzw. alle im String mit SPACE
						*-- und schreiben den neuen Wert zurck
						lcNewString = STRTRAN( lcString, CHR( lnASCII ), ' ' )
						REPLACE ( lcFieldName ) WITH lcNewString

						*-- Log-Information schreiben
						lcMessage = 'RNr.: ' + ALLTRIM( tcRechNr ) + ' - ' + 'Feld: ' + lcFieldName + ' - ' + ;
							lcString + CHR( 13 )
						STRTOFILE( lcMessage, lcTargetFile, 1 )
					ENDIF
				NEXT
			ENDIF
		ENDIF
	NEXT

	RETURN( lcString )

ENDPROC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Formatiert die Umsatzsteuer anhand den Vorgaben.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 22.08.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 22.08.22 HPG: Alle Lnderkrzel sind zweistellig, dann kommen die Zahlen.
*				Ausnahme sterreich: AT + U => ATU99999...
*				Ausnahme Schweiz: CHE-123.456.789
* 09.03.23 HPG: Den Text UID entfernen.
* 06.12.23 HPG: Neuer Parameter tlUI: Wenn gesetzt, drfen wir nicht auf das Feld COUNTRY prfen sondern nach LAND_ID.
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckVat( tcVat AS STRING, tlUI AS BOOLEAN, tlTableCheck AS BOOLEAN )
******************************************************************************

	LOCAL lcVat AS STRING, lcThirdCaracter AS STRING, lcLeft2 AS STRING, lcLeftSecond AS STRING, lcLeftFirst AS STRING
	LOCAL lcCountry AS STRING, lcPDF AS STRING
	LOCAL lnOffset AS INTEGER

	*-- Parameter setzen
	lcVat = ''
	lcPDF = FULLPATH( 'Aufbau_USt_IdNr_EUMitglieder.pdf' )

	*-- Parameter prfen/formatieren
	IF !EMPTY( ALLTRIM( tcVat ) )

		*-- Leerzeichen, Bindestrich, Schrgstrich und Punkt entfernen
		lcVat = STRTRAN( STRTRAN( STRTRAN( STRTRAN( ALLTRIM( UPPER( tcVat ) ), ' ', '' ), '-', '' ), '/', '' ), '.', '' )

		*-- An dritter Stelle muss bei AT ein U kommen
		lcThirdCaracter = SUBSTR( lcVat, 3, 1 )
		lcLeftFirst = LEFT( lcVat, 1 )
		lcLeft2 = LEFT( lcVat, 2 )

		*-- Auf zwei bzw. drei (ATU) fhrende Zeichen prfen
		DO CASE

			*-- Fhrendes U bzw. DE
			CASE lcLeftFirst == 'U' .OR. lcLeft2 == 'DE'

				*-- Haben wir ein DE in der Zeichenkette?
				lnOffset = AT( 'DE', lcVat )
				IF lnOffset > 0

					*-- Den fhrenden Mll entfernen
					lcVat = SUBSTR( lcVat, lnOffset )
				ENDIF

				*-- Wenn die Lnge nicht 9 ist
				IF LEN( lcVat ) <> 9 .AND. tlUI
					OpenDocument( lcPDF )
				ENDIF

			*-- sterreich
			CASE lcLeft2 == 'AT'

				*-- Alles OK, nix zu tun
				IF lcThirdCaracter == 'U'

				*-- Andernfalls
				ELSE

					*-- Das dritte Zeichen prfen
					IF ISDIGIT( lcThirdCaracter )

						*-- Wir haben eine Zahl => 'U' einfgen
						lcVat = 'ATU' + SUBSTR( lcVat, 3 )

					*-- Andernfalls ...
					ELSE

						*-- ... diese Umsatzsteuer zum Prfen markieren
						lcVat = '+ ' + lcVat
					ENDIF
				ENDIF

				*-- Wenn die Lnge nicht 9 ist
				IF LEN( lcVat ) <> 9 .AND. tlUI
					OpenDocument( lcPDF )
				ENDIF

			*-- Schweiz
			CASE lcLeft2 == 'CH'

				IF lcThirdCaracter == 'E'

				*-- Alles OK, nix zu tun
				ELSE

					*-- Das dritte Zeichen prfen
					IF ISDIGIT( lcThirdCaracter )

						*-- Wir haben eine Zahl => 'E' einfgen
						lcVat = 'CHE' + SUBSTR( lcVat, 4 )

					*-- Andernfalls ...
					ELSE

						*-- ... diese Umsatzsteuer zum Prfen markieren
						lcVat = '+ ' + lcVat
					ENDIF
				ENDIF

				*-- Wenn die Lnge nicht 9 ist
				IF LEN( lcVat ) <> 12 .AND. tlUI
					OpenDocument( lcPDF )
				ENDIF

			*-- Andernfalls
			OTHERWISE

				*-- Das zweite Zeichen holen
				lcLeftSecond = SUBSTR( lcVat, 2, 1 )

				*-- Haben wir zwei fhrende Zeichen
				IF ISALPHA( lcLeft2 ) .AND. ISALPHA( lcLeftSecond )

					*-- Das dritte Zeichen prfen
					IF ISDIGIT( lcThirdCaracter )

						*-- Wir haben eine Zahl => Alles OK, nix zu tun

					*-- Andernfalls ...
					ELSE

						*-- ... diese Umsatzsteuer zum Prfen markieren
						lcVat = '+ ' + lcVat

						IF tlUI

							*-- PDF ffnen
							OpenDocument( lcPDF )
						ENDIF
					ENDIF

				*-- Andernfalls ...
				ELSE

					*-- Prfen wird in der UI?
					IF tlUI .OR. tlTableCheck

						*-- Haben wir eine Land-ID?
						IF LAND_ID > 0 .AND. !EMPTY( lcVat )

							*-- Das Lnderkrzel holen und UPPERn
							lcCountry = UPPER( GetCountryShortCut4ID( LAND_ID ) )
							IF lcCountry = 'AT'
								lcCountry = 'ATU'
							ENDIF

							*-- Das Lnderkrzel davorstellen
							lcVat = lcCountry + lcVat

						*-- Andernfalls ...
						ELSE

							*-- ... diese Umsatzsteuer zum Prfen markieren
							lcVat = '+ ' + lcVat

							IF tlUI

								*-- PDF ffnen
								OpenDocument( lcPDF )
							ENDIF
						ENDIF

					*-- Andernfalls beim Import
					ELSE

						*-- Haben wir ein Lnderkrzel?
						IF !EMPTY( Country ) .AND. !EMPTY( lcVat )

							*-- Das Lnderkrzel holen und UPPERn
							lcCountry = UPPER( LEFT( Country, 2 ) )
							IF lcCountry = 'AT'
								lcCountry = 'ATU'
							ENDIF

							*-- Das Lnderkrzel davorstellen
							lcVat = lcCountry + lcVat

						*-- Andernfalls ...
						ELSE

							*-- ... diese Umsatzsteuer zum Prfen markieren
							lcVat = '+ ' + lcVat

							IF tlUI

								*-- PDF ffnen
								OpenDocument( lcPDF )
							ENDIF
						ENDIF
					ENDIF
				ENDIF
		ENDCASE

		*-- 09.03.23 HPG: Ein eventuell vorhandenes 'UID' entfernen
		lcVat = HTMLString2String( STRTRAN( lcVat, 'UID', '' ) )
	ENDIF

	RETURN( lcVat )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Alle Rechnungsartikel prfen, ob ein KA2-Artikel dabei ist.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 18.11.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckAllKA2Article()
*****************************

	LOCAL lcAlias AS STRING, lcArtNr AS STRING, lcUserURL AS STRING, lcBitmap AS STRING, lcKdNr AS STRING, lcRechNr AS STRING
	LOCAL lnRecNo AS INTEGER
	LOCAL llReturn AS BOOLEAN

	*-- Parameter setzen
	llReturn = .F.
	lcUserURL = ''

	*-- Umgebung sichern
	lcAlias = ALIAS()
	lnRecNo = RECNO()

	*-- Den Rechnungs-Cursor selektieren
	SELECT cursRechnungen

	*-- Die Kundennummer und Rechnungsnummer holen
	GOTO TOP
	lcKdNr = ALLTRIM( cursRechnungen.KdNr )
	lcRechNr = ALLTRIM( cursRechnungen.RechNr )

	*-- Schleife ber alle Rechnungs-Positionen
	SCAN

		*-- Die Artikelnummer holen und prfen
		lcArtNr = ALLTRIM( cursRechnungen.ArtNr )
		lcUserURL = CheckKA2Article( lcArtNr, lcKdNr, lcRechNr )

		*-- Es wurde eine URL zurckgegeben => dann Schleife beenden
		IF LEN( lcUserURL ) > 0
			llReturn = .T.
			EXIT
		ENDIF
	ENDSCAN

	*-- Wenn es einen KA2-Artikel gibt ...
	IF llReturn

		*-- ... erstellen wir den zugehrigen QRCode
		*-- Pfad und Name des Bitmaps
		lcBitmap = 'tmp\' + lcKdNr + '_' + lcRechNr + 'KA2.bmp'

		*-- Das QR-Code-Bitmap erstellen
		FastQRCode( lcUserURL, lcBitmap )

		*-- Etwas warten
		Sleep( 500 )
	ENDIF

	*-- Die Umgebung wieder herstellen
	SELECT ( lcAlias )
	TRY
		GOTO ( lnRecNo )
	CATCH
	ENDTRY

	*-- Rckgabe
	RETURN( lcUserURL )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die CSV-Datei fr den KA2-Artikel erstellen und auf den SFTP-Server hochladen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 19.11.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CreateCSV4KA2andUpload( tcArtNr AS STRING, tcKdNr AS STRING, tcRechNr AS STRING, tcGUID AS STRING )
************************************************************************************************************

	LOCAL lcServerName AS STRING, lcUserName AS STRING, lcPassWord AS STRING, lcTargetFile AS STRING, lcSourcePath AS STRING
	LOCAL lcFiles AS STRING, lcFile AS STRING, lcSourceFile AS STRING, lcTargetDir AS STRING, lcValue AS STRING
	LOCAL lcErrorMessage AS STRING, lcSourceFileText AS STRING
	LOCAL lnReturn AS INTEGER, lnFile AS INTEGER, lnFileCount AS INTEGER
	LOCAL llReturn AS BOOLEAN
	LOCAL loSFTP AS wwSFTP

	*-- Parameter fr FTP-Upload aus INI-Datei lesen
	lcValue = getKDVini( 'FTP', 'ServerName' )
	IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
		lcServerName = lcValue
	ELSE
		lcServerName = '85.215.117.220'
	ENDIF
	lcValue = getKDVini( 'FTP', 'UserName' )
	IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
		lcUserName = lcValue
	ELSE
		lcUserName = 'sftp_kdv_usr'
	ENDIF
	lcValue = getKDVini( 'FTP', 'PassWord' )
	IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
		lcPassWord = lcValue
	ELSE
		lcPassWord = '175!FD4E#E6E'
	ENDIF
	lcValue = getKDVini( 'FTP', 'TargetDirKA2' )
	IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
		lcTargetDir = lcValue
	ELSE
		lcTargetDir = 'Jobs_KA2'
	ENDIF

	*-- Parameter setzen
	lcErrorMessage = ''

	*-- CSV-Datei zusammenstellen
	lcSourceFileText = 'CustomerId;BillNumber;Uuid' + CR
	lcSourceFileText = lcSourceFileText + tcKdNr + ';' + tcRechNr + ';' + tcGUID
	lcSourceFile = tcKdNr + '_' + tcArtNr + '.csv'

	*-- 18.11.24 HPG: wir mssen den Text noch in UTF-8 wandeln
	*-- ANSI in UTF-8 umwandeln
	lcSourceFileText = STRCONV( lcSourceFileText, 5 )			&& Multi-Byte in Unicode umwandeln
	lcSourceFileText = STRCONV( lcSourceFileText, 10 )			&& Unicode in UTF-8 umwandeln

	*-- Und die CSV-Datei schreiben
	SET SAFETY OFF
	STRTOFILE( lcSourceFileText, lcSourceFile, 0 )
	SET SAFETY ON

	*-- Dateiname mit Pfad frs hochladen
	lcSourceFile = SYS( 5 ) + SYS( 2003 ) + '#' + lcSourceFile

	*-- Wenn wir alles einlesen konnten ...
	IF !EMPTY( lcServerName ) .AND. !EMPTY( lcUserName ) .AND. !EMPTY( lcPassWord ) .AND. !EMPTY( lcTargetDir )

		*-- ... FTP-Objekt erstellen, auf passiven Modus umstellen,
		*-- mit dem Server verbinden und Abbruch-Key setzen
		loSFTP = CREATEOBJECT( 'wwSFTP' )
		loSFTP.nFtpPort = 22
		lnReturn = loSFTP.FtpConnect( lcServerName, lcUserName, lcPassWord )

		*-- FTP-Verbindung konnte nicht erstellt werden
		IF lnReturn # 0

			*-- Info in Log-Datei schreiben
			lcErrorMessage = 'SFTP-Problem: ' + loSFTP.cErrorMsg
			STRTOFILE( lcErrorMessage + CR, 'SFTPConfig.log', 1 )

		*-- Andernfalls alles OK
		ELSE

			*-- Hotkey frs Abbrechen setzen
			ON KEY LABEL ALT-X loSFTP.lCancelDownload = .T.

			*-- In das Jobs-Verzeichnis wechseln
			llReturn = loSFTP.FtpSetDirectory( lcTargetDir )

			*-- Konnte in das Verzeichnis gewechselt werden?
			IF llReturn

				*-- Parameter zerlegen
				lcSourcePath = GETWORDNUM( lcSourceFile, 1, '#' )
				lcFiles = GETWORDNUM( lcSourceFile, 2, '#' )

				*-- Anzahl der Dateien bestimmen
				lnFileCount = GETWORDCOUNT( lcFiles, ',' )

				*-- Schleife ber alle Dateien
				FOR lnFile = 1 TO lnFileCount

					*-- Quell- und Zieldateinamen bestimmen
					lcFile = GETWORDNUM( lcFiles, lnFile, ',' )
					lcTargetFile = lcTargetDir + '/' + lcFile
					lcSourceFile = ADDBS( lcSourcePath ) + lcFile
					lnReturn = 0

					*-- Ist die Quelldatei vorhanden
					IF FileExist( lcSourceFile )

						*-- Datei/en per WWIPSTUFF hochladen
						lnReturn = loSFTP.FTPSendFileEx( lcSourceFile, lcTargetFile )

						*-- Wenn Problem beim hochladen
						IF lnReturn # 0

							*-- Info in Log-Datei schreiben
							lcErrorMessage = 'FTP-Problem mit Datei >' + lcTargetFile + '<: ' + loSFTP.cErrorMsg
							STRTOFILE( lcErrorMessage + CR, 'SFTPConfig.log', 1 )
						ENDIF

					*-- Andernfalls Quelldatei nicht gefunden
					ELSE

						*-- Info in Log-Datei schreiben
						lcErrorMessage = 'Problem mit Verzeichnis >' + lcTargetDir + '<'
						STRTOFILE( lcErrorMessage + CR, 'SFTPConfig.log', 1 )
					ENDIF
				ENDFOR

			*-- Andernfalls gibt es ein Problem mit dem Verzeichnis
			ELSE

				*-- Info in Log-Datei schreiben
				lcErrorMessage = 'SFTP-Problem: ' + loSFTP.cErrorMsg
				STRTOFILE( lcErrorMessage + CR, 'SFTPConfig.log', 1 )
			ENDIF

			*-- Hotkey wieder zurcksetzen und FTP-Verbindung wieder trennen
			ON KEY LABEL ALT-X
			loSFTP.FtpClose()
		ENDIF

	*-- Andernfalls fehlen bentigte SFTP-Daten
	ELSE

		*-- Info in Log-Datei schreiben
		lcErrorMessage = 'Einer oder mehrere Parameter fehlen fr die SFTP-bertragung: ' + CR + ;
			'ServerName, UserName, PassWord oder TargetDir'
		STRTOFILE( lcErrorMessage + CR, 'SFTPConfig.log', 1 )
	ENDIF

	*-- Es ist ein Fehler aufgetreten
	IF lcErrorMessage > ' '

		*-- Info fr den Anwender anzeigen
		MESSAGEBOX( 'Problem bei bertragung der CSV-Datei.' + CR + lcErrorMessage, 48+0+0, 'bertragungs-Fehler' )
		llReturn = .F.

	*-- Andernfalls war der Upload erfolgreich
	ELSE

		*-- Info fr den Anwender anzeigen
		MESSAGEBOX( 'bertragung der CSV-Datei erfolgreich.', 64+0+0, 'Datei-bertragung' )
		llReturn = .T.

		*-- CSV-Datei lschen
		DELETE FILE ( lcSourceFile )
	ENDIF

	*-- Status zurckgeben
	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Den Rechnungsartikel prfen, ob es ein besonderer Artikel ist.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 19.11.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Check4SpecialArticle( tcArtNr AS STRING, tcKdNr AS STRING, tcRechNr AS STRING, tdBillDate AS DATE )
************************************************************************************************************

	LOCAL lcUserURL AS STRING, lcReturn AS STRING, lcGUID AS STRING, lcBillText AS STRING, lcAlias AS STRING, lcJSON AS STRING
	LOCAL lcResult AS STRING, lcJSONOrder AS STRING, lcJSONArticle AS STRING
	LOCAL lnRecNo AS INTEGER
	LOCAL llReturn AS BOOLEAN

	*-- Umgebung sichern
	lcAlias = ALIAS()
	lnRecNo = RECNO()

	*-- Parameter setzen
	lcUserURL = ''
	lcBillText = ''
	lcReturn = ''

	*-- Eine GUID erzeugen wir immer und tragen diese
	*-- in das Feld BKB_AUFT1.UUID_PROD1 ein
	lcGUID = CreateBKIGUID()
	REPLACE Bkb_Auft1.UUID_PROD1 WITH lcGUID IN Bkb_Auft1

	*-- Haben wir ein KA2-Artikel?
	lcUserURL = CheckKA2Article( tcArtNr, tcKdNr, tcRechNr )
	IF !EMPTY( lcUserURL )

		*-- Ja, wir haben einen KA2-Artikel
		*-- CSV-Datei fr den KA2-Artikel erstellen
		*-- und per SFTP hochladen
		llReturn = CreateCSV4KA2andUpload( tcArtNr, tcKdNr, tcRechNr, lcGUID )
	ENDIF

	*-- Haben wir einen WebView-Artikel?
	lcBillText = CheckWebViewArticle( tcArtNr, tdBillDate )
	IF !EMPTY( lcBillText )

		*-- Ja, wir haben einen WebView-Artikel
		*-- Dann zuerst den JSON-String fr den Artikel
		*-- erstellen mit Platzhalter
		lcJSONArticle = Article2JSON( tcKdNr, tcRechNr, tcArtNr )

		*-- Dann die kompletten Auftragsdaten in einen
		*-- JSON-String schreiben
		lcJSONOrder = ArticleOrder2JSON( tcRechNr, tcArtNr, tdBillDate )

		*-- Den Auftrag als JSON in den Artikel-JSON-
		*-- String einbetten - den Platzhalter ersetzen.
		*-- GANZ WICHTIG: wir mssen auch die beiden
		*-- doppelten Hochkommas um den Platzhalter mit
		*-- ersetzen.
		lcJSON = STRTRAN( lcJSONArticle, '"1####1"', lcJSONOrder )

		*-- Die Meldung des WebView-Artikels durchfhren
		*-- tcMode:			'A' => ADD
		*-- tcJSON:			JSON-String mit der Artikel-Nummer und den
		*--					kompletten Auftrags-Daten
		*-- tlDontShowInfo:	Keine Meldung anzeigen
		*-- tlNoUpload:		Keine Daten hochladen, nur JSON-Datei erstellen
		*STRTOFILE( lcJSON, 'C:\tmp\WebViewArticle.JSON', 0 )
		lcJSON = HandleWebViewData( 'A', lcJSON, .F., .F. )
		IF LEFT( lcJSON, 5 ) <> 'ERROR'
			llReturn = .T. .AND. llReturn
		ELSE
			llReturn = .F. .AND. llReturn
		ENDIF
	ENDIF

	*-- Die Umgebung wieder herstellen
	SELECT ( lcAlias )
	TRY
		GOTO ( lnRecNo )
	CATCH
	ENDTRY

	*-- Rckgabe
	RETURN( lcReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Den Rechnungsartikel prfen, ob es ein KA2-Artikel ist.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 19.11.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckKA2Article( tcArtNr AS STRING, tcKdNr AS STRING, tcRechNr AS STRING )
*********************************************************************************

	LOCAL lcUserURL AS STRING

	*-- Parameter setzen
	lcUserURL = ''

	*-- Die Artikelnummer prfen
	IF tcArtNr $ gcKA2Article

		*-- Es ist ein KA2-Artikel => wir erstellen den zugehrigen Link
		lcUserURL = 'https://ebooks.bki.de/KA2_eLCA/BKI_KA2_' + tcKdNr + '_' + tcRechNr + '.zip'
	ENDIF

	*-- Rckgabe
	RETURN( lcUserURL )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Alle Rechnungsartikel prfen, ob ein WebView-Artikel dabei ist.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 14.10.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckAllWebViewArticle()
*********************************

	LOCAL lcAlias AS STRING, lcArtNr AS STRING, lcBitmap AS STRING, lcHash4Key AS STRING, lcKdNr AS STRING, lcURL AS STRING
	LOCAL lcMaxArtNr AS STRING, lcRechNr AS STRING, lcUserURL AS STRING
	LOCAL lnRecNo AS INTEGER, lnDays AS INTEGER, lnMaxDays AS INTEGER
	LOCAL llReturn AS BOOLEAN

	*-- Parameter setzen
	llReturn = .F.
	lnMaxDays = 0
	lcMaxArtNr = ''

	*-- Umgebung sichern
	lcAlias = ALIAS()
	lnRecNo = RECNO()

	*-- Den Rechnungs-Cursor selektieren
	SELECT cursRechnungen

	*-- Die Kundennummer und Rechnungsnummer holen
	GOTO TOP
	lcKdNr = ALLTRIM( cursRechnungen.KdNr )
	lcRechNr = ALLTRIM( cursRechnungen.RechNr )

	*-- Schleife ber alle Rechnungs-Positionen
	SCAN

		*-- Die Artikelnummer holen und prfen
		lcArtNr = ALLTRIM( cursRechnungen.ArtNr )
		IF SEEK( lcArtNr, 'cursWebViewArticle', 'ArtNr' )

			*-- Die Laufzeit holen
			lnDays = cursWebViewArticle.Tage

			*-- Den Sonderfall behandeln
			IF lnDays = -1
				lnDays = 1000
			ENDIF

			*-- Schleppzeiger-Prfung
			IF lnDays > lnMaxDays
				lnMaxDays = lnDays
				lcMaxArtNr = lcArtNr
			ENDIF
			llReturn = .T.
		ENDIF
	ENDSCAN

	*-- Die Umgebung wieder herstellen
	SELECT ( lcAlias )
	TRY
		GOTO ( lnRecNo )
	CATCH
	ENDTRY

	*-- Rckgabe
	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Den aktuellen Artikel prfen, ob es ein WebView-Artikel ist. Wenn ja, wird der zugehrige Infotext zurck-
*				gegeben.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 14.10.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckWebViewArticle( tcArtNr AS STRING, tdServiceDate AS DATE )
************************************************************************

	LOCAL lcBillText AS STRING, lcArtNr AS STRING, lcEndDate AS STRING, lcInfoText AS STRING
	LOCAL lnReleaseYear AS INTEGER, lnDays AS INTEGER
	LOCAL ldEndDate AS DATE

	*-- Parameter setzen
	lcBillText = ''

	*-- Parameter prfen
	lcArtNr = ALLTRIM( tcArtNr )

	*-- Die Artikel-Nummer prfen
	IF SEEK( lcArtNr, 'cursWebViewArticle', 'ArtNr' )

		*-- Das Verffentlichungsjahr und die Laufzeit holen
		lnReleaseYear = cursWebViewArticle.Jahr
		lnDays = cursWebViewArticle.Tage

		*-- Endedatum bestimmen
		ldEndDate = tdServiceDate + lnDays
		lcEndDate = DTOC( ldEndDate )

		*-- Daten fr die Kategorie prfen
		DO CASE
			CASE BETWEEN( lnDays, 0, 30 )
				lcBillText = 'Beim Kauf dieses Artikels haben Sie zustzlich kostenlos Zugriff auf die Baupreise-Online bis zum ' + lcEndDate

			CASE lnDays = -1 .OR. lnDays > 30
				IF lnDays = -1
					lcEndDate = '30.09.2026'
				ENDIF
				lcBillText = 'Beim Kauf dieses Artikels haben Sie Zugriff auf die Baupreise-Online bis zum ' + lcEndDate
		ENDCASE
	ENDIF

	*-- 22.10.24 HPG: Zum Debuggen
	*WAIT WINDOW lcArtNr + ' # Days: >' + Num2C( lnDays ) + '< # >' + lcBillText + '<'

	*-- Ist es ein WebView-Artikel, geben wir einen der beiden
	*-- Texte zurck, andernfalls eine leere Zeichenkette
	RETURN( lcBillText )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Webshop-ID holen und damit prfen, ob diese ID nicht schon einer anderen Adresse zugeordnet ist.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 27.09.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckWSID
******************

	LPARAMETERS tnIDWS AS INTEGER, tcKdNr AS STRING

	LOCAL lcKdNr AS STRING
	LOCAL lnFound AS INTEGER, lnRecCount AS INTEGER
	*:Global aKdNr[1]

	*-- Parameter setzen
	lcKdNr = '-1'

	*-- Parameter prfen
	IF VARTYPE( tnIDWS ) == 'N' .AND. VARTYPE( tcKdNr ) == 'C'

		*-- Ist die Webshop-ID nicht schon einer anderen
		*-- Adresse zugeordnet?
		SELECT KdNr FROM Bkb_Kund WHERE ID_WS = tnIDWS .AND. KdNr <> tcKdNr INTO ARRAY aKdNr
		lnRecCount = _TALLY

		*-- Wenn ja, dann alle Kunden-Nummern zurckgeben
		IF lnRecCount > 0

			*-- Die erste gefundene Kunden-Nummern merken
			lcKdNr = ALLTRIM( aKdNr[1] )

			*-- Schleife ber die restlichen gefundenen
			*-- Kunden-Nummern
			FOR lnFound = 2 TO lnRecCount
				lcKdNr = lcKdNr + ',' + aKdNr[lnFound]
			ENDFOR
		ENDIF

	*-- Parameterfehler
	ELSE
		Fehler( 'CheckWSID: Parameterfehler' )
		lcKdNr = '-1'
	ENDIF

	*-- Rckgabe
	RETURN( lcKdNr )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Anhand der Kunden-Nummer alle Email-Adressen des Kunden und dessen Ansprechpartner als CSV zusammenstellen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 11.10.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 29.10.24 HPG: Die erweiterten Daten bentigen wir hier nur fr die eBook-Aktivierung.
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 29.10.24 HPG: Auch die Rechnungs-Email-Adresse holen, falls vorhanden.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CollectEMails4Customer
*******************************

	LPARAMETERS tcKdNr AS INTEGER, tlExtendedData AS BOOLEAN

	LOCAL lcEmails AS STRING, lcEmail AS STRING, lcExtendedData AS STRING, lcAnrede AS STRING, lcName AS STRING
	LOCAL lcTitle AS STRING, lcRechEmail AS STRING
	LOCAL lnASPCount AS INTEGER, lnASP AS INTEGER, lnID AS INTEGER
	LOCAL llRechEmail AS BOOLEAN
	*:Global aExtendedData[1], aEmail[1]

	DECLARE aEmail[3]

	*-- Parameter setzen
	lcEmails = ''
	lcExtendedData = ''

	*-- Parameter prfen
	IF VARTYPE( tcKdNr ) == 'C'

		*-- Email-Adressen des Kunden holen
		*-- 29.10.24 HPG: Auch die Rechnungs-Adresse wenn vorhanden
		SELECT Rech_EMail, EMail_Rech, EMail FROM Bkb_Kund WHERE KdNr = tcKdNr INTO ARRAY aEmail

		*-- Wenn ja, dann die erste EMail-Adresse zurckgeben
		IF _TALLY > 0

			*-- Gibt es eine Rechnungs-EMail-Adresse?
			llRechEmail = aEmail[1]
			IF llRechEmail

				*-- Rechnungs-EMail-Adresse holen und anhngen
				lcRechEmail = ALLTRIM( aEmail[2] )
				lcEmails = lcEmails + ',' + lcRechEmail + ' ' + KDV_DEFAULT_BILLEMAIL
			ENDIF

			*-- EMail-Adresse holen
			lcEmail = ALLTRIM( aEmail[3] )

			*-- Ist eine Email-Adresse eingetragen?
			IF LEN( lcEmail ) > 0

				*-- Die Kunden-EMail-Adresse hinzufgen
				lcEmails = lcEmails + ',' + lcEmail

				*-- Haben wir eine erweiterte Datenabfrage
				IF tlExtendedData

					*-- 08.07.24 HPG: Verwendung der zentralen Funktion BuildSalutationEx
					*-- um die Anrede mit Titel und Name passend aufzubauen
					lcExtendedData = lcExtendedData + BuildSalutationEx( .F., .T. ) + '#'

					*-- 08.07.24 HPG: Zum Debuggen
					*STRTOFILE( lcExtendedData, 'C:\TMP\eBookTest.txt', 0 )
				ENDIF
			ENDIF
		ENDIF

		*-- Email-Adressen der Ansprechpartner der Adresse holen
		SELECT EMail, ID FROM Bki_Ansprech WHERE KdNr = tcKdNr INTO ARRAY aEmail

		*-- Wenn ja, dann alle EMail-Adressen zurckgeben
		lnASPCount = _TALLY
		IF lnASPCount > 0

			*-- Schleife ber alle Ansprechpartner
			FOR lnASP = 1 TO lnASPCount

				*-- EMail-Adresse und ID des Ansprechpartners holen
				lcEmail = ALLTRIM( aEmail[lnASP, 1] )
				lnID = aEmail[lnASP, 2]

				*-- Ist eine Email-Adresse eingetragen?
				IF LEN( lcEmail ) > 0
					lcEmails = lcEmails + ',' + lcEmail

					*-- Haben wir eine erweiterte Datenabfrage
					IF tlExtendedData

						*-- 08.07.24 HPG: Wir bentigen zustzlich zur EMail,
						*-- die Anrede, den Titel und den Namen
						SELECT Anrede, Titel, Name FROM Bki_Ansprech WHERE ID = lnID INTO ARRAY aExtendedData

						*-- 08.07.24 HPG: Verwendung der zentralen Funktion BuildSalutationEx
						*-- um die Anrede mit Titel und Name passend aufzubauen
						*lcAnrede = GetSalutation4ID( aExtendedData[1], 'B' )
						lcTitle = ALLTRIM( aExtendedData[2] )
						lcName = ALLTRIM( aExtendedData[3] )
						lcExtendedData = lcExtendedData + lcAnrede + ';' + lcTitle + ';' + lcName + '#'
					ENDIF
				ENDIF
			NEXT
		ENDIF

		*-- Eine Leerauswahl anhngen
		lcEmails = lcEmails + ', '
		lcEmails = IIF( LEFT( lcEmails, 1 ) = ',', SUBSTR( lcEmails, 2 ), lcEmails )
		lcEmails = STRTRAN( lcEmails, CR, '' )
		lcEmails = STRTRAN( lcEmails, LF, '' )

	*-- Parameterfehler
	ELSE
		Fehler( 'CollectEMails4Customer: Parameterfehler' )
		lcEmails = '-1'
	ENDIF

	*-- Sollen erweiterte Daten zurckgegeben werden?
	IF tlExtendedData

		*-- Das abschlieende '#' als Trenner lschen
		*-- dann die zustzlichen Daten mit '&' getrennt
		*-- an die EMail-CSV anhngen.
		lcExtendedData = StrDelR( lcExtendedData )
		lcEmails = lcEmails + '&' + lcExtendedData
	ENDIF

	*-- Rckgabe
	RETURN( lcEmails )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Funktion zur Erstellung eines EMail-Jobs.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 08.07.2021
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CreateEMailJob
***********************

	LPARAMETERS tcRechNr AS STRING, tcEMail AS STRING, tcBetreff AS STRING, tcText AS STRING, tcBelegArt AS STRING, tcPDF AS STRING

	LOCAL lnSelect AS INTEGER

	*-- Umgebung sichern
	lnSelect = SELECT()

	*-- Prfen ob die Tabellen auch offen sind
	IF NOT USED( 'Bkb_Auft' )
		USE ..\dat_kund.bki\Bkb_Auft IN 0
	ENDIF
	IF NOT USED( 'Bki_EMailJob' )
		USE ..\dat_kund.bki\Bki_EMailJob IN 0
	ENDIF

	*-- Ausgewhlte Rechnung selektieren
	SELECT * FROM Bkb_Auft WHERE RechNR == tcRechNr INTO CURSOR cursRechnung

	*-- Eintrag in der EMail-Job-Tabelle vornehmen
	INSERT INTO Bki_EMailJob ( Modi_Date, Modi_User, KdNr, RechNr, BelegArt, Datum, Leist_Datu, Email_Empf, Email_Betr, ;
		Email_Text, Email_PDFD ) VALUES ( DATETIME(), gu_user, cursRechnung.KdNr, cursRechnung.RechNr, tcBelegArt, ;
		cursRechnung.Datum, cursRechnung.LeistDatum, tcEMail, tcBetreff, tcText, tcPDF )

	*-- Umgebung wieder herstellen
	TRY
		SELECT lnSelect
	CATCH
	ENDTRY

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Erstellt einen QRCode fr den bergebenen Text.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 22.10.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CreateQRCodefromText( tcQRText AS STRING )
***************************************************

	LOCAL lcBitmap AS STRING

	*-- Pfad und Name des Bitmaps
	lcBitmap = 'tmp\' + SYS( 3 ) + '.bmp'

	*-- Das QR-Code-Bitmap erstellen
	FastQRCode( tcQRText, lcBitmap )

	*-- Verweis auf den QRCode zurckgeben
	RETURN( lcBitmap )

ENDFUNC

*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Gibt die URL fr den WebView-Login zurck.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 22.10.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 22.10.24 HPG: tcUserName: KdNr
*				tcPassword: RechNr
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetWebViewURL( tcUserName AS STRING, tcPassword AS STRING )
********************************************************************

	LOCAL lcPassword AS STRING, lcURL AS STRING, lcUserName AS STRING, lcWebViewURL AS STRING

	*-- Parameter prfen
	IF VARTYPE( tcUserName ) == 'C'
		lcUserName = ALLTRIM( tcUserName )
	ELSE
		lcUserName = 'MyLogin'
	ENDIF
	IF VARTYPE( tcPassword ) == 'C'
		lcPassword = ALLTRIM( tcPassword )
	ELSE
		lcPassword = 'MyUser'
	ENDIF

	*-- Parameter setzen
	lcURL = 'https://baupreise-online.bki.de/account/'
	lcWebViewURL = lcURL + CR + CR + 'Kundennummer: ' + lcUserName + CR + CR + 'Rechnungsnummer:' + lcPassword

	*-- URL fr den WebView-Login zurckgeben
	RETURN( lcWebViewURL )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Liste fr die Verpackungsgewicht-ComboBox erstellen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 20.04.2021
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CreateShippingCostList
*******************************

	LOCAL lcList AS STRING

	*-- Parameter setzen
	lcList = ''

	*-- Prfen ob die Tabelle auch offen sind
	IF NOT USED( 'Bki_Verpack' )
		USE ..\dat_kund.bki\Bki_Verpack IN 0 SHARED
	ENDIF

	*-- Daten fr die Liste holen
	SELECT LangText, Gewicht FROM Bki_Verpack ORDER BY Gewicht INTO CURSOR cursList
	SELECT cursList
	SCAN

		*-- Liste aufbauen
		lcList = lcList + Num2C( cursList.Gewicht ) + ' -> ' + ALLTRIM( cursList.LangText ) + ','
	ENDSCAN

	*-- Und wieder zumachen
	USE IN cursList

	RETURN( StrDelr( lcList ) )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Gibt das Datum im Format TimeStamp (Unix epoch) zurck.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 19.08.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION DateTime2TimeStamp( ttDateTime AS DATETIME )
*****************************************************

	LOCAL lnTimeStamp AS INTEGER

	*-- Parameter prfen
	IF VARTYPE( ttDateTime ) <> 'T'
		ttDateTime = DATETIME()
	ENDIF

	*-- Datum in Format TimeStamp (Unix epoch)
	*-- umwandeln und zurckgeben
	lnTimeStamp = ttDateTime - {01.01.1970 00:00:00}
	RETURN( lnTimeStamp )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Erweiterte Suche nach Adressen anhand von Vorgaben.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 16.03.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION DoSearch( toForm AS FORM, tcMessage AS STRING, tcKdNr AS STRING, tcName1 AS STRING, tcName2 AS STRING, ;
	tcPLZ AS STRING, tcOrt AS STRING, tcStrasse AS STRING, tcTelefon AS STRING, tcEMailAnmeld AS STRING )
*****************************************************************************************************************

	LOCAL lcCondition AS STRING, lcASPName AS STRING, lcASPVorname AS STRING, lcEngineBehavior AS STRING, lcName AS STRING
	LOCAL lcOrt AS STRING, lcPLZ AS STRING, lcStrasse AS STRING, lcTelefon AS STRING, lcKdNrAll AS STRING
	LOCAL lnKdNr AS INTEGER, lnTally AS INTEGER
	LOCAL llCaseSensitive AS BOOLEAN

	*-- Parameter fr die Suche setzen/lschen
	llCaseSensitive = toForm.lCaseSensitive

	*-- Suchbegriffe setzen
	lcName1 = tcName1
	lcName2 = tcName2
	lcStrasse = tcStrasse
	lcPLZ = tcPLZ
	lcOrt = tcOrt
	lcTelefon = tcTelefon
	lcEMailAnmeld = tcEMailAnmeld
	*lcASPVorname = tcASPVorname
	*lcASPName = tcASPName

	*-- Parameter setzen
	lcCondition = ''

	*-- Name1 prfen; Suche in Name1, Name2 und Titel
	IF !EMPTY( lcName1 )
		lcName1 = ALLTRIM( IIF( llCaseSensitive, lcName1, UPPER( lcName1 ) ) )
		lcCondition = lcCondition + "and (k.name1 = '" + lcName1 + "' "
		lcCondition = lcCondition + "or k.titel = '" + lcName1 + "' "
		lcCondition = lcCondition + "or k.name2 = '" + lcName1 + "') "
	ENDIF

	*-- Name2 prfen; Suche in Name1, Name2 und Titel
	IF !EMPTY( lcName2 )
		lcAndOr = IIF( toForm.pgfExtendedSearch.pagConfiguration.opgName.VALUE = 1, 'and ', 'or  ' )
		lcName2 = ALLTRIM( IIF( llCaseSensitive, lcName2, UPPER( lcName2 ) ) )
		lcCondition = lcCondition + lcAndOr + " (k.name1 = '" + lcName2 + "' "
		lcCondition = lcCondition + "or k.titel = '" + lcName2 + "' "
		lcCondition = lcCondition + "or k.name2 = '" + lcName2 + "') "
	ENDIF

	*-- Strasse prfen: Suchen in Strasse
	IF !EMPTY( lcStrasse )
		lcStrasse = ALLTRIM( IIF( llCaseSensitive, lcStrasse, UPPER( lcStrasse ) ) )
		lcCondition = lcCondition + "and k.strasse = '" + lcStrasse + "' "
	ENDIF

	*-- Postleitzahl prfen: Suchen in PLZ
	IF !EMPTY( lcPLZ )
		lcPLZ = ALLTRIM( lcPLZ )
		lcCondition = lcCondition + "and k.plz = '" + lcPLZ + "' "
	ENDIF

	*-- Ort prfen: Suchen in Ort
	IF !EMPTY( lcOrt )
		lcOrt = ALLTRIM( IIF( llCaseSensitive, lcOrt, UPPER( lcOrt ) ) )
		lcCondition = lcCondition + "and k.ort = '" + lcOrt + "' "
	ENDIF

	*-- Telefon/Fax/EMail prfen: Suchen in Tel, Tel2, Fax und EMail
	IF !EMPTY( lcTelefon )
		lcTelefon = ALLTRIM( lcTelefon )
		lcCondition = lcCondition + "and (k.tel = '" + lcTelefon + "' "
		lcCondition = lcCondition + "or k.tel2 = '" + lcTelefon + "' "
		lcCondition = lcCondition + "or k.fax = '" + lcTelefon + "' "
		lcCondition = lcCondition + "or k.email = '" + lcTelefon + "') "
	ENDIF

	*-- Anmelde-EMail-Adresse prfen: Suchen in eMail, eMail_Rech und eMail_Anme
	*-- 15.06.23 HPG: Mit ODER statt mit UND suchen
		*lcCondition = lcCondition + "and (UPPER(k.email) like '" + lcEMailAnmeld + "' "
	IF !EMPTY( lcEMailAnmeld )
		lcEMailAnmeld = ALLTRIM( IIF( llCaseSensitive, lcEMailAnmeld, UPPER( lcEMailAnmeld ) ) )

		*-- Die EMail-Adresse optimieren
		lcEMail = lcEMailAnmeld
		IF 'WEB.' $ lcEMail .OR. 'GMX.' $ lcEMail .OR. 'ONLINE.' $ lcEMail .OR. 'AOL.' $ lcEMail .OR. ;
				'ARCOR.' $ lcEMail .OR. 'BKI.' $ lcEMail .OR. 'CHELLO.' $ lcEMail .OR. 'GMAIL.' $ lcEMail .OR. 'GOOGLEMAIL.' $ lcEMail .OR. ;
				'HOTMAIL.' $ lcEMail .OR. 'ICLOUD.' $ lcEMail .OR. 'LIVE.' $ lcEMail .OR. 'MAIL.' $ lcEMail .OR. 'ME.' $ lcEMail .OR. ;
				'OUTLOOK.' $ lcEMail .OR. 'POSTEO.' $ lcEMail .OR. 'YAHOO.' $ lcEMail .OR. 'AON.' $ lcEMail
		*-- Nix zu tun
		ELSE

			*-- Fr die Wildcard-Suche setzen wir das Prozentzeichen davor
			lcEMail = '%' + GETWORDNUM( lcEMail, 2, '@' )
		ENDIF
		lcCondition = lcCondition + "or  (k.email like '" + lcEMail + "' "
		lcCondition = lcCondition + "or k.email_rech like '" + lcEMail + "' "
		lcCondition = lcCondition + "or k.email_anme like '" + lcEMail + "') "
	ENDIF

	*-- 16.03.23 HPG: Fr spter vielleicht noch notwendig
	IF .F.
		*-- Ansprechpartner-Vorname prfen
		IF !EMPTY( lcASPVorname )
			lcASPVorname = ALLTRIM( IIF( llCaseSensitive, lcASPVorname , UPPER( lcASPVorname ) ) )
			lcCondition = lcCondition + "and a.vorname = '" + lcASPVorname + "' "
		ENDIF

		*-- Ansprechpartner-Name prfen
		IF !EMPTY( lcASPName )
			lcASPName = ALLTRIM( IIF( llCaseSensitive, lcASPName, UPPER( lcASPName ) ) )
			lcCondition = lcCondition + "and (a.name = '" + lcASPName + "' "
			lcCondition = lcCondition + "or a.titel = '" + lcASPName + "') "
		ENDIF
	ENDIF

	*-- Keine Bedingungen gesetzt
	IF EMPTY( lcCondition )
		RETURN
	ENDIF

	*-- Das fhrende and entfernen
	lcCondition = SUBSTR( lcCondition, 4 )
	*STRTOFILE( lcCondition + CHR( 13 ), 'Condition.log', 1 )

	*-- Die Einstellung holen und auf 70 setzen
	lcEngineBehavior = SET( 'EngineBehavior' )
	SET ENGINEBEHAVIOR 70

	*-- Und suchen
	*-- 16.09.22 HPG: TRY/CATCH aufgrund Fehler 1300 - Fehlende ) bei Funktionsname - hinzugefgt
	TRY
		*-- 16.03.23 HPG: Fr spter vielleicht noch notwendig
		*SELECT K.*, Bki_Ansprech.NAME AS ASP_Name, Bki_Ansprech.Vorname AS ASP_Vorname, Bki_Ansprech.Titel AS ASP_Titel
		SELECT K.* FROM Kun K WHERE &lcCondition ORDER BY Suchbeg GROUP BY K.KdNr INTO CURSOR cursGefunden
		lnTally = _TALLY
		SELECT cursGefunden

		*-- Je nach Ergebnis
		DO CASE

			*-- Nichts gefunden
			CASE lnTally = 0

				*-- Log schreiben
				toForm.WriteLog( 'LOG', tcMessage + ' # Nichts gefunden' + CR )

			*-- Nur 1 Treffer gefunden
			CASE lnTally = 1

				*-- Die gefunde Kunden-Nummern ins Feld KdNr_Mat eintragen
				REPLACE Imp.KdNr WITH cursGefunden.KdNr

				*-- Log schreiben
				toForm.WriteLog( 'LOG', tcMessage + ' # Gefunden: ' + cursGefunden.KdNr + CR )

			*-- Mehr als 1 bis 20 Treffer gefunden
			CASE lnTally > 1 AND lnTally < 21

				*-- Log schreiben
				toForm.WriteLog( 'LOG', tcMessage + ' # Anzahl Treffer: ' + Num2C( lnTally ) + CR )

				*-- Die gefundenen Kunden-Nummern ins Feld KdNr_Mat eintragen
				lcKdNrAll = ''
				SCAN
					lcKdNrAll = lcKdNrAll + cursGefunden.KdNr + ','
				ENDSCAN

				*-- Kundennummern eintragen; zur Sicherheit das Kunden-Nummern-
				*-- Feld leeren
				SELECT Imp
				REPLACE Imp.KdNr WITH '', Imp.KdNr_Mat WITH StrDelR( lcKdNrAll )

			*-- Mehr 20 Treffer gefunden
			CASE lnTally > 20

				*-- Log schreiben
				toForm.WriteLog( 'LOG', tcMessage + ' # Kein Eintrag - Anzahl Treffer: ' + Num2C( lnTally ) + CR )
		ENDCASE

	*-- Fehler anzeigen
	CATCH

		*-- Log schreiben
		toForm.WriteLog( CR + 'Fehler bei der Suche - bitte den Suchbegriff ndern - Hochkommas sind nicht zulssig in der Suche!' + CR )
	ENDTRY

	*-- Die Einstellung wieder zurcksetzen
	SET ENGINEBEHAVIOR &lcEngineBehavior

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Eine Excel-Tabelle in eine DBF schreiben.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 23.02.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*				1.) Alle Sonderzeichen aus allen Zellen der Excel-Tabelle entfernen, indem wir die Daten zuerst in eine XML-
*					Datei schreiben, in diesem die Sonderzeichen - beginnen mit &# - ersetzen bzw. entfernen und die berar-
*					beitete XML-Datei wieder einlesen.
*				2.) Anzahl der Zeilen und Spalten in der Excel-Tabelle bestimmen, den Bereich mit Daten markieren und in die
*					Zwischenablage kopieren.
*				3.)	Die Zwischenablage - Daten mit TAB getrennt - in die CSV-Datei schreiben. Dadurch sind alle Daten vom
*					Datentyp Text.
*				4.)	Anhand der Anzahl der Spalten sowie den Feldnamem im ersten Datensatz, legen wir den passenden CURSOR
*					an und APPENDEN in diesen die CSV-Datei.
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Excel2DBF( tcExcelFileName AS STRING, tcImportFile AS STRING, toForm AS FORM, taTransform )
****************************************************************************************************

	LOCAL lcCSVFileName AS STRING, lcExtension AS STRING, lcFile AS STRING, lcFile1 AS STRING, lcFile2 AS STRING
	LOCAL lcFile3 AS STRING, lcFile4 AS STRING, lcFile5 AS STRING, lcXLSFileName AS STRING, lcXMLFileName AS STRING
	LOCAL lcXMLFileNameClean AS STRING, lcCommand AS STRING, lcField AS STRING, lcFieldXLS AS STRING, lcHeader AS STRING
	LOCAL lcMessage AS STRING, lcTABData AS STRING, lcExact AS STRING, lcChangedMessage AS STRING, lcNotImported AS STRING
	LOCAL lnColumnCount AS INTEGER, lnRowCount AS INTEGER, lnField AS INTEGER, lnRow AS INTEGER
	LOCAL llTransform AS BOOLEAN, llFieldsMissed AS BOOLEAN
	LOCAL loExcel AS 'Excel.Application'
	EXTERNAL ARRAY taTransform

	*-- Parameter prfen
	llTransform = IIF( VARTYPE( taTransform ) = 'L', .F., .T. )

	*-- Dateinamen vorbereiten
	lcXLSFileName = UPPER( tcExcelFileName )
	lcExtension = '.' + JUSTEXT( lcXLSFileName )
	lcXMLFileName = STRTRAN( lcXLSFileName, lcExtension, '.XML' )
	lcCSVFileName = STRTRAN( lcXLSFileName, lcExtension, '.CSV' )
	lcCSVFileName = STRTRAN( lcCSVFileName, '_MITCR', '_OHNECR' )
	lcXMLFileNameClean = ADDBS( JUSTPATH( lcXMLFileName ) ) + JUSTSTEM( lcXMLFileName ) + '_Clean.XML'

	*-- Excel-Objekt erstellen
	loExcel = CREATEOBJECT( 'Excel.Application' )

	*-- Fortschritts-Balken und -Info anzeigen
	toForm.cntProgressX.UPDATE( 3 )
	toForm.WriteLog( 'LOG', 'Excel-Objekt erstellt' + CR )

	*-- Excel-Tabelle ffnen
	loExcel.Workbooks.OPEN( lcXLSFileName )

	*-- Keine Excel-Warnung beim Speichern
	loExcel.APPLICATION.DisplayAlerts = .F.

	*-- Excel-Tabelle als XML speichern
	*-- xlXMLSpreadsheet 	46 	XML Spreadsheet
	*-- xlTextWindows 		20 	Windows Text
	loExcel.ActiveWorkbook.SAVEAS( lcXMLFileName, 46 )

	*-- Fortschritts-Balken und -Info anzeigen
	toForm.cntProgressX.UPDATE( 4 )
	toForm.WriteLog( 'LOG', 'Excel-Tabelle als XML gespeichert' + CR )

	*-- Excel-Tabelle wieder schlieen
	loExcel.Workbooks.CLOSE

	*-- Die XML-Datei als STRING ffnen und die
	*-- Sonderzeichen entfernen
	lcFile = FILETOSTR( lcXMLFileName )
	lcFile1 = STRTRAN( lcFile, '&#13;&#10;', '|' )
	lcFile2 = STRTRAN( lcFile1, '&#10;', '|' )
	lcFile3 = STRTRAN( lcFile2, '&#160;', '' )
	lcFile4 = STRTRAN( lcFile3, '&#45;', '' )
	lcFile5 = STRTRAN( lcFile4, '&#2;', '' )
	STRTOFILE( lcFile5, lcXMLFileNameClean )

	*-- Fortschritts-Balken und -Info anzeigen
	toForm.cntProgressX.UPDATE( 5 )
	toForm.WriteLog( 'LOG', 'Sonderzeichen aus XML-Datei entfernt' + CR )

	*-- XML-Excel-Tabelle ffnen
	loExcel.Workbooks.OPEN( lcXMLFileNameClean )

	*-- Keine Excel-Warnung beim Speichern
	loExcel.APPLICATION.DisplayAlerts = .F.

	*-- Anzahl der Zeilen und Spalten fr den Bereich bestimmen
	lnRowCount = loExcel.ActiveSheet.Cells( loExcel.ActiveSheet.ROWS.COUNT, 1 ).END( -4162 ).ROW					&& xlUp
	lnColumnCount = loExcel.ActiveSheet.Cells( 1, loExcel.ActiveSheet.Cells.COLUMNS.COUNT).END( -4159 ).COLUMN		&& xlToLeft

	*-- Alle Daten in die Zwischenablage kopieren,
	*-- aber ohne die Titelzeile mit den Feldnamen.
	*-- Dann den Inhalt der Zwischenablage in den
	*-- Parameter schreiben.
	*-- Und zum Schu die Zwischenablage in die CSV-
	*-- Datei mit TAB-getrennt schreiben
	loExcel.RANGE( loExcel.ActiveSheet.Cells( 2, 1 ), loExcel.ActiveSheet.Cells( lnRowCount, lnColumnCount ) ).COPY
	lcTABData = _CLIPTEXT
	STRTOFILE( lcTABData, lcCSVFileName, 0 )

	*-- Die Titelzeile mit den Feldnamen in die
	*-- Zwischenablage kopieren.
	*-- Dann den Inhalt der Zwischenablage in
	*-- den Parameter schreiben aber ohne ange-
	*-- hngtes CR.
	loExcel.RANGE( loExcel.ActiveSheet.Cells( 1, 1 ), loExcel.ActiveSheet.Cells( 1, lnColumnCount ) ).COPY
	lcHeader = SUBSTR( _CLIPTEXT, 1, AT( CR, _CLIPTEXT ) - 1 )

	*-- Fortschritts-Balken und -Info anzeigen
	toForm.cntProgressX.UPDATE( 6 )
	toForm.WriteLog( 'LOG', 'Bereinigte XML-Datei in die Zwischenablage kopiert' + CR )

	*-- Excel beenden
	loExcel.QUIT()

	*-- Excel-Objekt freigeben
	RELEASE loExcel

	*-- Befehlszeichen-Kette frs Erstellen des
	*-- Import-CURSORS aufbauen
	lcCommand = 'CREATE CURSOR cursImport ('

	*-- Fortschritts-Balken und -Info anzeigen
	toForm.cntProgressX.UPDATE( 7 )
	toForm.WriteLog( 'LOG', 'Befehlszeile frs Erstellen der Import-DBF aufbauen' + CR )

	*-- Wir suchen ganz genau
	lcExact = SET( 'Exact' )
	SET EXACT ON

	*-- Schleife ber alle Spalten
	lcChangedMessage = ''
	FOR lnField = 1 TO lnColumnCount

		*-- Feldnamen aus dem Tabellen-Header holen
		lcFieldXLS = UPPER( ALLTRIM( GETWORDNUM( lcHeader, lnField, HT ) ) )

		*-- Feldnamen prfen; wenn er sich gendert hat,
		*-- protokollieren wir das mit
		lcField = Name2DBFFieldName( lcFieldXLS )
		IF lcField <> lcFieldXLS
			lcMessage = 'Feldname >' + lcFieldXLS + '< gendert in >' + lcField + '<' + CR
			lcChangedMessage = lcChangedMessage + lcMessage
		ENDIF

		*-- Muss transformiert werden?
		IF llTransform

			*-- Mssen wir den Feldnamen umsetzen?
			*-- Array-Name:				taTransform
			*-- Zu suchender String:	lcField
			*-- Von Zeile:				1
			*-- Anzahl der Elemente:	-1 => alle
			*-- Suchen in Spalte:		1
			*-- Rckgabe:				8 = Zeilennummer
			lnRow = ASCAN( taTransform, lcField, 1, -1, 1, 8 )

			*-- Das Feld wurde im Transformations-Array gefunden
			IF lnRow > 0

				*-- Feld mit Datentyp und Gre holen
				lcField = taTransform[lnRow, 2]

				*-- Das Import-Kennzeichen setzen
				taTransform[lnRow, 4] = .T.

			*-- Andernfalls nicht gefunden
			ELSE

				*-- Feld mit Datentyp und Gre setzen
				lcField = 'FIELD' + TRANSFORM( lnField, '@L 999' ) + ' C(250)'
			ENDIF
		ENDIF

		*-- Befehlszeile erweitern
		lcCommand = lcCommand + lcField + ','
	ENDFOR

	*-- Musste ein Feldnamen gendert werden?
	IF LEN( ALLTRIM( lcChangedMessage ) ) > 5

		*-- Fortschritts-Balken und -Info anzeigen
		toForm.cntProgressX.UPDATE( 8 )
		toForm.WriteLog( 'LOG', lcChangedMessage )
	ENDIF

	*-- 14.04.23 HPG: Prfen ob alle Muss-Felder vor-
	*-- handen sind. Andernfalls Protokoll schreiben.
	*-- Schleife ber alle Felder des Import-Arrays.
	lcNotImported = 'Folgende Muss-Felder sind nicht vorhanden: '
	lnRowCount = ALEN( taTransform, 1 )
	llFieldsMissed = .F.
	FOR lnRow = 1 TO lnRowCount

		*-- Haben wir ein Muss-Feld?
		IF taTransform[lnRow, 3]

			*-- Ja, wurde es importiert?
			IF taTransform[lnRow, 4] = .F.

				*-- Nein, dann notieren und spter in Log-Datei schreiben
				lcNotImported = lcNotImported + taTransform[lnRow, 1] + ','
				llFieldsMissed = .T.
			ENDIF
		ENDIF
	ENDFOR

	*-- Fehlen Muss-Felder?
	IF llFieldsMissed

		*-- Das letzte Komma lschen
		lcNotImported = StrDelR( lcNotImported ) + CR

		*-- Fortschritts-Balken und -Info anzeigen
		toForm.cntProgressX.UPDATE( 9 )
		toForm.WriteLog( 'LOG', lcNotImported )
	ENDIF

	*-- Das letzte Zeichen lschen und mit ')' abschlieen
	lcCommand = StrDelR( lcCommand ) + ')'

	*-- Befehl zum Erstellen des CURSORs ausfhren
	TRY
		EXECSCRIPT( lcCommand )
	CATCH
		IF DirExist( 'C:\TMP' )
			STRTOFILE( lcCommand, 'C:\TMP\Command.prg' )
		ENDIF
	ENDTRY

	*-- Fortschritts-Balken und -Info anzeigen
	toForm.cntProgressX.UPDATE( 10 )
	toForm.WriteLog( 'LOG', 'Leere Import-DBF erstellt und daran die Daten angehngt' + CR )

	*-- Den Import-Cursor selektieren und die CSV-Datei
	*-- anhngen
	SELECT cursImport
	APPEND FROM ( lcCSVFileName ) TYPE DELIMITED WITH TAB

	*-- Fortschritts-Balken und -Info anzeigen
	toForm.cntProgressX.UPDATE( 11 )
	toForm.WriteLog( 'LOG', 'Aufrumen' + CR )

	*-- Die Import-Tabelle erstellen
	COPY TO ( tcImportFile )

	*-- Aufrumen
	_CLIPTEXT = ''
	DELETE FILE &lcXMLFileName
	DELETE FILE &lcXMLFileNameClean
	DELETE FILE &lcCSVFileName

	*-- Einstellung auf vorherige setzen
	SET EXACT &lcExact

	*-- Informieren
	WAIT CLEAR

	*-- Ende und aus
	RETURN

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Erweiterte Suche nach Adressen anhand von Vorgaben.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 16.03.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION ExtendedSearch( toForm AS FORM, tlWSImport AS BOOLEAN )
****************************************************************

	LOCAL lcCSV AS STRING, lcCSVChecked AS STRING, lcName AS STRING, lcPostName AS STRING, lcPreName AS STRING, lcLand AS STRING
	LOCAL lcEMail AS STRING, lcFax AS STRING, lcKdNr AS STRING, lcKdNrAll AS STRING, lcMessage AS STRING, lcName1 AS STRING
	LOCAL lcName2 AS STRING, lcPLZ AS STRING, lcRecNo AS STRING, lcStrasse AS STRING, lcTelefon AS STRING, lcEMail2 AS STRING
	LOCAL lcOrt AS STRING, lcEMailAnmeld AS STRING
	LOCAL lnKdNr AS INTEGER, lnLand AS INTEGER, lnRecNo AS INTEGER, lnStart AS INTEGER, lnTally AS INTEGER, lnRecCount AS INTEGER
	LOCAL lnStartAt AS INTEGER, lnEndAt AS INTEGER
	LOCAL llSearch AS BOOLEAN

	*-- Grundeinstellungen
	SET ESCAPE ON
	SET TALK OFF

	*-- Parameter setzen
	lnStart = SECONDS()

	*-- Die vorbereitete Kunden-Tabelle ffnen
	IF !USED( 'Kun' )
		USE ( toForm.cModifiedCustomer ) IN 0 ALIAS Kun
	ENDIF

	*-- Die sortierte Tabelle ffnen
	IF !USED( 'Imp' )
		USE ( toForm.cImport ) IN 0 ALIAS Imp
	ENDIF
	SELECT Imp

	*-- Die zweite Reiterseite aktivieren
	toForm.pgfExtendedSearch.ACTIVEPAGE = 2

	*-- Progress-Bar konfigurieren, zurcksetzen und fortschreiben
	lnStartAt = toForm.pgfExtendedSearch.pagConfiguration.txtStartAt.VALUE
	lnRecCount = toForm.pgfExtendedSearch.pagConfiguration.txtRecCount.VALUE
	lnEndAt = lnStartAt + lnRecCount
	toForm.cntProgressX.SetParameters( lnRecCount )
	toForm.cntProgressX.RESET()

	*-- Schleife ber alle Adressen
	SCAN FOR BETWEEN( RECNO(), lnStartAt, lnEndAt )

		*-- Wenn Webhop-Adressen zu importieren sind
		IF tlWSImport

			*-- Wenn die Kundennummer im Webshop schon
			*-- vorhanden ist ...
			IF ArtKZ = 'B91'

				*-- ... mssen wir nicht suchen
				llSearch = .F.

			*-- Andernfalls mssen wir suchen
			ELSE
				llSearch = .T.
			ENDIF

		*-- Andernfalls mssen wir suchen
		ELSE
			llSearch = .T.
		ENDIF

		*-- Mssen wir suchen?
		IF llSearch

			*-- Satz-Nummer
			lnRecNo = RECNO()
			lcRecNo = Num2C( lnRecNo )

			*-- Parameter holen
			lcKdNr = ALLTRIM( Imp.KdNr )

			*-- Nach Name1 suchen
			IF toForm.pgfExtendedSearch.pagConfiguration.chkName1.VALUE
				lcName1 = UPPER( ALLTRIM( Imp.Name1 ) )
			ELSE
				lcName1 = ''
			ENDIF

			*-- Nach Name2 suchen
			IF toForm.pgfExtendedSearch.pagConfiguration.chkName2.VALUE
				lcName2 = UPPER( ALLTRIM( Imp.Name2 ) )
			ELSE
				lcName2 = ''
			ENDIF

			*-- Nach PLZ suchen
			IF toForm.pgfExtendedSearch.pagConfiguration.chkPLZ.VALUE
				lcPLZ = ALLTRIM( Imp.PLZ )
			ELSE
				lcPLZ = ''
			ENDIF

			*-- Nach Ort suchen
			IF toForm.pgfExtendedSearch.pagConfiguration.chkOrt.VALUE
				lcOrt = ALLTRIM( Imp.Ort )
			ELSE
				lcOrt = ''
			ENDIF

			*-- Nach Strasse suchen
			IF toForm.pgfExtendedSearch.pagConfiguration.chkStrasse.VALUE
				lcStrasse = ALLTRIM( Imp.Strasse )
			ELSE
				lcStrasse = ''
			ENDIF

			*-- Nach Telefon suchen
			IF toForm.pgfExtendedSearch.pagConfiguration.chkTelefon.VALUE
				lcTelefon = ALLTRIM( Imp.Telefon )
			ELSE
				lcTelefon = ''
			ENDIF

			*-- Nach Anmelde-EMail-Adresse suchen
			IF toForm.pgfExtendedSearch.pagConfiguration.chkEMailAnmeld.VALUE
				lcEMailAnmeld = ALLTRIM( Imp.EMail_AnMe )
			ELSE
				lcEMailAnmeld = ''
			ENDIF

			*-- Nach Fax suchen
	*!*			IF toForm.pgfExtendedSearch.pagConfiguration.chkFax.VALUE
	*!*				lcFax = ALLTRIM( Imp.Fax )
	*!*			ELSE
	*!*				lcFax = ''
	*!*			ENDIF

	*!*			*-- Nach EMail suchen
	*!*			IF toForm.pgfExtendedSearch.pagConfiguration.chkEMail.VALUE
	*!*				lcEMail = ALLTRIM( Imp.EMail )
	*!*			ELSE
	*!*				lcEMail = ''
	*!*			ENDIF

			*-- Aktuell nicht verwendet
			lnLand = Imp.Land_ID
			lcFax = UPPER( ALLTRIM( Imp.Fax_Mat ) )
			lcEMail = UPPER( ALLTRIM( Imp.EMail_Mat ) )
			lcEMail2 = ALLTRIM( Imp.EMail_Ma2 )

			*-- Message aufbauen
			lcMessage = ALLTRIM( 'Satz: ' + lcRecNo + ' # ' + lcName1 + ' ' + lcName2 + ' ' + lcPLZ + ' ' + lcOrt + ' ' + ;
				lcStrasse + ' ' + lcTelefon + ' ' + lcEMailAnmeld )

			*-- Fortschrittsbalken updaten
			toForm.cntProgressX.UPDATE( RECNO() )

			*-- Und Suchen
			DoSearch( toForm, lcMessage, lcKdNr, lcName1, lcName2, lcPLZ, lcOrt, lcStrasse, lcTelefon, lcEMailAnmeld )

			*-- Den Schleifen-CUROSR wieder selektieren
			SELECT Imp

			*-- Fortschritt anzeigen
			toForm.cntProgressX.lblRecno.CAPTION = Num2C( RECNO() ) + ' / ' + Num2C( RECCOUNT() )
			DOEVENTS
		ENDIF
	ENDSCAN

	*-- 17.05.23 HPG: Das Feld Kunden-Nummer leeren,
	*-- wenn in KdNr_Mat etwas eingetragen ist
	BLANK FIELDS Imp.KdNr FOR !EMPTY( Imp.KdNr_Mat )

	*-- Und wieder zumachen
	USE IN Imp

	*-- Log schreiben
	toForm.WriteLog( 'LOG', 'Fertig nach: ' + Num2C( SECONDS() - lnStart, 2 ) + ' s' + CR )

	*-- ESCAPE wieder deaktivieren
	SET ESCAPE OFF
	SET TALK ON

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Funktion zum Extrahieren des Strassennamens und der Haus-Nummer aus einem kombinierten Feld.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 02.12.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

PROCEDURE ExtractStreetandNo
****************************

	LPARAMETERS tcStreet AS STRING

	LOCAL lcHouseNo AS STRING, lcChar AS STRING, lcSteet AS STRING, lcStrassenName AS STRING
	LOCAL lnLength AS INTEGER, lnOffset AS INTEGER
	LOCAL llHouseNoFound AS BOOLEAN

	*-- Strae holen und Leerzeichen abschneiden
	lcStreet = ALLTRIM( tcStreet )
	lnLength = LEN( lcStreet )
	llHouseNoFound = .F.

	*-- Erste Zahl suchen
	FOR lnOffset = 1 TO lnLength
		lcChar = SUBSTR( lcStreet, lnOffset, 1 )
		IF ISDIGIT( lcChar )
			llHouseNoFound = .T.
			EXIT
		ENDIF
	ENDFOR

	*-- Haben wir eine Hausnummer berhaupt gefunden
	IF llHouseNoFound

		*-- Ja, dann Eins zurck
		lnOffset = lnOffset - 1
		tcStreet = ALLTRIM( LEFT( lcStreet, lnOffset ) )
		lcHouseNo = ALLTRIM( SUBSTR( lcStreet, lnOffset ) )

	*-- Keine Hausnummer
	ELSE
		tcStreet = ALLTRIM( lcStreet )
		lcHouseNo = ''
	ENDIF

	RETURN( lcHouseNo )

ENDPROC


*----------------------------------------------------------------------------------------------------------------------------
*-- Fehler anzeigen.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Fehler
***************

	LPARAMETERS tcMessage AS STRING

	*-- Fehler anzeigen
	MESSAGEBOX( tcMessage, 0+48+0, 'Fehler ...' )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
*-- Den Pfad zur PDF-Rechnung holen.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetBillPath
********************

	LPARAMETERS tdDate AS DATE

	LOCAL lcPath AS STRING

	*-- Den Pfad zu der PDF-Rechnungs setzen
	lcPath = '..\Rechnungen\' + STR( YEAR( tdDate ), 4 ) + '-' + STR( MONTH( tdDate ), 2 ) + '\'

	RETURN( lcPath )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Gibt fr eine Gewichtsangabe den ComboBox-Text frs Verpackungsgewicht zurck.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 18.11.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetCBOTextfromWeight
*****************************

	LPARAMETERS tnPackagingWeight AS INTEGER

	LOCAL lcPackagingWeight AS STRING, lcText AS STRING
	LOCAL lnCount AS INTEGER, lnPackagingWeight AS INTEGER, lnWeight AS INTEGER, lnLoop AS INTEGER

	*-- Parameter prfen
	IF VARTYPE( tnPackagingWeight ) == 'N'
		lnPackagingWeight = tnPackagingWeight
		lcPackagingWeight = Num2C( lnPackagingWeight )

		*-- Aus der Vorgabeliste der ComboBox den richtigen
		*-- Eintrag auswhlen
		lnCount = GETWORDCOUNT( gcShippingCostList, ',' )
		FOR lnLoop = 1 TO lnCount

			*-- Eintrag und davon das Gewicht holen
			lcText = GETWORDNUM( gcShippingCostList, lnLoop, ',' )
			lnWeight = VAL( lcText )
			IF lnWeight = lnPackagingWeight
				lcPackagingWeight = lcText
			ENDIF
		ENDFOR

	*-- Andernfalls
	ELSE
		lcPackagingWeight = ''
	ENDIF

	RETURN( lcPackagingWeight )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
*-- Den Pfad zur PDF-Rechnungskopie holen.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetCopyBillPath
************************

	LPARAMETERS tcRechNr AS STRING

	LOCAL lcPath AS STRING

	*-- Den Pfad zu der PDF-Rechnungs-Kopie setzen
	lcPath = '..\Rechnungskopie\00' + SUBSTR( tcRechNr, 1, 3 ) + '\'

	RETURN( lcPath )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
*-- Dateinamen mit Pfad der PDF-Rechnungskorrektur holen.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetCorrectBillFileName
*******************************

	LPARAMETERS tcRechNr AS STRING, tdDate AS DATE

	LOCAL lcFileName AS STRING

	*-- Den Dateinamen mit Pfad der PDF-Rechnungskorrektur setzen
	lcFileName = FULLPATH( '..\Rechnungen\Rechnungskorrektur\' + STR( YEAR( tdDate ), 4 ) + '-' + STR( MONTH( tdDate ), 2 ) + ;
		'\RK' + tcRechNr + '-S.pdf' )

	RETURN( lcFileName )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Holt den Freischalt-Text fr den Artikel ab.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 07.10.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetFreiSText( tcArtNr AS STRING )
******************************************

	LOCAL lcFreiSText AS STRING, lcArtNr AS STRING
	DIMENSION aFreiSText[1], aID[1]

	*-- Parameter setzen
	lcArtNr = ALLTRIM( tcArtNr )
	lcFreiSText = ''

	*-- Die Freischalt-ID des Artikels holen
	SELECT ID_ArtFTxt FROM Bkb_Art WHERE ArtNr == tcArtNr INTO ARRAY aID
	IF _TALLY > 0

		*-- Den FreiSchalt-Text anhand der ID holen
		SELECT FreiSText FROM Bki_ArtFreiText WHERE ID == aID[1] INTO ARRAY aFreiSText
		IF _TALLY > 0
			lcFreiSText = aFreiSText[1]
		ENDIF
	ENDIF

	*-- Das Freischalt-Text zurckgeben
	RETURN( lcFreiSText )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Wandelt die KDV-Lnder-ID in ein zweistelliges Lnderkrzel um.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 19.08.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetCountryShortCut4ID( tnCountryID AS INTEGER )
********************************************************

	LOCAL lcCountryShortCut AS STRING
	LOCAL lnCountryID AS INTEGER
	DIMENSION aCountryShortCut[1]

	*-- Die Lnder-ID holen
	SELECT Kuerzel FROM Bki_Land WHERE ID == tnCountryID INTO ARRAY aCountryShortCut
	IF _TALLY > 0
		lcCountryShortCut = LOWER( aCountryShortCut[1] )
	ELSE
		lcCountryShortCut = ''
	ENDIF

	*-- Das Lnderkrzel zurckgeben
	RETURN( lcCountryShortCut )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Artikel-Email-Info anhand der ID holen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 23.07.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.: Voraussetzung: Die Tabelle Bki_ArtEMailText muss geffnet sein.
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetArtEMailInfo4ID( tnID AS INTEGER )
**********************************************

	LOCAL lcEMailInfo AS STRING
	LOCAL aInfo[1]

	*-- Parameter setzen
	lcEMailInfo = 'ERROR'

	*-- Ist die Tabelle Bki_ArtEMailText geffnet
	IF USED( 'Bki_ArtEMailText' )

		*-- Den Inhalt des Feldes EMailText holen
		SELECT Info FROM Bki_ArtEMailText WHERE ID = tnID INTO ARRAY aInfo
		IF _TALLY > 0
			lcEMailInfo = ALLTRIM( aInfo[1] )
		ENDIF

	*-- Andernfalls den Anwender informieren
	ELSE
		Fehler( 'Die Tabelle Bki_ArtEMailText ist nicht geffnet!' )
	ENDIF

	*-- Rckgabe
	RETURN( lcEMailInfo )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Artikel-Email-Text anhand der ID holen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 31.01.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.: Voraussetzung: Die Tabelle Bki_ArtEMailText muss geffnet sein.
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 28.10.24 HPG: Wir holen auch den EMail-Betreff aus der Tabelle Bki_ArtEMailText.dbf.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetArtEMailText4ID( tnID AS INTEGER )
**********************************************

	LOCAL lcEMailText AS STRING
	LOCAL aText[1]

	*-- Parameter setzen
	lcEMailText = 'ERROR'

	*-- Ist die Tabelle Bki_ArtEMailText geffnet
	IF USED( 'Bki_ArtEMailText' )

		*-- Den Inhalt des Feldes EMailText und EMailBetreff holen
		SELECT EMailText, EMailBetre FROM Bki_ArtEMailText WHERE ID = tnID INTO ARRAY aText
		IF _TALLY > 0

			*-- Den EMailText und EMailBetreff holen
			lcEMailText = ALLTRIM( aText[1] ) + '#' + ALLTRIM( aText[2] )
		ENDIF

	*-- Andernfalls den Anwender informieren
	ELSE
		Fehler( 'Die Tabelle Bki_ArtEMailText ist nicht geffnet!' )
	ENDIF

	*-- Rckgabe
	RETURN( lcEMailText )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Gibt zu einer HTTP-Fehlernummer den zugehrigen Fehlertext zurck.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 31.08.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetHTTPErrorText( tcHTTPError AS STRING, tnMode AS INTEGER )
*********************************************************************

	LOCAL lcHTTPError AS STRING
	LOCAL lnMode AS INTEGER, lnRow AS INTEGER
	LOCAL laHTTPError

	*-- Parameter prfen
	IF VARTYPE( tnMode ) == 'I'
		lnMode = tnMode
	ELSE
		lnMode = 1
	ENDIF

	*-- Fehler-Array fr die RestAPI
	DECLARE laHTTPError[9, 3]
	laHTTPError[1, 1] = '400'
	laHTTPError[1, 2] = 'Bad Request'
	laHTTPError[1, 3] = 'Die Anfrage-Nachricht war fehlerhaft aufgebaut'
	laHTTPError[2, 1] = '401'
	laHTTPError[2, 2] = 'Unauthorized'
	laHTTPError[2, 3] = 'Die Anfrage kann nicht ohne gltige Authentifizierung durchgefhrt werden'
	laHTTPError[3, 1] = '402'
	laHTTPError[3, 2] = 'Payment Required'
	laHTTPError[3, 3] = 'Bezahlung bentigt'
	laHTTPError[4, 1] = '403'
	laHTTPError[4, 2] = 'Forbidden'
	laHTTPError[4, 3] = 'Die Anfrage wurde mangels Berechtigung des Clients nicht durchgefhrt, bspw. weil der authentifizierte Benutzer nicht berechtigt ist oder ein als HTTPS konfigurierter URL nur mit HTTP aufgerufen wurde'
	laHTTPError[5, 1] = '404'
	laHTTPError[5, 2] = 'Not Found'
	laHTTPError[5, 3] = 'Die angeforderte Ressource wurde nicht gefunden. Dieser Statuscode kann ebenfalls verwendet werden, um eine Anfrage ohne nheren Grund abzuweisen'
	laHTTPError[6, 1] = '405'
	laHTTPError[6, 2] = 'Method Not Allowed'
	laHTTPError[6, 3] = 'Die Anfrage darf nur mit anderen HTTP-Methoden (zum Beispiel GET statt POST) gestellt werden'
	laHTTPError[7, 1] = '406'
	laHTTPError[7, 2] = 'Not Acceptable'
	laHTTPError[7, 3] = 'Die angeforderte Ressource steht nicht in der gewnschten Form zur Verfgung. Gltige Content-Type-Werte knnen in der Antwort bermittelt werden'
	laHTTPError[8, 1] = '407'
	laHTTPError[8, 2] = 'Proxy Authentication Required'
	laHTTPError[8, 3] = 'Analog zum Statuscode 401 ist hier zunchst eine Authentifizierung des Clients gegenber dem verwendeten Proxy erforderlich'
	laHTTPError[9, 1] = '408'
	laHTTPError[9, 2] = 'Request Timeout'
	laHTTPError[9, 3] = 'Innerhalb der vom Server erlaubten Zeitspanne wurde keine vollstndige Anfrage des Clients empfangen'

	*-- Die Fehlernummer suchen
	*-- Array-Name:				laHTTPError
	*-- Zu suchender String:	tcHTTPError
	*-- Von Zeile:				1
	*-- Anzahl der Elemente:	-1 => alle
	*-- Suchen in Spalte:		1
	*-- Rckgabe:				8 = Zeilennummer
	lnRow = ASCAN( laHTTPError, tcHTTPError, 1, -1, 1, 8 )

	*-- Die Fehlernummer wurde gefunden
	IF lnRow > 0

		*-- Je nach Modus den Fehlertext zusammenstellen
		DO CASE
			CASE lnMode = 1
				lcHTTPError = laHTTPError[lnRow, 2]

			CASE lnMode = 2
				lcHTTPError = laHTTPError[lnRow, 3]

			CASE lnMode = 3
				lcHTTPError = laHTTPError[lnRow, 2] + ' # ' + laHTTPError[lnRow, 3]
		ENDCASE

	*-- Andernfalls nicht gefunden
	ELSE

		*-- Freier Text
		lcHTTPError = 'noch nicht hinterlegter Fehler - Info an den Entwickler'
	ENDIF

	*-- Den Fehlertext zurckgeben
	RETURN( lcHTTPError )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Wandelt ein zweistelliges Lnderkrzel in die KDV-Lnder-ID.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 19.08.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetID4CountryShortCut( tcCountryShortCut AS STRING, tcLand AS STRING )
*******************************************************************************

	LOCAL lcCountryShortCut AS STRING
	LOCAL lnCountryID AS INTEGER
	DIMENSION aCountryID[1]

	*-- Parameter prfen
	lcCountryShortCut = UPPER( tcCountryShortCut )

	*-- Die Lnder-ID holen
	SELECT ID, LAND FROM Bki_Land WHERE Kuerzel == lcCountryShortCut INTO ARRAY aCountryID
	IF _TALLY > 0
		lnCountryID = aCountryID[1]
		tcLand = aCountryID[2]
	ELSE
		lnCountryID = 0
		tcLand = ''
	ENDIF

	*-- Die Lnder-ID zurckgeben
	RETURN( lnCountryID )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Gibt anhand der Anrede-ID den Anrede-Text zurck.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 25.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetSalutation4ID4WS( tnID AS INTEGER )
***********************************************

	LOCAL lcSalutation AS STRING

	*-- Anrede holen und die Rckgabe prfen
	*lcSalutation = GetSalutation4ID( tnID, 'A' )
	IF lcSalutation == 'ERROR'
		lcSalutation = 'Nicht gefunden'
	ENDIF

	*-- Rckgabe
	RETURN( lcSalutation )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Funktion zur Bestimmung des Modus frs Ermitteln des Versandgewichtes.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 13.04.2021
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 19.04.21 HPG: Wir mssen hier erkennen, welchen Fall wir zur Berechnung der Versand-Gewichtes haben.
*				Es ist eine Versandkosten-Position vorhanden: Artikel-Nummern: 9982, 9992, 9993, 9994, 9995 (in gcShippingCostPosition)
*				1.)	1 Artikel mit einem Produkt										Verpackungsgewicht immer vom einzigen Produkt
*				2.)	1 Artikel mit mehreren Produkten								Verpackungsgewicht immer vom Versandkosten-Produkt
*																					Alle Gewichte der Produkte mit dem Versandkosten-Produkt
*																					summieren ohne das Gewicht des Versandkosten-Position
*				3.)	Beliebige Kombination von Artikeln, Bundles, Anzahl der Artikel	Verpackungsgewicht immer von Versandkosten-Position
*				Ohne Versandkosten-Position:
*				4.) 1 Artikel mit einem Produkt										Verpackungsgewicht immer vom einzigen Produkt
*				5.) 1 Artikel mit mehreren Produkten (Bundle) + Versand-Produkt		Gewicht immer vom Versand-Produkt
*				Der Anwender muss das Verpackungs-Gewicht ber die ComboBox festlegen:
*				6.) 10 gleiche Artikel mit einem Produkt							ComboBox fr die Anwenderauswahl
*				6.) 1 Artikel mit einem Produkt und 1 Bundle						ComboBox fr die Anwenderauswahl
*				6.) X Artikel mit einem Produkt und Y Bundles						ComboBox fr die Anwenderauswahl
*
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION GetShippingWeightMode
******************************

	LPARAMETERS tcArtikelNr AS STRING, tcRechNr AS STRING

	LOCAL lcArtikelNr AS STRING
	LOCAL lnMode AS INTEGER, lnArtikel AS INTEGER, lnArtikelCount AS INTEGER, lnFirstAmount AS INTEGER
	LOCAL llWithSendingPosition AS BOOLEAN
	DECLARE aAmount[1]

	*-- Parameter setzen
	llWithSendingPosition = .F.

	*-- Prfen ob die Tabellen auch offen sind
	IF NOT USED( 'Bki_Art_Prod' )
		USE ..\dat_kund.bki\Bki_Art_Prod IN 0
	ENDIF
	IF NOT USED( 'Bkb_Art' )
		USE ..\dat_kund.bki\Bkb_Art IN 0
	ENDIF
	IF NOT USED( 'Bkb_Auft' )
		USE ..\dat_kund.bki\Bkb_Auft IN 0
	ENDIF

	*-- Die Artikel-Nummern in ein Array bertragen
	lnArtikelCount = GETWORDCOUNT( tcArtikelNr, ',' )
	IF lnArtikelCount > 0
		DECLARE aArtikelNr[lnArtikelCount]
		FOR lnArtikel = 1 TO lnArtikelCount

			*-- Die einzelne Artikel-Nummer holen
			lcArtikelNr = ALLTRIM( GETWORDNUM( tcArtikelNr, lnArtikel, ',' ) )
			aArtikelNr[lnArtikel] = lcArtikelNr

			*-- Haben wir eine Versandkosten-Position?
			IF lcArtikelNr $ gcShippingCostPosition
				llWithSendingPosition = .T.
			ENDIF
		NEXT

		*-- Rechnungs-Nummer prfen
		*-- Wir bergeben '0000' als Rechnungs-Nummer von der
		*-- automatischen Auftrags-Erstellung und vom Marketing-
		*-- Formular - Hans-Peter Tool - weil die Menge des
		*-- Artikels hier immer 1 ist.
		IF tcRechNr == '0000'

			*-- Menge des Artikels auf 1 setzen (fr Fall 1.
			*-- und 2. sowie 4. und 5.)
			lnFirstAmount = 1

		*-- Andernfalls haben wir eine Rechnungs-Nummer
		ELSE

			*-- Menge im Auftrag prfen
			*-- Dazu gengt es, sich nur die Menge des ersten
			*-- Artikels zu merken. Sind es mehr als ein Artikel,
			*-- kann es Fall 1. oder 2. bzw. 4. oder 5. sowieso
			*-- nicht mehr sein.
			SELECT Menge FROM Bkb_Auft WHERE RechNr = tcRechNr INTO ARRAY aAmount
			IF _TALLY > 0
				lnFirstAmount = aAmount[1]
			ELSE
				lnFirstAmount = 1
			ENDIF
		ENDIF

		*-- 17.11.22 HPG: Zum Debuggen
		*ShowWindow( 'ArtikelCount: ' + Num2C( lnArtikelCount ) + ' # WithSendingPosition: ' + IIF( llWithSendingPosition, 'Ja', 'Nein' ) + ' # FirstAmount: ' + Num2C( lnFirstAmount ), .T. )

		*-- Wenn wir eine Versandkosten-Position haben
		IF llWithSendingPosition

			*-- Bei nur zwei Artikeln mit Versandkosten-Position
			*-- knnte es sich um einen Artikel mit einem Produkt
			*-- (lnMode = 1) oder um einen Artikel mit mehreren
			*-- Produkten - Bundle (lnMode = 2) handeln und der
			*-- Menge 1 sein. Ist die Menge bei einem Artikel hher,
			*-- nehmen wir das Verpackungsgewicht von der Versand-
			*-- kosten-Position.
			*-- Hierzu zhlen wir einfach die Anzahl der Produkte
			*-- im Artikel sowie die Menge der ersten Position
			*-- Zwei Artikel: Artikel und Versandkosten-Artikel
			IF lnArtikelCount = 2 .AND. lnFirstAmount = 1

				*-- Artikel-Nummer holen
				lcArtikelNr = aArtikelNr[1]

				*-- Anzahl der Produkte je Artikel bestimmen
				SELECT Art.ArtNr FROM Bkb_Art Art LEFT JOIN Bki_Art_Prod AProd ON Art.ID = AProd.ID_Art ;
					WHERE Art.ArtNr = lcArtikelNr INTO ARRAY aDummy

				*-- Wenn dem Artikel nur 1 Produkt zugeordnet ist
				IF _TALLY = 1
					lnMode = 1

				*-- Andernfalls mehr als 1 Produkt zugeordnet
				ELSE
					lnMode = 2
				ENDIF

			*-- Andernfalls haben wir mehrere Artikel oder ein
			*-- Artikel mit einer Menge grer 1. In diesem Fall
			*-- nehmen wir das Verpackungsgewicht von der
			*-- Versandkosten-Position
			ELSE
				lnMode = 3
			ENDIF

		*-- Andernfalls haben wir keine Versandkosten-Position
		ELSE

			*-- Bei nur einem Artikel ohne Versandkosten-Position
			*-- knnte es sich um einen Artikel mit einem Produkt
			*-- (lnMode = 4) oder um einen Artikel mit mehreren
			*-- Produkten - Bundle (lnMode = 5) handeln und der
			*-- Menge 1 sein. Ist die Menge bei einem Artikel hher,
			*-- MUSS der Anwender ber die ComboBox das Verpackungs-
			*-- Gewicht selbst setzen.
			*-- Hierzu zhlen wir einfach die Anzahl der Produkte
			*-- im Artikel sowie die Menge der ersten Position
			IF lnArtikelCount = 1 .AND. lnFirstAmount = 1

				*-- Artikel-Nummer holen
				lcArtikelNr = aArtikelNr[1]

				*-- Anzahl der Produkte je Artikel bestimmen
				SELECT Art.ArtNr FROM Bkb_Art Art LEFT JOIN Bki_Art_Prod AProd ON Art.ID = AProd.ID_Art ;
					WHERE Art.ArtNr = lcArtikelNr INTO ARRAY aDummy

				*-- Wenn dem Artikel nur 1 Produkt zugeordnet ist
				IF _TALLY = 1
					lnMode = 4

				*-- Andernfalls mehr als 1 Produkt zugeordnet
				ELSE
					lnMode = 5
				ENDIF

			*-- Andernfalls haben wir mehrere Artikel oder ein
			*-- Artikel mit einer Menge grer 1. In diesem Fall
			*-- muss der Anwender das Verpackungs-Gewicht ber
			*-- die ComboBox selbst festlegen.
			ELSE
				lnMode = 6
			ENDIF
		ENDIF

		*-- 05.02.23 HPG: Zum Debuggen
		IF .F.
			SET SAFETY OFF
			TRY
				STRTOFILE( 'Modus: ' + Num2C( lnMode ) + CR, 'C:\TMP\Modus.txt', 0 )
			CATCH
			ENDTRY
			SET SAFETY ON
		ENDIF

	*-- Andernfalls ist kein Artikel vorhanden
	ELSE

		*-- Fehler anzeigen
		Fehler( 'Kein Artikel vorhanden fr die Gewichtsberechnung - Auftragsanlage wird abgebrochen!' )
		lnMode = 0
	ENDIF

	RETURN( lnMode )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
*-- Hinweis anzeigen.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Hinweis
****************

	LPARAMETERS tcMessage AS STRING

	*-- Hinweis anzeigen
	MESSAGEBOX( tcMessage, 0+64+0, 'Hinweis ...' )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Wandelt die Feld-Bezeichnung von Importdateien in einen gltigen DBF-Feldnamen um.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 23.02.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Name2DBFFieldName( tcFieldName AS STRING )
***************************************************

	LOCAL lcFieldName AS STRING

	*-- Gltigen DBF-Feldnamen bilden
	lcFieldName = STRTRAN( ALLTRIM( tcFieldName ), '', 'ae' )	&&  mit ae ersetzen
	lcFieldName = STRTRAN( lcFieldName, '', 'oe' )				&&  mit oe ersetzen
	lcFieldName = STRTRAN( lcFieldName, '', 'ue' )				&&  mit ue ersetzen
	lcFieldName = STRTRAN( lcFieldName, '', 'AE' )				&&  mit Ae ersetzen
	lcFieldName = STRTRAN( lcFieldName, '', 'OE' )				&&  mit Oe ersetzen
	lcFieldName = STRTRAN( lcFieldName, '', 'UE' )				&&  mit Ue ersetzen
	lcFieldName = STRTRAN( lcFieldName, '', 'ss' )				&&  mit ss ersetzen
	lcFieldName = STRTRAN( lcFieldName, '', 'Grad' )			&&  mit Grad ersetzen
	lcFieldName = STRTRAN( lcFieldName, ' ', '_' )				&&   mit _ ersetzen
	lcFieldName = STRTRAN( lcFieldName, '.', '_' )				&& . mit _ ersetzen
	lcFieldName = STRTRAN( lcFieldName, ':', '_' )				&& : mit _ ersetzen
	lcFieldName = STRTRAN( lcFieldName, '-', '_' )				&& - mit _ ersetzen
	lcFieldName = STRTRAN( lcFieldName, '/', '_' )				&& / mit _ ersetzen
	lcFieldName = STRTRAN( lcFieldName, '\', '_' )				&& \ mit _ ersetzen
	lcFieldName = STRTRAN( lcFieldName, '(', '_' )				&& ( mit _ ersetzen
	lcFieldName = STRTRAN( lcFieldName, ')', '_' )				&& ) mit _ ersetzen
	lcFieldName = STRTRAN( lcFieldName, '@', '_' )				&& @ mit _ ersetzen
	lcFieldName = STRTRAN( lcFieldName, '[', '_' )				&& [ mit _ ersetzen
	lcFieldName = STRTRAN( lcFieldName, ']', '_' )				&& ] mit _ ersetzen

	*-- Gltigen Feldnamen zurckgeben
	RETURN( lcFieldName )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Formatiert die Telefon- und Faxnummer fr die KDV.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 09.03.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Phone2Tel( tcString AS STRING, tcCountry AS STRING )
*************************************************************

	LOCAL lcString AS STRING

	lcString = STRTRAN( ALLTRIM( tcString ), '&#40;', '(' )		&& (
	lcString = STRTRAN( lcString, '&#41;', ')' )				&& )
	lcString = STRTRAN( lcString, '(', '' )
	lcString = STRTRAN( lcString, ')', '' )
	lcString = STRTRAN( lcString, ' - ', ' ' )
	lcString = STRTRAN( lcString, '-', ' ' )
	lcString = STRTRAN( lcString, ' / ', ' ' )
	lcString = STRTRAN( lcString, '/', ' ' )
	lcString = STRTRAN( lcString, '+0039', '0039' )
	lcString = STRTRAN( lcString, '+01', '01' )
	lcString = STRTRAN( lcString, '+02', '02' )
	lcString = STRTRAN( lcString, '+1 ', '001' )
	lcString = STRTRAN( lcString, '+1-', '001' )
	*-- Griechenland
	IF LOWER( tcCountry ) = 'gr'
		lcString = STRTRAN( lcString, '+30 ', '0030' )
	ENDIF
	*-- Niederlande
	IF LOWER( tcCountry ) = 'nl'
		lcString = STRTRAN( lcString, '+31 ', '0031' )
		lcString = STRTRAN( lcString, '+31', '0031' )
	ENDIF
	*-- Belgien
	IF LOWER( tcCountry ) = 'be'
		lcString = STRTRAN( lcString, '+32 ', '0032' )
		lcString = STRTRAN( lcString, '+32', '0032' )
	ENDIF
	*-- Frankreich
	IF LOWER( tcCountry ) = 'fr'
		lcString = STRTRAN( lcString, '+33 ', '0033' )
		lcString = STRTRAN( lcString, '+33', '0033' )
	ENDIF
	*-- Spanien
	IF LOWER( tcCountry ) = 'es'
		lcString = STRTRAN( lcString, '+34 ', '0034' )
		lcString = STRTRAN( lcString, '+34', '0034' )
	ENDIF
	*-- Luxemburg
	IF LOWER( tcCountry ) = 'lu'
		lcString = STRTRAN( lcString, '+352 ', '00352' )
		lcString = STRTRAN( lcString, '+352', '00352' )
	ENDIF
	*-- Irland
	IF LOWER( tcCountry ) = 'ie'
		lcString = STRTRAN( lcString, '+353 ', '00353' )
		lcString = STRTRAN( lcString, '+353', '00353' )
	ENDIF
	*-- Finnland
	IF LOWER( tcCountry ) = 'fi'
		lcString = STRTRAN( lcString, '+358 ', '00358' )
		lcString = STRTRAN( lcString, '+358', '00358' )
	ENDIF
	*-- Bulgarien
	IF LOWER( tcCountry ) = 'bg'
		lcString = STRTRAN( lcString, '+359 ', '00359' )
		lcString = STRTRAN( lcString, '+359', '00359' )
	ENDIF
	*-- Ungarn
	IF LOWER( tcCountry ) = 'hu'
		lcString = STRTRAN( lcString, '+36 ', '0036' )
		lcString = STRTRAN( lcString, '+36', '0036' )
	ENDIF
	*-- Litauen
	IF LOWER( tcCountry ) = 'lt'
		lcString = STRTRAN( lcString, '+370 ', '00370' )
		lcString = STRTRAN( lcString, '+370', '00370' )
	ENDIF
	*-- Lettland
	IF LOWER( tcCountry ) = 'lv'
		lcString = STRTRAN( lcString, '+371 ', '00371' )
		lcString = STRTRAN( lcString, '+371', '00371' )
	ENDIF
	*-- Estland
	IF LOWER( tcCountry ) = 'ee'
		lcString = STRTRAN( lcString, '+372 ', '00372' )
		lcString = STRTRAN( lcString, '+372', '00372' )
	ENDIF
	*-- Norwegen Korrektur
	IF LOWER( tcCountry ) = 'no' .AND. LEFT( tcCountry, 4 ) = '+372'
		lcString = STRTRAN( lcString, '+372 ', '0047' )
		lcString = STRTRAN( lcString, '+372', '0047' )
	ENDIF
	*-- ??
	lcString = STRTRAN( lcString, '+377', '00377' )
	lcString = STRTRAN( lcString, '+381', '00381' )
	*-- Kroatien
	IF LOWER( tcCountry ) = 'hr'
		lcString = STRTRAN( lcString, '+385 ', '00385' )
		lcString = STRTRAN( lcString, '+385', '00385' )
	ENDIF
	*-- Slowenien
	IF LOWER( tcCountry ) = 'si'
		lcString = STRTRAN( lcString, '+386 ', '00386' )
		lcString = STRTRAN( lcString, '+386', '00386' )
	ENDIF
	*-- Italien
	IF LOWER( tcCountry ) = 'it'
		lcString = STRTRAN( lcString, '+39 ', '0039' )
		lcString = STRTRAN( lcString, '+39', '0039' )
	ENDIF
	*-- Rumnien
	IF LOWER( tcCountry ) = 'ro'
		lcString = STRTRAN( lcString, '+40 ', '0040' )
		lcString = STRTRAN( lcString, '+40', '0040' )
	ENDIF
	*-- Schweiz
	IF LOWER( tcCountry ) = 'ch'
		lcString = STRTRAN( lcString, '+41 ', '0041' )
		lcString = STRTRAN( lcString, '+41', '0041' )
	ENDIF
	*-- Tschechien
	IF LOWER( tcCountry ) = 'cz'
		lcString = STRTRAN( lcString, '+420 ', '00420' )
		lcString = STRTRAN( lcString, '+420', '00420' )
	ENDIF
	*-- Slowakei
	IF LOWER( tcCountry ) = 'sk'
		lcString = STRTRAN( lcString, '+421 ', '00421' )
		lcString = STRTRAN( lcString, '+421', '00421' )
	ENDIF
	*-- sterreich
	IF LOWER( tcCountry ) = 'at'
		lcString = STRTRAN( lcString, '+43 ', '0043' )
		lcString = STRTRAN( lcString, '+43', '0043' )
	ENDIF
	*-- Grobritanien
	IF LOWER( tcCountry ) = 'gb'
		lcString = STRTRAN( lcString, '+44 ', '0044' )
		lcString = STRTRAN( lcString, '+44', '0044' )
	ENDIF
	*-- Dnemark
	IF LOWER( tcCountry ) = 'dk'
		lcString = STRTRAN( lcString, '+45 ', '0045' )
		lcString = STRTRAN( lcString, '+45', '0045' )
	ENDIF
	*-- Schweden
	IF LOWER( tcCountry ) = 'se'
		lcString = STRTRAN( lcString, '+46 ', '0046' )
		lcString = STRTRAN( lcString, '+46', '0046' )
	ENDIF
	*-- Norwegen
	IF LOWER( tcCountry ) = 'no'
		lcString = STRTRAN( lcString, '+47 ', '0047' )
		lcString = STRTRAN( lcString, '+47', '0047' )
		lcString = STRTRAN( lcString, '+047', '0047' )
	ENDIF
	*-- Polen
	IF LOWER( tcCountry ) = 'pl'
		lcString = STRTRAN( lcString, '+48 ', '0048' )
		lcString = STRTRAN( lcString, '+48', '0048' )
	ENDIF
	*-- Deutschland
	IF LOWER( tcCountry ) = 'de' .AND. LEFT( lcString, 2 ) = '49'
		lcString = STRTRAN( lcString, '49', '0' )
	ENDIF
	*-- Deutschland
	IF LOWER( tcCountry ) = 'de'
		lcString = STRTRAN( lcString, '+00490', '0' )
		lcString = STRTRAN( lcString, '+0049', '0' )
		lcString = STRTRAN( lcString, '+0490', '0' )
		lcString = STRTRAN( lcString, '+049', '0' )
		lcString = STRTRAN( lcString, '+49 ', '0' )
		lcString = STRTRAN( lcString, '+49', '0' )
		lcString = STRTRAN( lcString, '+', '0' )
	ENDIF

RETURN( lcString )


*----------------------------------------------------------------------------------------------------------------------------
*-- Eine PDF-Rechnungs-Kopie mittels SumatraPDF drucken.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION PrintCopyBill
**********************

	LPARAMETERS lcRechNr AS STRING

	LOCAL lcFileName AS STRING, lcBatchString AS STRING, lcBatchFile AS STRING, lcSafety AS STRING, lcReturn AS STRING

	*-- Den Namen und den Pfad zu der Rechnungs-Kopie setzen
	lcFileName = FULLPATH( '..\Rechnungskopie\00' + SUBSTR( lcRechNr, 1, 3 ) + '\' ) + 'RK' + lcRechNr + '.PDF'

	*-- Den Batch-Befehl setzen und in die Batchdatei schreiben
	lcBatchString = '@echo off' + CRLF + FULLPATH( './SumatraPDF/SumatraPDF.exe' ) + ' -print-to-default ' + lcFileName
	lcBatchString = STRCONV( lcBatchString, 5 )			&& Multi-Byte in Unicode umwandeln
	lcBatchString = STRCONV( lcBatchString, 10 )		&& Unicode in UTF-8 umwandeln
	lcBatchFile = 'SumatraPDF\SumatraPDF.bat'
	lcSafety = SET( 'Safety' )
	SET SAFETY OFF
	STRTOFILE( lcBatchString, lcBatchFile, 0 )
	SET SAFETY &lcSafety

	*-- Wenn die PDF-Datei und die Batchdatei vorhanden ist ...
	IF FileExist( lcFileName ) .AND. FileExist( lcBatchFile )

		*-- ... die erstellte Batch-Datei ausfhren
		ShellExecute( 0, 'open', lcBatchFile, '', '', 0 )

		*-- Wir prfen maximal 5000 ms, ob die SumatraPDF.exe noch luft
		FOR lnLoop = 1 TO 100

			*-- 50 ms warten
			Sleep(50)

			*-- Luft die SumatraPDF.exe noch?
			IF IsExeRunning( 'sumatrapdf.exe' ) = .F.

				*-- 200 ms warten
				Sleep(200)

				*-- HaPe fertig
				EXIT
			ENDIF
		NEXT

		*-- Alles OK
		lcReturn = ''

	*-- Andernfalls geben wir die Rechnungs-Nummer zurck
	ELSE
		lcReturn = lcRechNr + CR
	ENDIF

	*-- Rckgabe setzen
	RETURN( lcReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: PDF-Rechnung ausdrucken.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 28.10.2020
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION PrintPDF
*****************

	LPARAMETERS lcRechPath AS STRING

	LOCAL lcString AS STRING, lcAction AS STRING, lcFileName AS STRING

	*-- Zeichenkette fr Batchdatei aufbauen
	lcString = '@echo off' + CHR( 13 ) + CHR( 10 ) + FULLPATH( './SumatraPDF/SumatraPDF.exe' ) + ' -print-to-default ' + ;
		lcRechPath
	lcString = STRCONV( lcString, 5 )		&& convert a String from VFP multi-byte to Unicode
	lcString = STRCONV( lcString, 10 )		&& Convert from Unicode to UTF-8

	*-- Alte Batchdatei lschen und neue erstellen
	DELETE FILE '.\SumatraPDF\SumatraPDF.bat'
	STRTOFILE( lcString, './SumatraPDF/SumatraPDF.bat' )
	lcFileName = '.\SumatraPDF\SumatraPDF.bat'

	*-- Die PDF-Datei drucken
	IF FILE( lcFileName )
		lcAction = 'open'
		ShellExecute( 0, lcAction, lcFileName, '', '', 0 )
	ENDIF

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
*-- Formular-Position wieder herstellen.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION RestoreFormPosition
****************************

	LPARAMETERS toForm AS FORM

	LOCAL lcINIFile AS STRING, lcUser AS STRING, lcFormName AS STRING

	*-- Parameter prfen
	IF VARTYPE( gu_user ) == 'C'
		lcUser = ALLTRIM( gu_user )
	ELSE
		lcUser = 'HPG'
	ENDIF

	*-- Name des Formulars
	lcFormName = UPPER( toForm.NAME )

	*-- Name der INI-Datei
	lcINIFile = GETWORDNUM( gcINIFileName, 1, '.' ) + '_' + lcUser + '.ini'

	*-- Bei TOP-LEVEL-Formularen gibt es eine pcabhngige Sektion
	IF toForm.SHOWWINDOW = 2

		*-- Name der Sektion pcabhngig setzen
		lcFormName = lcFormName + '_' + gcMachineName
	ENDIF

	*-- Wir lesen die Fenster-Daten aus der INI-Datei
	toForm.TOP = VAL( ReadINI( lcFormName, 'Top', 60, Num2C( toForm.TOP ), lcINIFile ) )
	toForm.LEFT = VAL( ReadINI( lcFormName, 'Left', 60, Num2C( toForm.LEFT ), lcINIFile  ) )
	toForm.WIDTH = VAL( ReadINI( lcFormName, 'Width', 60, Num2C( toForm.WIDTH ), lcINIFile ) )
	toForm.HEIGHT = VAL( ReadINI( lcFormName, 'Height', 60, Num2C( toForm.HEIGHT ), lcINIFile ) )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
*-- Applikations-Fenster wieder herstellen.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION RestoreScreenPosition
******************************

	LOCAL lcINIFile AS STRING, lcUser AS STRING, lcConfigSection AS STRING
	LOCAL lnValue AS INTEGER

	*-- Parameter prfen
	IF VARTYPE( gu_user ) == 'C'
		lcUser = ALLTRIM( gu_user )
	ELSE
		lcUser = 'HPG'
	ENDIF

	*-- Sektion fr Hauptfenster bestimmen
	lcConfigSection = 'CONFIG_' + gcMachineName

	*-- Name der INI-Datei
	lcINIFile = GETWORDNUM( gcINIFileName, 1, '.' ) + '_' + lcUser + '.ini'

	*-- Wir lesen die Fenster-Daten aus der INI-Datei
	*-- Prfen ob die pcabhngige Sektion vorhanden ist
	lnValue = VAL( ReadINI( lcConfigSection, 'Width', 60, '0', lcINIFile ) )
	IF lnValue > 0
		_VFP.TOP = VAL( ReadINI( lcConfigSection, 'Top', 60, '0', lcINIFile ) )
		_VFP.LEFT = VAL( ReadINI( lcConfigSection, 'Left', 60, '0', lcINIFile  ) )
		_VFP.WIDTH = lnValue
		_VFP.HEIGHT = VAL( ReadINI( lcConfigSection, 'Height', 60, '800', lcINIFile ) )
	ELSE
		_VFP.TOP = VAL( ReadINI( 'CONFIG', 'Top', 60, '0', lcINIFile ) )
		_VFP.LEFT = VAL( ReadINI( 'CONFIG', 'Left', 60, '0', lcINIFile  ) )
		_VFP.WIDTH = VAL( ReadINI( 'CONFIG', 'Width', 60, '1400', lcINIFile ) )
		_VFP.HEIGHT = VAL( ReadINI( 'CONFIG', 'Height', 60, '800', lcINIFile ) )
	ENDIF

	*-- Und Applikation anzeigen
	_VFP.VISIBLE = .T.

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
*-- Position des Formulars sichern.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION SaveFormPosition
*************************

	LPARAMETERS toForm AS FORM

	LOCAL lcINIFile AS STRING, lcUser AS STRING, lcFormName AS STRING

	*-- Parameter prfen
	IF VARTYPE( gu_user ) == 'C'
		lcUser = ALLTRIM( gu_user )
	ELSE
		lcUser = 'HPG'
	ENDIF

	*-- Name des Formulars
	lcFormName = UPPER( toForm.NAME )

	*-- Name der INI-Datei
	lcINIFile = GETWORDNUM( gcINIFileName, 1, '.' ) + '_' + lcUser + '.ini'

	*-- Bei TOP-LEVEL-Formularen gibt es eine pcabhngige Sektion
	IF toForm.SHOWWINDOW = 2

		*-- Name der Sektion pcabhngig setzen
		lcFormName = lcFormName + '_' + gcMachineName
	ENDIF

	*-- Wir schreiben die Fenster-Daten in der INI-Datei
	WriteINI( lcFormName, 'Top', Num2C( toForm.TOP ), lcINIFile )
	WriteINI( lcFormName, 'Left', Num2C( toForm.LEFT ), lcINIFile )
	WriteINI( lcFormName, 'Width', Num2C( toForm.WIDTH ), lcINIFile )
	WriteINI( lcFormName, 'Height', Num2C( toForm.HEIGHT ), lcINIFile )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
*-- Position des Applikations-Fenster sichern.
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION SaveScreenPosition
***************************

	LOCAL lcINIFile AS STRING, lcUser AS STRING, lcConfigSection AS STRING

	*-- Parameter prfen
	IF VARTYPE( gu_user ) == 'C'
		lcUser = ALLTRIM( gu_user )
	ELSE
		lcUser = 'HPG'
	ENDIF

	*-- Sektion fr Hauptfenster bestimmen
	lcConfigSection = 'CONFIG_' + gcMachineName

	*-- Name der INI-Datei
	lcINIFile = GETWORDNUM( gcINIFileName, 1, '.' ) + '_' + lcUser + '.ini'

	*-- Wir schreiben die Fenster-Daten in die INI-Datei
	WriteINI( lcConfigSection, 'Top', Num2C( _VFP.TOP ), lcINIFile )
	WriteINI( lcConfigSection, 'Left', Num2C( _VFP.LEFT ), lcINIFile )
	WriteINI( lcConfigSection, 'Width', Num2C( _VFP.WIDTH ), lcINIFile )
	WriteINI( lcConfigSection, 'Height', Num2C( _VFP.HEIGHT ), lcINIFile )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Gibt das Datum im Format DateTime zurck.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 19.08.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION TimeStamp2DateTime( tnTimeStamp AS INTEGER )
*****************************************************

	LOCAL ltDateTime AS DATETIME

	*-- Parameter prfen
	IF VARTYPE( tnTimeStamp) <> 'N'
		tnTimeStamp = 0
	ENDIF

	*-- Datum in Format DateTime umwandeln
	*-- und zurckgeben
	ltDateTime = {01.01.1970 00:00:00} + tnTimeStamp
	RETURN( ltDateTime )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Schreibt Logging-Infos auf die Festplatte.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 15.12.2022
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION WriteError
*******************

	LPARAMETERS tcText AS STRING

	LOCAL lcDateTime AS STRING, lcMessage AS STRING

	*-- Aktuelle Uhrzeit holen
	lcDateTime = TTOC( DATETIME() )
	lcDateTimeFile = gu_User + '_' + TTOC( DATETIME(), 1 )

	*-- Meldung aufbauen
	lcMessage = lcDateTime + ' # ' + tcText + CRLF

	*-- Logging-Info wegschreiben
	STRTOFILE( lcMessage, 'Error\KDV_Error_' + lcDateTimeFile + '.log', 1 )

ENDFUNC


*############################################################################################################################
* RestAPI-Funktionen
*############################################################################################################################

*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Artikeldaten in einen Cursor holen und daraus den JSON-String erstellen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 25.11.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Article2JSON( tcKdNr AS STRING, tcRechNr AS STRING, tcArtNr AS STRING )
********************************************************************************

	LOCAL lcJSON AS STRING
	LOCAL loJSON AS 'wwJsonSerializer'

	*-- Die JSON-Klasse laden und erstellen
	DO wwJsonSerializer
	loJSON = CREATEOBJECT( 'wwJsonSerializer' )

	*-- Mit den bergebenen Daten einen CURSOR aufbauen
	*-- fllen und den Ergebnis-Cursor selektieren
	CREATE CURSOR cursArticle ( KdNr C(6), RechNr C(6), ArtNr C(6), Auftrag C(6) )
	INSERT INTO cursArticle( KdNr, RechNr, ArtNr, Auftrag ) VALUES ( ALLTRIM( tcKdNr ), ALLTRIM( tcRechNr ), ;
		ALLTRIM( tcArtNr ), '1####1' )
	SELECT cursArticle

	*-- Den CURSOR ins JSON-Format wandeln
	*-- Tabellen und Cursor
	lcJSON = loJSON.Serialize( 'cursor:cursArticle' )
	USE IN cursArticle

	*-- Die eckigen Klammern vorne und hinten
	*-- entfernen, falls vorhanden
	IF LEFT( lcJSON, 1 ) == '['
		lcJSON = StrDelL( StrDelR( lcJSON ) )
	ENDIF

	*-- ANSI in UTF-8 umwandeln
	lcJSON = STRCONV( lcJSON, 5 )			&& Multi-Byte in Unicode umwandeln
	lcJSON = STRCONV( lcJSON, 10 )			&& Unicode in UTF-8 umwandeln

#IFDEF  RESTAPI_WRITELOG
	*-- 25.11.24 HPG: Zum Debuggen
	*-- Und wegschreiben in eine JSON-Datei
	TRY
		SET SAFETY OFF
		STRTOFILE( lcJson, 'C:\tmp\Article.JSON' )
		SET SAFETY ON
	CATCH
	ENDTRY
#ENDIF

	*-- Den JSON-String zurckgeben
	RETURN( lcJSON )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Den WebShop-Ansprechpartner anhand der ID auf Vorhandensein prfen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 14.09.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckWSContactbyID( tnID_WS AS INTEGER )
*************************************************

	LOCAL llReturn AS BOOLEAN

	*-- Ist diese WebShop-Ansprechpartner vorhanden?
	lcJSON = HandleWebShopData( '', tnID_WS, 'G', 'CO', .T. )
	IF LEFT( lcJSON, 5 ) <> 'ERROR'
		llReturn = .T.
	ELSE
		llReturn = .F.
	ENDIF

	*-- Rckgabe des Kennzeichens
	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die WebShop-Adresse anhand der ID auf Vorhandensein prfen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 14.09.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CheckWSCustomerbyID( tnID_WS AS INTEGER )
**************************************************

	LOCAL llReturn AS BOOLEAN

	*-- Ist diese WebShop-Adresse vorhanden?
	lcJSON = HandleWebShopData( '', tnID_WS, 'G', 'CU', .T. )
	IF LEFT( lcJSON, 5 ) <> 'ERROR'
		llReturn = .T.
	ELSE
		llReturn = .F.
	ENDIF

	*-- Rckgabe des Kennzeichens
	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Ansprechpartnerdaten in einen Cursor holen und daraus den JSON-String erstellen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 11.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 16.08.23 HPG: Die Felder Vorwahl und Zusatz wurden entfernt.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Contact2JSON( tnID AS INTEGER )
****************************************

	LOCAL lcJSON AS STRING
	LOCAL loJSON AS 'wwJsonSerializer'

	*-- Ist die SELECT-Zeichenkette schon deklariert?
	IF VARTYPE( gcContact ) <> 'C'

		*-- Nein, dann dies durchfhren
		CreateWebShopArrays()
	ENDIF

	*-- Die JSON-Klasse laden und erstellen
	DO wwJsonSerializer
	loJSON = CREATEOBJECT( 'wwJsonSerializer' )

	*-- Die gesuchte ID in den SELECT einfgen
	gcContactSelect = STRTRAN( gcContactSelect, '#', Num2C( tnID ) )

	*-- Einen Ansprechpartner holen
	EXECSCRIPT( gcContactSelect )

	*-- Den CURSOR ins JSON-Format wandeln
	*-- Tabellen und Cursor
	lcJson = loJSON.Serialize( 'cursor:cursContact' )
	USE IN cursContact

	*-- ANSI in UTF-8 umwandeln
	lcJson = STRCONV( lcJson, 5 )			&& Multi-Byte in Unicode umwandeln
	lcJson = STRCONV( lcJson, 10 )			&& Unicode in UTF-8 umwandeln

#IFDEF  RESTAPI_WRITELOG
	*-- 05.07.23 HPG: Zum Debuggen
	*-- Und wegschreiben in eine JSON-Datei
	TRY
		SET SAFETY OFF
		STRTOFILE( lcJson, 'C:\tmp\Contact.JSON' )
		SET SAFETY ON
	CATCH
	ENDTRY
#ENDIF

	*-- Den JSON-String zurckgeben
	RETURN( lcJSON )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Fllt anhand des ID-Cursors cursID_WS die Import-Tabelle.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 25.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CreateImportRecords( tcMode AS STRING, tcImportFile AS STRING, toForm AS FORM )
****************************************************************************************

	LOCAL lcJSON AS STRING, lcRestAPIErrorPost AS STRING, lcRestAPIErrorPre AS STRING
	LOCAL lnID_WS AS INTEGER, lnID AS INTEGER
	LOCAL llReturn AS BOOLEAN, llWriteLog AS BOOLEAN, llErrorOccured AS BOOLEAN

	*-- Parameter setzen
	llErrorOccured = .F.

	*-- Parameter prfen
	llWriteLog = ( VARTYPE( toForm ) == 'O' )

	*-- RestAPI-Fehlerlog einlesen
	TRY
		lcRestAPIErrorPre = FILETOSTR( 'RestAPIError.log' )
	CATCH
		lcRestAPIErrorPre = ''
	ENDTRY

	*-- Schleife ber den Cursor mit den WebShop-IDs zum
	*-- Einlesen der Adressen/Ansprechpartner-Daten
	SELECT cursID_WS
	SCAN

		*-- ID holen
		lnID_WS = ID_WS

		*-- Adresse/Ansprechpartner aus dem WebShop als
		*-- JSON-String holen
		lcJSON = HandleWebShopData( .F., lnID_WS, 'G', tcMode, .T. )

		*-- Wir schreiben ein Log
		IF llWriteLog
			toForm.WriteLog( 'LOG', 'Daten von ID >' + Num2C( lnID_WS ) + '< vom Webshop holen' + CR )
		ENDIF

		*-- Je nach Modus die Daten in die Import-Tabelle schreiben
		DO CASE

			*-- Kunden-Adresse
			CASE tcMode == 'CU'

				*-- Diesen JSON-String in die Import-Tabelle schreiben
				JSON2CustomerImport( lcJSON, tcImportFile )

			*-- Ansprechpartner-Adresse
			CASE tcMode == 'CO'

				*-- Diesen JSON-String in die Import-Tabelle schreiben
				JSON2ContactImport( lcJSON, tcImportFile )
		ENDCASE
	ENDSCAN

	*-- Schleife ber den Cursor mit den nderungs-IDs
	*-- zum Lschen der zugehrigen nderungen.
	SELECT cursID_WSAll
	SCAN

		*-- ID holen
		lnID = ID_CHG

		*-- nderungs-Eintrag lschen
		lcJSON = HandleWebShopData( .F., lnID, 'D', 'SH', .T. )

		*-- Wir schreiben ein Log
		IF llWriteLog
			toForm.WriteLog( 'LOG', 'nderungseintrag von ID >' + Num2C( lnID ) + '< lschen' + CR )
		ENDIF
	ENDSCAN

	*-- Prfen ob RestAPI-Fehler aufgetreten sind
	*-- RestAPI-Fehlerlog einlesen
	TRY
		lcRestAPIErrorPost = FILETOSTR( 'RestAPIError.log' )
	CATCH
		lcRestAPIErrorPost = ''
	ENDTRY

	*-- Es gab Fehlereintrge
	IF LEN( lcRestAPIErrorPost ) > LEN( lcRestAPIErrorPre )

		*-- Fehlermeldung ausgeben
		MESSAGEBOX( 'Es sind RestAPI-Fehler aufgetreten; bitte wenden Sie sich an die IT.', 48+0+0, 'RestAPI-Fehler' )
	ENDIF

	*-- Temporre Cursor schlieen
	USE IN cursID_WS
	USE IN cursID_WSAll

	*-- Rckgabe
	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Erstellt fr einen WebView-Artikel den JSON-String fr den Rest-API-Request.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 26.11.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION ArticleOrder2JSON( tcRechNr AS STRING, tcArtNr AS STRING, tdDate AS DATE )
***********************************************************************************

	LOCAL lcBillPDF AS STRING, lcCopyBillPDF AS STRING, lcJSON AS STRING
	LOCAL lnReturn AS INTEGER, lnASCII AS INTEGER
	LOCAL llReturn AS BOOLEAN

	*-- Pfad und Name der Rechnung und der Rechnungs-Kopie
	*-- 26.11.24 HPG: Wichtig: Wir mssen das Datum und nicht
	*-- das Leistungsdatum (LeistDatum) verwenden.
	lcBillPDF = GetBillPath( tdDate ) + 'R' + tcRechNr + '.pdf'
	lcCopyBillPDF = GetCopyBillPath( tcRechNr ) + 'RK' + tcRechNr + '.pdf'

	*-- Die Daten fr den WebView-Artikel zusamenstellen
	SELECT Kun.KdNr, Kun.SuchBeg, Kun.Anrede, Kun.Titel, Kun.Name1, Kun.Name2, Kun.Name3, Kun.PLZ, Kun.Ort, Kun.Strasse, ;
		Kun.Land_ID, Kun.Vorwahl, Kun.Tel, Kun.Tel2, Kun.Mobil, Kun.Fax, Kun.Email, Kun.Email_Rech, Kun.USStIDNr, ;
		Auf.Jahr, Auf.RechNr, Auf.Zahl_Ziel, Auf.ArtNr, Auf.Menge, Auf.KurzText, Auf.MWSt, Auf.Preis, Auf.Notiz, Auf.TEXT, ;
		Auf.PROG, Auf.Datum, Auf.Zeit, Auf.Transfer, Auf.Erledigt, Auf.BH_Konto, Auf.Storno, Auf.STransfer, Auf.SDatum, ;
		Auf.SModi_User, Auf.SBS_Rech, Auf.SBS_Storno, Auf.Storno_Nr, Auf.LeistDatum, Auf.Medium, Auf.Bez_Datum, Auf.MahnStufe, ;
		Auf.St_WebSper, Art.ArtGruppe, Art.LangText, Art.Preis1, Art.Preis2, Art.Preis3, Art.Prog_Para, Art.BH_KostSt, ;
		Art.BH_KostTr, Art.BKB_EDV, Art.Lizenz_Proz, Art.Lizenz_Bundle, Art.Lizenz_Flag, Art.Lizenz_Abzug, Art.Versand, ;
		Art.Verpackung, Art.ISBN, CAST( Art.Medium AS C(10) ) AS ArtMedium, Rec.Adresse, CAST( Rec.Strasse AS C(50) ) AS RecStrasse, ;
		CAST( Rec.PLZ AS C(5) ) AS RecPLZ, CAST( Rec.Ort AS C(50) ) AS RecOrt, Rec.BestellNr, Rec.Lieferschein, ;
		CAST( Rec.Land_ID AS I(4) ) AS RecLand_ID, Rec.ZahlArt, Rec.ASP_ID, Rec.VerpackGew, Fre.Anwender, Fre.Freischalt, ;
		CAST( Fre.ArtNr AS C(6) ) AS FreiArtNr, Fre.Datum AS FreiDatum, Fre.Rech_Datum, CAST( Fre.Email AS C(100) ) AS FreiEmail, ;
		Fre.VersandArt, MwS.Satz, CAST( lcBillPDF AS C(150) ) AS RechPDF, CAST( lcCopyBillPDF AS C(150) ) AS RechKopPDF ;
		FROM Bkb_Kund Kun ;
		LEFT JOIN Bkb_Auft Auf ON Kun.KdNr = Auf.KdNr ;
		LEFT JOIN Bki_MwSt MwS ON MwS.ID = Auf.MWSt ;
		LEFT JOIN Bki_Rech Rec ON Rec.RechNr = Auf.RechNr ;
		LEFT JOIN Bkb_Art Art ON Art.ArtNr = Auf.ArtNr ;
		LEFT JOIN Bki_Frei Fre ON Fre.Rech_Nr = Auf.RechNr ;
		WHERE Auf.ArtNr = tcArtNr .AND. Auf.RechNr = tcRechNr INTO CURSOR cursJSON4Article READWRITE

	*-- Den Cursor in einen JSON-String umwandeln
	lcJSON = Cursor2JSON( 'cursJSON4Article', .T. )
	FOR lnASCII = 0 TO 8
		lcJSON = STRTRAN( lcJSON, CHR( lnASCII ), '' )
	NEXT

	*-- Die eckigen Klammern vorne und hinten
	*-- entfernen, falls vorhanden
	IF LEFT( lcJSON, 1 ) == '['
		lcJSON = StrDelL( StrDelR( lcJSON ) )
	ENDIF

	*-- ANSI in UTF-8 umwandeln
	lcJSON = STRCONV( lcJSON, 5 )			&& Multi-Byte in Unicode umwandeln
	lcJSON = STRCONV( lcJSON, 10 )			&& Unicode in UTF-8 umwandeln

	*-- Rckgabe des JSON-Strings
	RETURN( lcJSON )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die beiden Arrays fr den WebShop-Import erstellen und fllen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 10.08.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 10.08.23 HPG: In der ersten Spalte ist der Namen der Felder die zu bertragen sind.
*				In der zweiten Spalte knnen stehen fr die Umwandlung beim Download, also JSON zu DBF:
*				 - FU: Umwandlungs-Funktion fr den Quell-Feldnamen, zb. bei anrede
*				 - FN: Ein Ziel-Feldname, zb. bei name und vorname beim Ansprechpartner.
* 20.12.23 HPG: In der dritten Spalte knnen stehen fr die Umwandlung beim Upload, also DBF zu JSON:
*				 - FU: Umwandlungs-Funktion fr den Quell-Feldnamen, zb. bei anrede
*				 - FN: Ein Ziel-Feldname, zb. bei name und vorname beim Ansprechpartner.
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 19.12.23 HPG: Feld passwort fr den Kunden hinzugefgt.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CreateWebShopArrays
****************************

	PUBLIC gcCustomerFields
	PUBLIC gcContactFields
	PUBLIC gcCustomerSelect
	PUBLIC gcContactSelect

	*-- Array fr die Kunden
*-- Wenn definiert, wird das WebPasswort und die EMail-
*-- Anmelde-Adresse hochgeladen, andernfalls nicht.
#IFDEF  UPLOAD_PASSWORD
	DECLARE gcCustomerFields[26, 3]
#ELSE
	DECLARE gcCustomerFields[24, 3]
#ENDIF
	gcCustomerFields[1, 1]  = 'id'
	gcCustomerFields[2, 1]  = 'modi_date'
	gcCustomerFields[3, 1]  = 'modi_user'
	gcCustomerFields[4, 1]  = 'kdnr'
	gcCustomerFields[5, 1]  = 'merkmal'
	gcCustomerFields[6, 1]  = 'suchbeg'
	gcCustomerFields[7, 1]  = 'kdart'
	gcCustomerFields[8, 1]  = 'anrede'
	gcCustomerFields[8, 2]  = 'FU:GetSalutation4ID4WS'
	gcCustomerFields[8, 3]  = 'FU:MapKDVSalutation2WS'
	gcCustomerFields[9, 1]  = 'titel'
	gcCustomerFields[10, 1] = 'name1'
	gcCustomerFields[11, 1] = 'name2'
	gcCustomerFields[12, 1] = 'plz'
	gcCustomerFields[13, 1] = 'ort'
	gcCustomerFields[14, 1] = 'strasse'
	gcCustomerFields[15, 1] = 'land'
	gcCustomerFields[16, 1] = 'land_id'
	gcCustomerFields[17, 1] = 'tel'
	gcCustomerFields[18, 1] = 'tel2'
	gcCustomerFields[19, 1] = 'mobil'
	gcCustomerFields[20, 1] = 'fax'
	gcCustomerFields[21, 1] = 'email'
	gcCustomerFields[22, 1] = 'email_rech'
	gcCustomerFields[23, 1] = 'rech_email'
	gcCustomerFields[24, 1] = 'usstidnr'

*-- Wenn definiert, wird das WebPasswort und die EMail-
*-- Anmelde-Adresse hochgeladen, andernfalls nicht.
#IFDEF  UPLOAD_PASSWORD
	gcCustomerFields[23, 1] = 'email_anme'
	gcCustomerFields[24, 1] = 'rech_email'
	gcCustomerFields[25, 1] = 'usstidnr'
	gcCustomerFields[26, 1] = 'passwort'
#ENDIF

	*-- Array fr die Ansprechpartner
	DECLARE gcContactFields[24, 3]
	gcContactFields[1, 1]  = 'id'
	gcContactFields[2, 1]  = 'modi_date'
	gcContactFields[3, 1]  = 'modi_user'
	gcContactFields[4, 1]  = 'kdnr'
	gcContactFields[5, 1]  = 'id_kund'
	gcContactFields[6, 1]  = 'asp_aspnr'
	gcContactFields[7, 1]  = 'anrede'
	gcContactFields[7, 2]  = 'FU:GetSalutation4ID4WS'
	gcContactFields[7, 3]  = 'FU:MapKDVSalutation2WS'
	gcContactFields[8, 1]  = 'titel'
	gcContactFields[9, 1]  = 'funktion'
	gcContactFields[10, 1] = 'name'
	gcContactFields[10, 2] = 'FN:name1'
	gcContactFields[11, 1] = 'vorname'
	gcContactFields[11, 2] = 'FN:name2'
	gcContactFields[12, 1] = 'hauspostco'
	gcContactFields[13, 1] = 'plz'
	gcContactFields[14, 1] = 'ort'
	gcContactFields[15, 1] = 'strasse'
	gcContactFields[16, 1] = 'land'
	gcContactFields[17, 1] = 'id_land'
	gcContactFields[18, 1] = 'tel'
	gcContactFields[19, 1] = 'tel2'
	gcContactFields[20, 1] = 'mobil'
	gcContactFields[21, 1] = 'fax'
	gcContactFields[22, 1] = 'email'
	gcContactFields[23, 1] = 'aspinaktiv'
	gcContactFields[24, 1] = 'zusatz'

	*-- Die SELECT-Zeichenkette fr den Kunden erstellen
	gcCustomerSelect = 'SELECT '
	lnFieldCount = ALEN( gcCustomerFields, 1 )

	*-- Schleife ber alle Felder
	FOR lnField = 1 TO lnFieldCount

		*-- Feldnamen aus dem Array holen und anhngen
		lcField = ALLTRIM( gcCustomerFields[lnField, 1] ) + ', '
		gcCustomerSelect = gcCustomerSelect + lcField
	NEXT

	*-- Das letzte Komma lschen
	gcCustomerSelect = StrDelR( gcCustomerSelect, 2 ) + ' '

	*-- Die SELECT-Zeichenkette fr den Kunden vervollstndigen
	gcCustomerSelect = gcCustomerSelect + 'FROM BKB_KUND WHERE KdNr = "#" INTO CURSOR cursCustomer READWRITE'

	*-- Die SELECT-Zeichenkette fr den Ansprechpartner erstellen
	gcContactSelect = 'SELECT '
	lnFieldCount = ALEN( gcContactFields, 1 )

	*-- Schleife ber alle Felder
	FOR lnField = 1 TO lnFieldCount

		*-- Feldnamen aus dem Array holen und anhngen
		lcField = ALLTRIM( gcContactFields[lnField, 1] ) + ', '
		gcContactSelect = gcContactSelect + lcField
	NEXT

	*-- Das letzte Komma lschen
	gcContactSelect = StrDelR( gcContactSelect, 2 ) + ' '

	*-- Die SELECT-Zeichenkette fr den Ansprechpartner vervollstndigen
	gcContactSelect = gcContactSelect + 'FROM BKI_ANSPRECH WHERE ID = # INTO CURSOR cursContact READWRITE'

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Einen Foxpro-CURSOR in einen JSON-String umwandeln.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 26.11.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Cursor2JSON( tcCursorName AS STRING, tlCloseCursor AS BOOLEAN )
************************************************************************

	LOCAL lcJSON AS STRING
	LOCAL loJSON AS 'wwJsonSerializer'

	*-- Die JSON-Klasse laden und erstellen
	DO wwJsonSerializer
	loJSON = CREATEOBJECT( 'wwJsonSerializer' )

	*-- Den CURSOR ins JSON-Format wandeln
	*-- Tabellen und Cursor
	lcJSON = loJSON.Serialize( 'cursor:' + tcCursorName )

	*-- Wenn der Cursor geschlossen werden soll
	IF tlCloseCursor
		USE IN ( tcCursorName )
	ENDIF

	*-- ANSI in UTF-8 umwandeln
	lcJSON = STRCONV( lcJSON, 5 )			&& Multi-Byte in Unicode umwandeln
	lcJSON = STRCONV( lcJSON, 10 )			&& Unicode in UTF-8 umwandeln

	*-- Und wegschreiben
	RETURN( lcJSON )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Kundendaten in einen Cursor holen und daraus den JSON-String erstellen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 05.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 16.08.23 HPG: Die Felder Name3, Ansprech und Vorwahl wurden entfernt.
* 19.12.23 HPG: Wir mssen das Kennwort-Feld faken: aus "passwort": wird "@pass": in der JSON-Datei.
* 20.12.23 HPG: Die Anrede-ID mssen wir ummappen: dafr wird beim Anrede-Feld die neue Funktion MapKDVSalutation2WS ver-
*				wendet.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION Customer2JSON( tcKdNr AS STRING )
*****************************************

	LOCAL lcJSON AS STRING, lcCustomerSelect AS STRING, lcDataType AS STRING, lcField AS STRING, lcFunction AS STRING
	LOCAL lcTargetField AS STRING, lcThirdElement AS STRING, lcType AS STRING, lcAnrede AS STRING, lcNeuerTitel AS STRING
	LOCAL lcTitel AS STRING
	LOCAL lnLoop AS INTEGER, lnMaxLoop AS INTEGER, lnValue AS INTEGER, lnID AS INTEGER
	LOCAL luThirdElement AS VARIANT, luValue AS VARIANT
	LOCAL loJSON AS 'wwJsonSerializer'

	*-- Ist die SELECT-Zeichenkette schon deklariert?
	IF VARTYPE( gcCustomerSelect ) <> 'C'

		*-- Nein, dann dies durchfhren
		CreateWebShopArrays()
	ENDIF

	*-- Die JSON-Klasse laden und erstellen
	DO wwJsonSerializer
	loJSON = CREATEOBJECT( 'wwJsonSerializer' )

	*-- Die gesuchte Kunden-Nummer in den SELECT einfgen
	*-- Dazu den Musterselect bernehmen und erst in der
	*-- Kopie die Ersetzung vornehmen.
	lcCustomerSelect = gcCustomerSelect
	lcCustomerSelect = STRTRAN( lcCustomerSelect, '#', ALLTRIM( tcKdNr ) )

	*-- Eine Adresse holen
	EXECSCRIPT( lcCustomerSelect )

	*-- Den Ergebnis-Cursor selektieren
	SELECT cursCustomer

	*-- 20.12.23 HPG: Feste Funktion fr das Mappen von
	*-- Anrede in Anrede und Titel.
	*-- Muss VOR der nachstehende Schleife durchgefhrt
	*-- werden, weil dort die Anrede-ID schon umgesetzt
	*-- wird. Hier bentigen wir noch die Anrede-ID der
	*-- KDV und nicht vom WebShop.
	*-- Sind die Felder Anrede und Titel vorhanden?
	IF LEN( FIELD( 'ANREDE' ) ) > 0 .AND. LEN( FIELD( 'TITEL' ) ) > 0

		*-- Inhalte der Felder holen
		lnID = Anrede
		lcTitel = ALLTRIM( Titel )

		*-- Nur wenn das Titel-Feld leer ist und die ID
		*-- grer als 2, mssen wir was tun.
		IF LEN( lcTitel ) = 0 .AND. lnID > 2

			*-- Den Anrede-Text holen und den neuer Standard-
			*-- Titel setzen
			*lcAnrede = GetSalutation4ID( lnID, 'A' )
			lcNeuerTitel = ''

			*-- Ist im Anrede-Text ein Titel enthalten?
			DO CASE

				*-- Der DR und der PROF steht schon in der Anrede ...
				CASE 'DR' $ UPPER( lcAnrede ) .AND. 'PROF' $ UPPER( lcAnrede )

					*-- Bei ID = 10
					IF lnID = 10
						lcNeuerTitel = 'Prof. Dr. Dr.'
					ELSE
						lcNeuerTitel = 'Prof. Dr.'
					ENDIF

				*-- Der DR steht schon in der Anrede ...
				CASE 'DR' $ UPPER( lcAnrede )
					lcNeuerTitel = 'Dr.'

				*-- Der Prof steht schon in der Anrede
				CASE 'PROF' $ UPPER( lcAnrede )
					lcNeuerTitel = 'Prof.'
			ENDCASE

			*-- Den neuen Titel setzen
			REPLACE Titel WITH lcNeuerTitel
		ENDIF
	ENDIF

	*-- Schleife ber alle Felder zum Prfen auf
	*-- eventuelle Funktionen bzw. Feld-Umsetzungen
	lnMaxLoop = ALEN( gcCustomerFields, 1 )

	*-- Schleife ber alle Felder des Arrays
	FOR lnLoop = 1 TO lnMaxLoop

		*-- Parameter setzen
		lcFunction = ''
		luThirdElement = .F.

		*-- Das Feld, eine eventuelle Konvertierfunktion
		*-- bzw. ein Ziel-Feldname holen
		lcField = ALLTRIM( gcCustomerFields[lnLoop, 1] )
		luThirdElement = gcCustomerFields[lnLoop, 3]

		*-- Gibt es einen Inhalt im dritten Array-Element?
		IF VARTYPE( luThirdElement ) == 'C'

			*-- Ja, dann die Art und den Inhalt bestimmen
			lcType = LEFT( luThirdElement, 3 )
			lcThirdElement = SUBSTR( luThirdElement, 4 )

			*-- Je nach Art verzweigen
			DO CASE

				*-- Wir haben eine Funktion
				CASE lcType == 'FU:'

					*-- Die Funktion holen
					lcFunction = lcThirdElement

					*-- Den Wert und damit den Datentyp bestimmen
					luValue = EVALUATE( lcField )
					lcDataType = VARTYPE( luValue )

					*-- Je nach Datentyp verzweigen
					DO CASE

						*-- Wir haben ein numerisches oder ein Whrungsfeld
						CASE lcDataType $ 'NY'
							
							*-- Funktion ausfhren und den Ergebnis-Wert
							*-- wieder in das Tabellen-feld zurckschreiben
							lnValue = &lcFunction( luValue )
							REPLACE &lcField WITH lnValue
					ENDCASE

				*-- Wir haben einen Ziel-Feldnamen
				CASE lcType == 'FN:'
					lcTargetField = lcThirdElement
			ENDCASE
		ENDIF
	ENDFOR

	*-- 14.09.23 HPG: Zum Hochladen von Adressen ohne
	*-- Kundennummer, diese lschen wenn negativ
*!*		SELECT cursCustomer
*!*		SET SAFETY OFF
*!*		COPY TO 'C:\TMP\' + tcKdNr
*!*		STRTOFILE( lcCustomerSelect, 'C:\TMP\' + tcKdNr + '.txt' )
*!*		SET SAFETY ON
*!*		GOTO TOP
*!*		IF VAL( cursCustomer.KdNr ) < 0
*!*			REPLACE cursCustomer.KdNr WITH ''
*!*		ENDIF
*!*		*-- Keine mehrfachen Stze zulassen
*!*		DELETE FOR RECNO() > 1

	*-- Den CURSOR ins JSON-Format wandeln
	*-- Tabellen und Cursor
	lcJson = loJSON.Serialize( 'cursor:cursCustomer' )
	USE IN cursCustomer

	*-- ANSI in UTF-8 umwandeln
	lcJson = STRCONV( lcJson, 5 )			&& Multi-Byte in Unicode umwandeln
	lcJson = STRCONV( lcJson, 10 )			&& Unicode in UTF-8 umwandeln

	*-- 19.12.23 HPG: Das Kennwort-Feld faken: aus "passwort": wird "@pass":
	*-- Wenn definiert, wird das WebPasswort hochgeladen, andernfalls nicht.
#IFDEF  UPLOAD_PASSWORD
	lcJson = STRTRAN( lcJson, '"passwort":', '"@pass":' )
#ENDIF

#IFDEF  RESTAPI_WRITELOG
	*-- 05.07.23 HPG: Zum Debuggen
	*-- Und wegschreiben in eine JSON-Datei
	TRY
		SET SAFETY OFF
		STRTOFILE( lcJson, 'C:\tmp\Customer.JSON' )
		SET SAFETY ON
	CATCH
	ENDTRY
#ENDIF

	*-- Den JSON-String zurckgeben
	RETURN( lcJSON )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Pflicht-Kundendaten in einen Cursor holen und daraus den JSON-String erstellen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 14.09.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 14.09.23 HPG: Die Pflichfelder sind salutation, name1, address1, zip, city und phone.
*				bzw. Anrede, Name1, PLZ, Ort, Strasse und Tel.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION CustomerChange2JSON( tcKdNr AS STRING )
************************************************

	LOCAL lcJSON AS STRING, lcAddress AS STRING, lcCity AS STRING, lcID_WS AS STRING, lcName1 AS STRING, lcPhone AS STRING
	LOCAL lcSalutation AS STRING, lcZIP AS STRING

	*-- Eine Adresse holen
	*-- Die Feldliste muss in Anzahl und Reihenfolge
	*-- dem Array gcCustomerFields entsprechen.
	SELECT ID_WS, ;
		ANREDE, ;
		NAME1, ;
		PLZ, ;
		ORT, ;
		STRASSE, ;
		TEL ;
		FROM BKB_KUND WHERE KdNr = tcKdNr INTO CURSOR cursCustomer READWRITE

	*-- Die Daten in die JSON-Zeichenkette eintragen
	lcID_WS = Num2C( cursCustomer.ID_WS )
	lcSalutation = IIF( cursCustomer.ANREDE > 0, Num2C( cursCustomer.ANREDE ), '' )
	lcName1 = ALLTRIM( cursCustomer.NAME1 )
	lcAddress = ALLTRIM( cursCustomer.STRASSE )
	lcZIP = ALLTRIM( cursCustomer.PLZ )
	lcCity = ALLTRIM( cursCustomer.ORT )
	lcPhone = ALLTRIM( cursCustomer.TEL )
	lcJson = '{"shop_id":' + lcID_WS + ',"action":"update_billing_address","address":{"salutation":"' + lcSalutation + ;
		'","name1":"' + lcName1 + '","address1":"' + lcAddress + '","zip":"' + lcZIP + '","city":"' + lcCity + ;
		'","phone":"' + lcPhone + '"}}'

	*-- ANSI in UTF-8 umwandeln
	lcJson = STRCONV( lcJson, 5 )			&& Multi-Byte in Unicode umwandeln
	lcJson = STRCONV( lcJson, 10 )			&& Unicode in UTF-8 umwandeln

#IFDEF  RESTAPI_WRITELOG
	*-- 05.07.23 HPG: Zum Debuggen
	*-- Und wegschreiben in eine JSON-Datei
	TRY
		SET SAFETY OFF
		STRTOFILE( tcKdNr + '#' + lcJson + CR, 'C:\tmp\CustomerChange.JSON', 1 )
		SET SAFETY ON
	CATCH
	ENDTRY
#ENDIF

	*-- Den JSON-String zurckgeben
	RETURN( lcJSON )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Bringt die WebShop-Adresse bzw. einen WebShop-Ansprechpartner per RestAPI auf den neuesten Stand.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 05.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 05.07.23 HPG: Modus fr die Adressen-/Ansprechpartner-Abfrage/Update/Delete:
*				 - 'L' = Liste der Adressen/Ansprechpartner/nderungen/Produkte
*				 - 'LU' = Liste der genderten Adressen
*				 - 'LO' = Liste der genderten Ansprechpartner
*				 - 'A' = Add einer Adresse/Ansprechpartner.
*				 - 'G' = Get einer Adresse/Ansprechpartner.
*				 - 'U' = Update einer vorhandenen Adresse/Ansprechpartner.
*				 - 'D' = Delete einer vorhandenen Adresse/Ansprechpartner.
* 11.07.23 HPG: Datentyp;
*				 - 'SH' = nderungen im Shop
*				 - 'CU' = Adresse
*				 - 'CO' = Ansprechpartner
*				 - 'PR' = Produkt
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 31.08.23 HPG: Es wird jetzt auch ein Fehlertext zur HTTP-Fehler-Nummer zurckgegeben.
* 13.09.23 HPG: Einen verstndlichen Fehlertext fr den Anwender hinzufgen.
* 14.09.23 HPG: Einen verstndlichen OK-Text fr den Anwender hinzufgen.
* 20.12.23 HPG: Parameter tlNoUpload hinzugefgt, wenn gesetzt, werden keine Uploads durchgefhrt sondern nur die JSON-Datei
*				erstellt.
* 18.01.24 HPG: Parameter tcKeyNumber hinzugefgt, um optional den zu verwendenden Key setzen zu knnen, zb. fr Tests.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION HandleWebShopData
**************************

	LPARAMETERS tuID AS STRING, tnID_WS AS INTEGER, tcMode AS STRING, tcDataType AS STRING, tlDontShowInfo AS BOOLEAN, ;
		tlNoUpload AS BOOLEAN, tcKeyNumber AS STRING

	LOCAL lcUserName AS STRING, lcPassWord AS STRING, lcValue AS STRING, lcOKCode AS STRING, lcCommand AS STRING
	LOCAL lcURL4Customer AS STRING, lcResult AS STRING, lcURL AS STRING, lcJSON AS STRING, lcURL4Contact AS STRING
	LOCAL lcURL4WebShop AS STRING, lcURL4Product AS STRING, lcKey AS STRING, lcErrorMessage AS STRING, lcResultCode AS STRING
	LOCAL lcHTTPResultCode AS STRING, lcUserErrorMessage AS STRING, lcUserOKMessage AS STRING, lcURL4Change AS STRING
	LOCAL lcMessage AS STRING, lcLimit AS STRING, lcKeyText4Key AS STRING
	LOCAL llLoadData AS BOOLEAN, llCheckCode AS BOOLEAN, llCreateNewAllowed AS BOOLEAN, llNoIDNecessary AS BOOLEAN
	LOCAL loHttp AS 'wwHttp'

	*-- Parameter setzen
	lcResult = ''
	llCheckCode = .F.
	lcLimit = '_limit=1000'

	*-- Sollen die RestAPI-Funktionen verwendet werden
	IF glUseRestAPI

		*-- Parameter prfen
		IF VARTYPE( tcKeyNumber ) == 'C'
			lcKeyText4Key = 'Key' + tcKeyNumber
		ELSE
			lcKeyText4Key = 'Key'
		ENDIF

		*-- Parameter fr RestAPI aus INI-Datei lesen
		lcValue = getKDVini( 'RestAPI', 'UserName' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcUserName = lcValue
		ELSE
			lcUserName = 'bki'
		ENDIF
		lcValue = getKDVini( 'RestAPI', 'PassWord' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcPassWord = lcValue
		ELSE
			lcPassWord = 'seelberg4'
		ENDIF
		lcValue = getKDVini( 'RestAPI', lcKeyText4Key )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcKey = lcValue
		ELSE
			lcKey = '' && 'a7QMEIdQd5lcfJ3X9d0I1WMb'
		ENDIF
		lcValue = getKDVini( 'RestAPI', 'URL4WebShop' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcURL4WebShop = lcValue
		ELSE
			lcURL4WebShop = 'https://bki.fuf.de/api/bki/shop_change'
		ENDIF
		lcValue = getKDVini( 'RestAPI', 'URL4Customer' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcURL4Customer = lcValue
		ELSE
			lcURL4Customer = 'https://bki.fuf.de/api/bki/customer'
		ENDIF
		lcValue = getKDVini( 'RestAPI', 'URL4Contact' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcURL4Contact = lcValue
		ELSE
			lcURL4Contact = 'https://bki.fuf.de/api/bki/contact'
		ENDIF
		lcValue = getKDVini( 'RestAPI', 'URL4Product' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcURL4Product = lcValue
		ELSE
			lcURL4Product = 'https://bki.fuf.de/api/bki/product'
		ENDIF
		lcValue = getKDVini( 'RestAPI', 'URL4Change' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcURL4Change = lcValue
		ELSE
			lcURL4Change = 'https://bki.fuf.de/api/bki/test'
		ENDIF
		lcValue = getKDVini( 'RestAPI', 'CreateNewAllowed' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			llCreateNewAllowed = ( lcValue == '1' )
		ELSE
			llCreateNewAllowed = .F.
		ENDIF

		*-- Die West-Wind-Bibliotheken laden
		DO wwhttp

		*-- Das HTTP-Objekt erstellen
		loHttp = CREATEOBJECT( 'wwHttp' )

		*-- Den Content-Typ setzen
		loHttp.cContentType = 'application/json'

		*-- Die Zugangsdaten eintragen
		loHttp.AddBasicAuthenticationHeader( lcUserName, lcPassWord )

		*-- Den Key eintragen
		loHttp.AddHeader( 'x-api-key', lcKey )

		*-- Je nach Datentyp verzweigen
		DO CASE

			*-- nderungen im WebShop
			CASE tcDataType = 'SH'

				*-- Je nach Modus den passenden Befehl und
				*-- die URL-Ergnzung setzen
				DO CASE

					*-- List Adressen
					CASE tcMode = 'LU'
						lcURL = lcURL4WebShop + '?entity=customer' + '&' + lcLimit
						lcCommand = 'GET'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .T.
						lcUserErrorMessage = 'Fehler bei Abfrage der genderten Adressen!'
						lcUserOKMessage = 'Abfrage der genderten Adressen erfolgreich'

					*-- List Ansprechpartner
					CASE tcMode = 'LO'
						lcURL = lcURL4WebShop + '?entity=contact' + '&' + lcLimit
						lcCommand = 'GET'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .T.
						lcUserErrorMessage = 'Fehler bei Abfrage der genderten Ansprechpartner!'
						lcUserOKMessage = 'Abfrage der genderten Ansprechpartner erfolgreich'

					*-- List Adressen und Ansprechpartner
					CASE tcMode = 'L'
						lcURL = lcURL4WebShop + '?' + lcLimit
						lcCommand = 'GET'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .T.
						lcUserErrorMessage = 'Fehler bei Abfrage der genderten Adressen und Ansprechpartner!'
						lcUserOKMessage = 'Abfrage der genderten Adressen und Ansprechpartner erfolgreich'

					*-- Get
					CASE tcMode = 'G'
						lcURL = lcURL4WebShop + '/' + Num2C( tnID_WS )
						lcCommand = 'GET'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler bei Abfrage der nderung mit ID:>' + Num2C( tnID_WS ) + '<!'
						lcUserOKMessage = 'Abfrage des nderungseintrages erfolgreich'

					*-- Delete
					CASE tcMode = 'D'
						lcURL = lcURL4WebShop + '/' + Num2C( tnID_WS )
						lcCommand = 'DELETE'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler bei Lschung der nderung mit ID:>' + Num2C( tnID_WS ) + '<!'
						lcUserOKMessage = 'Lschung des nderungseintrages erfolgreich'
				ENDCASE

			*-- Adresse
			CASE tcDataType = 'CU'

				*-- Je nach Modus den passenden Befehl und
				*-- die URL-Ergnzung setzen
				DO CASE

					*-- List
					CASE tcMode = 'L'
						lcURL = lcURL4Customer + '?' + lcLimit
						lcCommand = 'GET'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .T.
						lcUserErrorMessage = 'Fehler bei Abfrage der vorhandenen Adressen!'
						lcUserOKMessage = 'Abfrage der vorhandenen Adressen erfolgreich'

					*-- Add
					CASE tcMode = 'A'
						lcURL = lcURL4Customer
						lcCommand = 'POST'
						lcOKCode = '201'
						llLoadData = .T.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler beim Hinzufgen der Adresse!'
						lcUserOKMessage = 'Neue Adresse hinzugefgt'

					*-- Get
					CASE tcMode = 'G'
						lcURL = lcURL4Customer + '/' + Num2C( tnID_WS )
						lcCommand = 'GET'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler bei Abfrage der Adresse mit ID:>' + Num2C( tnID_WS ) + '<!'
						lcUserOKMessage = 'Adresse bernommen'

					*-- Update
					CASE tcMode = 'U'
						lcURL = lcURL4Customer + '/' + Num2C( tnID_WS )
						lcCommand = 'PUT'
						lcOKCode = '200'
						llLoadData = .T.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler beim Aktualisieren der Adresse mit ID:>' + Num2C( tnID_WS ) + '<!'
						lcUserOKMessage = 'Adresse bertragen'

					*-- Delete
					CASE tcMode = 'D'
						lcURL = lcURL4Customer + '/' + Num2C( tnID_WS )
						lcCommand = 'DELETE'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler beim Lschen der Adresse mit ID:>' + Num2C( tnID_WS ) + '<!'
						lcUserOKMessage = 'Adresse gelscht'
				ENDCASE

				*-- Mssen wir die Adresse/Ansprechpartner als JSON mitsenden?
				IF llLoadData

					*-- Ja, dann die Kundendaten in einen Cursor holen und
					*-- daraus den JSON-String erstellen
					lcJSON = Customer2JSON( tuID )

					*-- 14.12.23 HPG: Zum Debuggen
					IF tlNoUpload
						SET SAFETY OFF
						STRTOFILE( lcJSON, 'C:\TMP\AddCustomer.JSON', 0 )
						SET SAFETY ON
					ENDIF

					*-- Die JSON-Datei/den an den JSON-String anhngen
					loHttp.AddPostKey( lcJSON )
				ENDIF

			*-- Ansprechpartner
			CASE tcDataType = 'CO'

				*-- Je nach Modus den passenden Befehl und
				*-- die URL-Ergnzung setzen
				DO CASE

					*-- List
					CASE tcMode = 'L'
						lcURL = lcURL4Contact + '?' + lcLimit
						lcCommand = 'GET'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .T.
						lcUserErrorMessage = 'Fehler bei Abfrage der vorhandenen Ansprechpartner!'
						lcUserOKMessage = 'Abfrage der vorhandenen Ansprechpartner erfolgreich'

					*-- Add
					CASE tcMode = 'A'
						lcURL = lcURL4Contact
						lcCommand = 'POST'
						lcOKCode = '201'
						llLoadData = .T.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler beim Hinzufgen des Ansprechpartners!'
						lcUserOKMessage = 'Neuer Ansprechpartner hinzugefgt'

					*-- Get
					CASE tcMode = 'G'
						lcURL = lcURL4Contact + '/' + Num2C( tnID_WS )
						lcCommand = 'GET'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler bei Abfrage des Ansprechpartners mit ID:>' + Num2C( tnID_WS ) + '<!'
						lcUserOKMessage = 'Ansprechpartner bernommen'

					*-- Update
					CASE tcMode = 'U'
						lcURL = lcURL4Contact + '/' + Num2C( tnID_WS )
						lcCommand = 'PUT'
						lcOKCode = '200'
						llLoadData = .T.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler beim Aktualisieren des Ansprechpartners mit ID:>' + Num2C( tnID_WS ) + '<!'
						lcUserOKMessage = 'Ansprechpartner bertragen'

					*-- Delete
					CASE tcMode = 'D'
						lcURL = lcURL4Contact + '/' + Num2C( tnID_WS )
						lcCommand = 'DELETE'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler beim Lschen des Ansprechpartners mit ID:>' + Num2C( tnID_WS ) + '<!'
						lcUserOKMessage = 'Ansprechpartner gelscht'
				ENDCASE

				*-- Mssen wir den Ansprechpartner als JSON mitsenden?
				IF llLoadData

					*-- Ja, dann die Ansprechpartnerdaten in einen Cursor
					*-- holen und daraus den JSON-String erstellen
					lcJSON = Contact2JSON( tuID )

					*-- 14.12.23 HPG: Zum Debuggen
					IF tlNoUpload
						SET SAFETY OFF
						STRTOFILE( lcJSON, 'C:\TMP\AddContact.JSON', 0 )
						SET SAFETY ON
					ENDIF

					*-- Die JSON-Datei/den JSON-String anhngen
					loHttp.AddPostKey( lcJSON )
				ENDIF

			*-- Produkte
			CASE tcDataType = 'PR'

				*-- Je nach Modus den passenden Befehl und
				*-- die URL-Ergnzung setzen
				DO CASE

					*-- List
					CASE tcMode = 'L'
						lcURL = lcURL4Product + '?' + lcLimit
						lcCommand = 'GET'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .T.
						lcUserErrorMessage = 'Fehler bei Abfrage der vorhandenen Produkte!'
						lcUserOKMessage = 'Abfrage der vorhandenen Produkten erfolgreich'

					*-- Get
					CASE tcMode = 'G'
						lcURL = lcURL4Product + '/' + Num2C( tnID_WS )
						lcCommand = 'GET'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler bei Abfrage des Produktes mit ID:>' + Num2C( tnID_WS ) + '<!'
						lcUserOKMessage = 'Produkt bernommen'

					*-- Delete
					CASE tcMode = 'D'
						lcURL = lcURL4Product + '/' + Num2C( tnID_WS )
						lcCommand = 'DELETE'
						lcOKCode = '200'
						llLoadData = .F.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler beim Lschen des Produktes mit ID:>' + Num2C( tnID_WS ) + '<!'
						lcUserOKMessage = 'Produkt gelscht'
				ENDCASE

			*-- Change
			CASE tcDataType = 'CH'

				*-- Je nach Modus den passenden Befehl und
				*-- die URL-Ergnzung setzen
				DO CASE

					*-- Post
					CASE tcMode = 'P'
						lcURL = lcURL4Change
						lcCommand = 'POST'
						lcOKCode = '200'
						llLoadData = .T.
						llNoIDNecessary = .F.
						lcUserErrorMessage = 'Fehler beim Absetzen des nderungsbefehls!'
						lcUserOKMessage = 'nderungsbefehl erfolgreich abgesetzt'
				ENDCASE

				*-- Mssen wir Daten als JSON mitsenden?
				IF llLoadData

					*-- Ja, dann die Adressdaten holen und in den
					*-- nderungs-JSON-String schreiben
					lcJSON = CustomerChange2JSON( tuID )

					*-- Die JSON-Datei/den JSON-String anhngen
					loHttp.AddPostKey( lcJSON )
				ENDIF
		ENDCASE

		*-- Haben wir berhaupt eine WebShop-ID fr diesen
		*-- Kunden/Ansprechpartner bzw. ist berhaupt eine
		*-- ID fr die Message notwendig?
		IF tnID_WS > 0 .OR. llNoIDNecessary

			*-- Ja, nur dann macht ein Update bzw. ein Delete
			*-- berhaupt Sinn.
			*-- Per GET, PUT oder DEL die Daten holen, updaten
			*-- oder lschen
			*-- Nur wenn das Kennzeichen tlNoUpload nicht
			*-- gesetzt ist, fhren wir den Befehl aus
			IF .NOT. tlNoUpload
				lcResult = loHttp.&lcCommand( lcURL )

				*-- 07.02.24 HPG: Zum Debuggen
				IF .F.
					SET SAFETY OFF
					TRY
						STRTOFILE( lcURL + CR, 'C:\TMP\WebShopURL.log', 1 )
					CATCH
					ENDTRY
					SET SAFETY ON
				ENDIF

				*-- Der Rckgabe-Code ist zu prfen
				llCheckCode = .T.
			ENDIF

		*-- Andernfalls wird die Adresse/Ansprechpartner
		*-- neu angelegt
		ELSE

			*-- Darf eine Neuanlage erfolgen?
			IF llCreateNewAllowed

				*-- Ja, dann ausfhren
				*-- Nur wenn das Kennzeichen tlNoUpload nicht
				*-- gesetzt ist, fhren wir den Befehl aus
				IF .NOT. tlNoUpload
					lcResult = loHttp.POST( lcURL )

					*-- Der Rckgabe-Code ist zu prfen
					llCheckCode = .T.
				ENDIF
			ENDIF
		ENDIF

		*-- Ist der Rckgabe-Code zu prfen?
		IF llCheckCode

			*-- Die Rckgabe behandeln
			IF loHttp.cResultCode # lcOKCode

				*-- bertragungsfehler
				lcErrorMessage = 'Fehler: ' + loHttp.cErrorMsg

				*-- Steht etwas im JSON-String?
				IF LEN( ALLTRIM( lcResult ) ) > 0
					lcErrorMessage = JSON2Message( lcResult ) + CR + lcErrorMessage
				ELSE
					lcErrorMessage = lcErrorMessage + CR + 'Extraheader: ' + loHttp.cExtraHeaders
				ENDIF

				*-- Steht etwas in cResult-Code?
				IF LEN( ALLTRIM( loHttp.cResultCode ) ) > 0
					lcResultCode = loHttp.cResultCode
				ELSE

					*-- Die Anfrage kann nicht ohne gltige
					*-- Authentifizierung durchgefhrt werden.
					*-- Es fehlt der X-Api-Key
					lcResultCode = '401'
				ENDIF

				*-- Zur Fehlernummer den Fehlertext holen
				lcHTTPResultCode = lcResultCode + ' = ' + GetHTTPErrorText( lcResultCode )

				*-- Fehlermeldung ausgeben
				lcMessage = lcUserErrorMessage + CR + CR + lcCommand + ': HTTP Antwort: ' + lcHTTPResultCode + CR + lcErrorMessage

				*-- Nur wenn die Information zu zeigen ist
				IF !tlDontShowInfo
					Fehler( lcMessage )
				ENDIF
				lcResult = 'ERROR'

				*-- Protokoll immer in die Error-Log-Datei schreiben
				STRTOFILE( TTOC( DATETIME(), 1 ) + HT + STRTRAN( lcMessage, CR, '#' ) + CR, 'RestAPIError.log', 1 )

			*-- Andernfalls
			ELSE

				*-- Nur wenn die Information zu zeigen ist
				IF !tlDontShowInfo
					Hinweis( lcUserOKMessage + CR + CR + lcCommand + ': Alles OK' )
				ENDIF
			ENDIF
		ENDIF
	ENDIF

	*-- Rckgabe
	RETURN( lcResult )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Bringt die WebView-Artikel-Daten per RestAPI auf den neuesten Stand.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 25.11.2024
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 25.11.24 HPG: Modus fr die WebView-Artikel-Abfrage/Update/Delete:
*				 - 'A' = Add eines WebView-Artikel
*				 - 'G' = Get eines WebView-Artikel
*				 - 'U' = Update eines WebView-Artikel
*				 - 'D' = Delete eines WebView-Artikel
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION HandleWebViewData( tcMode AS STRING, tcJSON AS STRING, tlDontShowInfo AS BOOLEAN, tlNoUpload AS BOOLEAN )
******************************************************************************************************************

	LOCAL lcUserName AS STRING, lcPassWord AS STRING, lcURL AS STRING, lcValue AS STRING, lcOKCode AS STRING
	LOCAL lcCommand AS STRING, lcURL4WebView AS STRING, lcResult AS STRING, lcErrorMessage AS STRING, lcResultCode AS STRING
	LOCAL lcHTTPResultCode AS STRING, lcUserErrorMessage AS STRING, lcMessage AS STRING, lcUserOKMessage AS STRING
	LOCAL llLoadData AS BOOLEAN, llCheckCode AS BOOLEAN, llNoIDNecessary AS BOOLEAN
	LOCAL loHttp AS 'wwHttp'
	*:Global tnID_WS

	*-- Parameter setzen
	lcResult = ''
	llCheckCode = .F.
	tnID_WS = 0

	*-- Sollen die RestAPI-Funktionen verwendet werden
	IF glUseRestAPI

		*-- Parameter fr RestAPI aus INI-Datei lesen
		lcValue = getKDVini( 'RestAPIWebView', 'UserName' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcUserName = lcValue
		ELSE
			lcUserName = 'bki'
		ENDIF
		lcValue = getKDVini( 'RestAPIWebView', 'PassWord' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcPassWord = lcValue
		ELSE
			lcPassWord = 'seelberg4'
		ENDIF
		lcValue = getKDVini( 'RestAPIWebView', 'URL4WebView' )
		IF !ISNULL( lcValue ) AND !EMPTY( lcValue )
			lcURL4WebView = lcValue
		ELSE
			lcURL4WebView = 'https://bki.fuf.de/api/bki/webview'
		ENDIF

		*-- Die West-Wind-Bibliotheken laden
		DO wwhttp

		*-- Das HTTP-Objekt erstellen
		loHttp = CREATEOBJECT( 'wwHttp' )

		*-- Den Content-Typ setzen
		loHttp.cContentType = 'application/json'

		*-- Die Zugangsdaten eintragen
		loHttp.AddBasicAuthenticationHeader( lcUserName, lcPassWord )

		*-- Je nach Modus den passenden Befehl und
		*-- die URL-Ergnzung setzen
		DO CASE

			*-- Add
			CASE tcMode = 'A'
				lcURL = lcURL4WebView
				lcCommand = 'POST'
				lcOKCode = '200'
				llLoadData = .T.
				llNoIDNecessary = .F.
				*-- 09.12.24 HPG: Aussagekrftigere Meldungen bei Fehler bzw. Erfolg des Uploads
				*lcUserErrorMessage = 'Fehler beim Hinzufgen des WebView-Artikels!'
				*lcUserOKMessage = 'Neuen WebView-Artikel hinzugefgt'
				lcUserErrorMessage = 'Der Zugang konnte nicht aktiviert werden!'
				lcUserOKMessage = 'Der Zugang wurde erfolgreich aktiviert.'

			*-- Get
			CASE tcMode = 'G'
				lcURL = lcURL4WebView + '/' + Num2C( tnID_WS )
				lcCommand = 'GET'
				lcOKCode = '200'
				llLoadData = .F.
				llNoIDNecessary = .F.
				lcUserErrorMessage = 'Fehler bei Abfrage des WebView-Artikels mit ID:>' + Num2C( tnID_WS ) + '<!'
				lcUserOKMessage = 'WebView-Artikel bernommen'

			*-- Update
			CASE tcMode = 'U'
				lcURL = lcURL4WebView + '/' + Num2C( tnID_WS )
				lcCommand = 'PUT'
				lcOKCode = '200'
				llLoadData = .T.
				llNoIDNecessary = .F.
				lcUserErrorMessage = 'Fehler beim Aktualisieren des WebView-Artikels mit ID:>' + Num2C( tnID_WS ) + '<!'
				lcUserOKMessage = 'WebView-Artikel bertragen'

			*-- Delete
			CASE tcMode = 'D'
				lcURL = lcURL4WebView + '/' + Num2C( tnID_WS )
				lcCommand = 'DELETE'
				lcOKCode = '200'
				llLoadData = .F.
				llNoIDNecessary = .F.
				lcUserErrorMessage = 'Fehler beim Lschen des WebView-Artikels mit ID:>' + Num2C( tnID_WS ) + '<!'
				lcUserOKMessage = 'WebView-Artikel gelscht'
		ENDCASE

		*-- Mssen wir den WebView-Artikel als JSON mitsenden?
		IF llLoadData

			*-- 25.11.24 HPG: Zum Debuggen
			IF tlNoUpload
				SET SAFETY OFF
				STRTOFILE( tcJSON, 'C:\TMP\AddWebViewArticle.JSON', 0 )
				SET SAFETY ON
			ENDIF

			*-- Die JSON-Datei/den an den JSON-String anhngen
			loHttp.AddPostKey( tcJSON )
		ENDIF

		*-- Haben wir berhaupt eine WebView-Artikel-ID fr diesen
		*-- Artikel bzw. ist berhaupt eine ID fr die Message notwendig?
		IF tnID_WS > 0 .OR. llNoIDNecessary

			*-- Ja, nur dann macht ein Update bzw. ein Delete
			*-- berhaupt Sinn.
			*-- Per GET, PUT oder DEL die Daten holen, updaten
			*-- oder lschen
			*-- Nur wenn das Kennzeichen tlNoUpload nicht
			*-- gesetzt ist, fhren wir den Befehl aus
			IF .NOT. tlNoUpload
				lcResult = loHttp.&lcCommand( lcURL )

				*-- 25.11.24 HPG: Zum Debuggen
				IF .F.
					SET SAFETY OFF
					TRY
						STRTOFILE( lcURL + CR, 'C:\TMP\WebViewArticle.log', 1 )
					CATCH
					ENDTRY
					SET SAFETY ON
				ENDIF

				*-- Der Rckgabe-Code ist zu prfen
				llCheckCode = .T.
			ENDIF

		*-- Andernfalls wird der WebView-Artikel neu angelegt
		ELSE

			*-- Nur wenn das Kennzeichen tlNoUpload nicht
			*-- gesetzt ist, fhren wir den Befehl aus
			IF .NOT. tlNoUpload
				lcResult = loHttp.POST( lcURL )

				*-- Der Rckgabe-Code ist zu prfen
				llCheckCode = .T.
			ENDIF
		ENDIF

		*-- Ist der Rckgabe-Code zu prfen?
		IF llCheckCode

			*-- Die Rckgabe behandeln
			IF loHttp.cResultCode # lcOKCode

				*-- bertragungsfehler
				lcErrorMessage = 'Fehler: ' + loHttp.cErrorMsg

				*-- Steht etwas im JSON-String?
				IF LEN( ALLTRIM( lcResult ) ) > 0
					lcErrorMessage = JSON2Message( lcResult ) + CR + lcErrorMessage
				ELSE
					lcErrorMessage = lcErrorMessage + CR + 'Extraheader: ' + loHttp.cExtraHeaders
				ENDIF

				*-- Steht etwas in cResult-Code?
				IF LEN( ALLTRIM( loHttp.cResultCode ) ) > 0
					lcResultCode = loHttp.cResultCode
				ELSE

					*-- Die Anfrage kann nicht ohne gltige
					*-- Authentifizierung durchgefhrt werden.
					*-- Es fehlt der X-Api-Key
					lcResultCode = '401'
				ENDIF

				*-- Zur Fehlernummer den Fehlertext holen
				lcHTTPResultCode = lcResultCode + ' = ' + GetHTTPErrorText( lcResultCode )

				*-- Fehlermeldung ausgeben
				lcMessage = lcUserErrorMessage + CR + CR + lcCommand + ': HTTP Antwort: ' + lcHTTPResultCode + CR + lcErrorMessage

				*-- Nur wenn die Information zu zeigen ist
				IF !tlDontShowInfo

					*-- 09.12.24 HPG: Verstndlichere Meldung fr den Anwender
					*Fehler( lcMessage )
					Fehler( lcUserErrorMessage )
				ENDIF
				lcResult = 'ERROR'

				*-- Protokoll immer in die Error-Log-Datei schreiben
				STRTOFILE( TTOC( DATETIME(), 1 ) + HT + STRTRAN( lcMessage, CR, '#' ) + CR, 'RestAPIErrorWebView.log', 1 )

			*-- Andernfalls
			ELSE

				*-- Nur wenn die Information zu zeigen ist
				IF !tlDontShowInfo

					*-- 09.12.24 HPG: Verstndlichere Meldung fr den Anwender
					*Hinweis( lcUserOKMessage + CR + CR + lcCommand + ': Alles OK' )
					Hinweis( lcUserOKMessage )
				ENDIF
			ENDIF
		ENDIF
	ENDIF

	*-- Rckgabe
	RETURN( lcResult )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Ansprechpartnerdaten als JSON-String in einen Cursor schreiben und den zugehrigen Ansprechpartner
*				updaten.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 11.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION JSON2Contact( tcJSON AS STRING )
*****************************************

	LOCAL lcJSON AS STRING, lcValue AS STRING
	LOCAL lnASPID AS INTEGER
	LOCAL loJSON AS 'wwJsonSerializer'
	LOCAL loCollection AS COLLECTION, loContact AS COLLECTION
	*:Global aID[1]

	*-- Sollen die RestAPI-Funktionen verwendet werden
	IF glUseRestAPI

#IFDEF  RESTAPI_WRITELOG
		*-- Zum Testen
		TRY
			SET SAFETY OFF
			STRTOFILE( tcJSON, 'c:\TMP\Contact.JSON' )
			SET SAFETY ON
		CATCH
		ENDTRY
#ENDIF

		*-- UTF-8 in ANSI umwandeln
		lcJson = STRCONV( tcJson, 12 )			&& UTF-8 in Unicode umwandeln
		lcJson = STRCONV( lcJson, 6 )			&& Unicode in Multi-Byte umwandeln

		*-- Die JSON-Klasse laden und erstellen
		DO wwJsonSerializer
		loJSON = CREATEOBJECT( 'wwJsonSerializer' )

		*-- Den kompletten JSON-String in eine Collection umwandeln
		loCollection = loJSON.DeserializeJson( lcJSON )

		*-- Den Verweis auf die erste Adresse der KDV-Collection
		*-- holen
		loContact = loCollection.KDV

		*-- Die Ansprechpartner-ID holen und prfen ob diese
		*-- berhaupt vorhanden ist
		lnASPID = loContact.ID
		SELECT ID FROM BKI_ANSPRECH WHERE ID = lnASPID INTO ARRAY aID
		IF _TALLY > 0

			*-- Ja, diese Ansprechpartner-ID ist vorhanden
			*-- Dann diese Daten eintragen
			TransferDatafromJSON( loCollection, loContact, 'Bki_Ansprech', 'CU' )

		*-- Andernfalls Info an den Anwender ausgeben
		ELSE
			Fehler( 'Ansprechpartner mit ID >' + Num2C( lnASPID ) + '< nicht gefunden!' )
		ENDIF
	ENDIF

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Ansprechpartnerdaten als JSON-String in einen Cursor schreiben und an die Importtabelle anhngen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 09.08.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION JSON2ContactImport( tcJSON AS STRING, tcImportFile AS STRING )
***********************************************************************

	LOCAL lcJSON AS STRING, lcKdNr AS STRING, lcValue AS STRING, lcImportTable AS STRING
	LOCAL loJSON AS 'wwJsonSerializer'
	LOCAL loCollection AS COLLECTION, loContact AS COLLECTION
	*:Global aID[1]

	*-- Parameter setzen
	lcImportTable = tcImportFile + '.DBF'

	*-- Sollen die RestAPI-Funktionen verwendet werden
	IF glUseRestAPI

#IFDEF  RESTAPI_WRITELOG
		*-- Zum Testen
		TRY
			SET SAFETY OFF
			STRTOFILE( tcJSON, 'c:\TMP\ContactImport.JSON' )
			SET SAFETY ON
		CATCH
		ENDTRY
#ENDIF

		*-- UTF-8 in ANSI umwandeln
		lcJson = STRCONV( tcJson, 12 )			&& UTF-8 in Unicode umwandeln
		lcJson = STRCONV( lcJson, 6 )			&& Unicode in Multi-Byte umwandeln

		*-- Die JSON-Klasse laden und erstellen
		DO wwJsonSerializer
		loJSON = CREATEOBJECT( 'wwJsonSerializer' )

		*-- Den kompletten JSON-String in eine Collection umwandeln
		loCollection = loJSON.DeserializeJson( lcJSON )

		*-- Den Verweis auf die erste Adresse der KDV-Collection
		*-- holen
		loContact = loCollection.KDV

		*-- Ist die Import-Tabelle vorhanden?
		IF FileExist( lcImportTable )

			*-- Ist diese schon geffnet?
			IF !USED( 'Imp' )

				*-- Nein, dann aufmachen
				USE ( lcImportTable ) IN 0 ALIAS Imp
			ENDIF

			*-- Einen leeren Datensatz anhngen und dann
			*-- die Daten eintragen
			SELECT Imp
			APPEND BLANK
			TransferDatafromJSON( loCollection, loContact, 'Imp', 'CO' )

		*-- Andernfalls Info an den Anwender ausgeben
		ELSE
			Fehler( 'Import-Tabelle >' + lcImportTable + '< nicht gefunden!' )
		ENDIF
	ENDIF

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Von den neuen/genderten Ansprechpartnerdaten als JSON-String, die WebShop-ID (ID_WS) in die Bki_Ansprech
*				schreiben.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 24.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION JSON2ContactWSID( tcJSON AS STRING )
*********************************************

	LOCAL lcJSON AS STRING, lcValue AS STRING
	LOCAL lnASPID AS INTEGER
	LOCAL loJSON AS 'wwJsonSerializer'
	LOCAL loCollection AS COLLECTION, loContact AS COLLECTION
	*:Global aID[1]

	*-- Sollen die RestAPI-Funktionen verwendet werden
	IF glUseRestAPI

#IFDEF  RESTAPI_WRITELOG
		*-- Zum Testen
		TRY
			SET SAFETY OFF
			STRTOFILE( tcJSON, 'c:\TMP\Contact.JSON' )
			SET SAFETY ON
		CATCH
		ENDTRY
#ENDIF

		*-- UTF-8 in ANSI umwandeln
		lcJson = STRCONV( tcJson, 12 )			&& UTF-8 in Unicode umwandeln
		lcJson = STRCONV( lcJson, 6 )			&& Unicode in Multi-Byte umwandeln

		*-- Die JSON-Klasse laden und erstellen
		DO wwJsonSerializer
		loJSON = CREATEOBJECT( 'wwJsonSerializer' )

		*-- Den kompletten JSON-String in eine Collection umwandeln
		loCollection = loJSON.DeserializeJson( lcJSON )

		*-- Zuerst den Status abfragen
		IF loCollection.Status = 'ok'

			*-- Den Verweis auf den erste Ansprechpartner der KDV-Collection
			*-- holen
			loContact = loCollection.Item.KDV

			*-- Die Ansprechpartner-ID holen und prfen ob diese
			*-- berhaupt vorhanden ist
			lnASPID = loContact.ID
			SELECT ID FROM BKI_ANSPRECH WHERE ID = lnASPID INTO ARRAY aID
			IF _TALLY > 0

				*-- Ja, diese Ansprechpartner-ID ist vorhanden
				*-- Dann die WebShop-ID eintragen
				SELECT Bki_Ansprech
				REPLACE Bki_Ansprech.ID_WS WITH loCollection.ID FOR Bki_Ansprech.ID = lnASPID

			*-- Andernfalls Info an den Anwender ausgeben
			ELSE
				Fehler( 'Ansprechpartner-ID >' + Num2C( lnASPID ) + '< nicht gefunden!' )
			ENDIF

		*-- Andernfalls Info an den Anwender ausgeben
		ELSE
			Fehler( 'Fehler beim Abfragen der Ansprechpartnerdaten von ID >' + Num2C( lnASPID ) + '<!' )
		ENDIF
	ENDIF

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Kundendaten als JSON-String in einen Cursor schreiben und die zugehrige Adresse updaten.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 05.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION JSON2Customer( tcJSON AS STRING )
*****************************************

	LOCAL lcJSON AS STRING, lcKdNr AS STRING, lcValue AS STRING
	LOCAL loJSON AS 'wwJsonSerializer'
	LOCAL loCollection AS COLLECTION, loCustomer AS COLLECTION
	*:Global aID[1]

	*-- Sollen die RestAPI-Funktionen verwendet werden
	IF glUseRestAPI

#IFDEF  RESTAPI_WRITELOG
		*-- Zum Testen
		TRY
			SET SAFETY OFF
			STRTOFILE( tcJSON, 'c:\TMP\CustomerAddress.JSON' )
			SET SAFETY ON
		CATCH
		ENDTRY
#ENDIF

		*-- UTF-8 in ANSI umwandeln
		lcJson = STRCONV( tcJson, 12 )			&& UTF-8 in Unicode umwandeln
		lcJson = STRCONV( lcJson, 6 )			&& Unicode in Multi-Byte umwandeln

		*-- Die JSON-Klasse laden und erstellen
		DO wwJsonSerializer
		loJSON = CREATEOBJECT( 'wwJsonSerializer' )

		*-- Den kompletten JSON-String in eine Collection umwandeln
		loCollection = loJSON.DeserializeJson( lcJSON )

		*-- Den Verweis auf die erste Adresse der KDV-Collection
		*-- holen
		loCustomer = loCollection.KDV

		*-- Die Kunden-Nummer holen und prfen ob diese
		*-- berhaupt vorhanden ist
		lcKdNr = ALLTRIM( loCustomer.KdNr )
		SELECT ID FROM Bkb_Kund WHERE KdNr = lcKdNr INTO ARRAY aID
		IF _TALLY > 0

			*-- Ja, diese Kunden-Nummer ist vorhanden
			*-- Dann diese Daten eintragen
			TransferDatafromJSON( loCollection, loCustomer, 'Bkb_Kund', 'CU' )

		*-- Andernfalls Info an den Anwender ausgeben
		ELSE
			Fehler( 'Kundennummer >' + lcKdNr + '< nicht gefunden!' )
		ENDIF
	ENDIF

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Die Kundendaten als JSON-String in einen Cursor schreiben und an die Importtabelle anhngen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 25.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION JSON2CustomerImport( tcJSON AS STRING, tcImportFile AS STRING )
************************************************************************

	LOCAL lcJSON AS STRING, lcKdNr AS STRING, lcValue AS STRING, lcImportTable AS STRING
	LOCAL loJSON AS 'wwJsonSerializer'
	LOCAL loCollection AS COLLECTION, loCustomer AS COLLECTION
	*:Global aID[1]

	*-- Parameter setzen
	lcImportTable = tcImportFile + '.DBF'

	*-- Sollen die RestAPI-Funktionen verwendet werden
	IF glUseRestAPI

		*-- UTF-8 in ANSI umwandeln
		lcJson = STRCONV( tcJson, 12 )			&& UTF-8 in Unicode umwandeln
		lcJson = STRCONV( lcJson, 6 )			&& Unicode in Multi-Byte umwandeln

		*-- Die JSON-Klasse laden und erstellen
		DO wwJsonSerializer
		loJSON = CREATEOBJECT( 'wwJsonSerializer' )

		*-- Den kompletten JSON-String in eine Collection umwandeln
		loCollection = loJSON.DeserializeJson( lcJSON )

		*-- Den Verweis auf die erste Adresse der KDV-Collection
		*-- holen
		loCustomer = loCollection.KDV

#IFDEF  RESTAPI_WRITELOG
		*-- Zum Testen
		TRY
			SET SAFETY OFF
			STRTOFILE( tcJSON, 'c:\TMP\CustomerImport_' + Num2C( loCollection.ID ) + '.JSON' )
			SET SAFETY ON
		CATCH
		ENDTRY
#ENDIF

		*-- Ist die Import-Tabelle vorhanden?
		IF FileExist( lcImportTable )

			*-- Ist diese schon geffnet?
			IF !USED( 'Imp' )

				*-- Nein, dann aufmachen
				USE ( lcImportTable ) IN 0 ALIAS Imp
			ENDIF

			*-- Einen leeren Datensatz anhngen und dann
			*-- die Daten eintragen
			SELECT Imp
			APPEND BLANK
			TransferDatafromJSON( loCollection, loCustomer, 'Imp', 'CU' )

		*-- Andernfalls Info an den Anwender ausgeben
		ELSE
			Fehler( 'Import-Tabelle >' + lcImportTable + '< nicht gefunden!' )
		ENDIF
	ENDIF

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Von den neuen/genderten Kundendaten als JSON-String, die WebShop-ID (ID_WS) in die Bkb_Kund schreiben.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 24.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION JSON2CustomerWSID( tcJSON AS STRING )
**********************************************

	LOCAL lcJSON AS STRING, lcKdNr AS STRING, lcValue AS STRING
	LOCAL loJSON AS 'wwJsonSerializer'
	LOCAL loCollection AS COLLECTION, loCustomer AS COLLECTION
	*:Global aID[1]

	*-- Sollen die RestAPI-Funktionen verwendet werden
	IF glUseRestAPI

#IFDEF  RESTAPI_WRITELOG
		*-- Zum Testen
		TRY
			SET SAFETY OFF
			STRTOFILE( tcJSON, 'c:\TMP\CustomerAddress.JSON' )
			SET SAFETY ON
		CATCH
		ENDTRY
#ENDIF

		*-- UTF-8 in ANSI umwandeln
		lcJson = STRCONV( tcJson, 12 )			&& UTF-8 in Unicode umwandeln
		lcJson = STRCONV( lcJson, 6 )			&& Unicode in Multi-Byte umwandeln

		*-- Die JSON-Klasse laden und erstellen
		DO wwJsonSerializer
		loJSON = CREATEOBJECT( 'wwJsonSerializer' )

		*-- Den kompletten JSON-String in eine Collection umwandeln
		loCollection = loJSON.DeserializeJson( lcJSON )

		*-- Zuerst den Status abfragen
		IF loCollection.Status = 'ok'

			*-- Den Verweis auf die erste Adresse der KDV-Collection
			*-- holen
			loCustomer = loCollection.Item.KDV

			*-- Die Kunden-Nummer holen und prfen ob diese
			*-- berhaupt vorhanden ist
			lcKdNr = ALLTRIM( loCustomer.KdNr )
			SELECT ID FROM Bkb_Kund WHERE KdNr = lcKdNr INTO ARRAY aID
			IF _TALLY > 0

				*-- Ja, diese Kunden-Nummer ist vorhanden
				*-- Dann die WebShop-ID eintragen
				SELECT Bkb_Kund
				REPLACE Bkb_Kund.ID_WS WITH loCollection.ID FOR Bkb_Kund.KDNR == lcKdNr

			*-- Andernfalls Info an den Anwender ausgeben
			ELSE
				Fehler( 'Kundennummer >' + lcKdNr + '< nicht gefunden!' )
			ENDIF

		*-- Andernfalls Info an den Anwender ausgeben
		ELSE
			Fehler( 'Fehler beim Abfragen der Kundendaten!' )
		ENDIF
	ENDIF

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Holt die IDs der genderten Entitten und schreibt diese in einen CURSOR.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 25.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 25.07.23 HPG: Es werden dabei je ID nur ein Eintrag in den CURSOR geschrieben.
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION JSON2IDS( tcJSON AS STRING )
*************************************

	LOCAL lcValue AS STRING
	LOCAL lnLoop AS INTEGER, lnMaxLoop AS INTEGER, lnID_CHG AS INTEGER, lnID_WS AS INTEGER
	LOCAL llReturn AS BOOLEAN
	LOCAL loJSON AS 'wwJsonSerializer'
	LOCAL loCollection AS COLLECTION
	*:Global aID[1]

	*-- Sollen die RestAPI-Funktionen verwendet werden
	IF glUseRestAPI

#IFDEF  RESTAPI_WRITELOG
		*-- Zum Testen
		TRY
			SET SAFETY OFF
			STRTOFILE( tcJSON, 'c:\TMP\Contact.JSON' )
			SET SAFETY ON
		CATCH
		ENDTRY
#ENDIF

		*-- Die JSON-Klasse laden und erstellen
		DO wwJsonSerializer
		loJSON = CREATEOBJECT( 'wwJsonSerializer' )

		*-- Den kompletten JSON-String in eine Collection umwandeln
		loCollection = loJSON.DeserializeJson( tcJSON )

		*-- Zuerst die Anzahl abfragen
		lnMaxLoop = MIN( loCollection._Total, loCollection._Limit )
		IF lnMaxLoop > 0

			*-- Temporren Cursor erstellen
			CREATE CURSOR cursID_WSAll ( ID_WS I( 4 ), ID_CHG I( 4 ) )
			SELECT cursID_WSAll

			*-- TRY-/CATCH-Block
			TRY

				*-- Schleife ber alle nderungen zum Sammeln der IDs
				FOR lnLoop = 1 TO lnMaxLoop

					*-- ID aus der Collection holen
					lnID_WS = loCollection.Items[lnLoop].Shop_ID
					lnID_CHG = loCollection.Items[lnLoop].ID
					INSERT INTO cursID_WSAll ( ID_WS, ID_CHG ) VALUES ( lnID_WS, lnID_CHG )
				NEXT

			*-- CATCH
			CATCH
			ENDTRY

			*-- GROUP BY zum Sicherstellen dass eine Entitt
			*-- nur einmal in den CURSOR kommt
			*-- 15.02.24 HPG: Muss bei einem GROUP BY auch sortiert werden?
			SELECT ID_WS FROM cursID_WSAll GROUP BY ID_WS INTO CURSOR cursID_WS
			IF _TALLY > 0

				*-- Die Rckgabe setzen
				llReturn = .T.

			*-- Andernfalls nix gefunden
			ELSE

				*-- Die Rckgabe setzen
				llReturn = .F.
			ENDIF

		*-- Andernfalls Info an den Anwender ausgeben
		ELSE
			Fehler( 'Fehler beim Abfragen der IDs!' + CHR( 13 ) + 'Keine nderungen vorhanden' )
			llReturn = .F.
		ENDIF
	ENDIF

	*-- Rckgabe
	RETURN( llReturn )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Bei einem Fehler aus dem JSON-String die Message holen.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 24.07.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
* 28.11.24 HPG: Erweiterte Prfung des JSON-Strings:
*				 - Prfung auf ein JSON => wenn kein JSON, den String 1:1 zurckgeben.
*				 - JSON prfen:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION JSON2Message( tcJSON AS STRING )
*****************************************

	LOCAL lcValue AS STRING, lcMessage AS STRING
	LOCAL loJSON AS 'wwJsonSerializer'
	LOCAL loCollection AS COLLECTION
	*:Global aID[1]

	*-- Parameter setzen
	lcMessage = ''

	*-- Sollen die RestAPI-Funktionen verwendet werden
	IF glUseRestAPI

#IFDEF  RESTAPI_WRITELOG
		*-- Zum Testen
		TRY
			SET SAFETY OFF
			STRTOFILE( tcJSON, 'c:\TMP\Message.JSON' )
			SET SAFETY ON
		CATCH
		ENDTRY
#ENDIF

		*-- Erstes und letztes Zeichen
		lcJSON = ALLTRIM( tcJSON )
		lcLeftChar = LEFT( lcJSON, 1 )
		lcRightChar = RIGHT( TRIM( lcJSON ), 1 )

		*-- Auf JSON prfen
		IF lcLeftChar = '{' .AND. lcRightChar = '}'

			*-- Die JSON-Klasse laden und erstellen
			DO wwJsonSerializer
			loJSON = CREATEOBJECT( 'wwJsonSerializer' )

			*-- Protected .NET-Eigenschaften nicht einlesen
			loJSON.IgnoreDollarVars = .T.

			*-- Den kompletten JSON-String in eine Collection umwandeln
			loCollection = loJSON.DeserializeJson( tcJSON )

			*-- Den Status-Typ abfragen
			luState = loCollection.STATUS
			lcStateType = VARTYPE( luState )

			*-- Je nach Datentyp verzweigen
			DO CASE

				*-- Status vom Typ numerisch
				CASE lcStateType = 'N'

					DO CASE

						*-- Status 400
						CASE loCollection.STATUS = 400

							*-- Den Titel holen
							lcMessage = loCollection.TITLE

						*-- Andernfalls Info an den Anwender ausgeben
						OTHERWISE
							Fehler( 'Fehler beim Abfragen der Fehlermeldung!' )
					ENDCASE

				*-- Status vom Typ Zeichen
				CASE lcStateType = 'C'

					DO CASE

						*-- Error
						CASE loCollection.STATUS = 'error'

							*-- Die Fehlermeldung holen
							lcMessage = loCollection.MESSAGE

						*-- OK
						CASE loCollection.STATUS = 'ok'

							*-- Die Fehlermeldung setzen
							lcMessage = 'Alles OK'

						*-- Andernfalls Info an den Anwender ausgeben
						OTHERWISE
							Fehler( 'Fehler beim Abfragen der Fehlermeldung!' )
					ENDCASE
			ENDCASE

		*-- Andernfalls kein JSON
		ELSE

			*-- Zeichenkette direkt zurckgeben
			lcMessage = tcJSON
		ENDIF
	ENDIF

	*-- Rckgabe
	RETURN( lcMessage )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: Gibt anhand der KDV-Anrede-ID die zugehrige WebShop-Anrede-ID zurck.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 20.12.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
* 09.01.24 HPG: Wenn kein Eintrag in der Ansprechpartner-Tabelle gefunden wird, setzen wird die ID auf 3 fr keine Anrede im
*				WebShop.
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION MapKDVSalutation2WS( tnID AS INTEGER )
***********************************************

	LOCAL lcField AS STRING
	LOCAL lnIDWebShop AS INTEGER
	LOCAL aID[1]

	*-- Parameter setzen
	*-- 09.01.24 HPG: Wenn kein Eintrag in der Ansprechpartner-Tabelle
	*-- gefunden wird, setzen wird die ID auf 3 fr keine Anrede im
	*-- WebShop.
	lnIDWebShop = 3

	*-- Ist die Tabelle Bki_Ansprech geffnet
	IF USED( 'Bki_Ansprech' )

		*-- Den Inhalt des Feldes Anrede bzw. Brief holen
		SELECT ID_WS FROM Bki_Anrede WHERE ID = tnID INTO ARRAY aID
		IF _TALLY > 0
			lnIDWebShop = aID[1]
		ELSE
			*-- 09.01.24 HPG: Wenn kein Eintrag in der Ansprechpartner-Tabelle
			*-- gefunden wird, setzen wird die ID auf 3 fr keine Anrede im
			*-- WebShop.
			lnIDWebShop = 3
		ENDIF

	*-- Andernfalls den Anwender informieren
	ELSE
		Fehler( 'Die Tabelle Bki_Ansprech ist nicht geffnet!' )
	ENDIF

	*-- Rckgabe
	RETURN( lnIDWebShop )

ENDFUNC


*----------------------------------------------------------------------------------------------------------------------------
* Beschreibung: bertrgt die Kunden-Daten des WebShops in eine DBF-Tabelle.
* Autor.......: Hans-Peter Grzinger
* Datum.......: 10.08.2023
*----------------------------------------------------------------------------------------------------------------------------
* Bemerkungen.:
* 10.08.23 HPG:  - tcMode:
*					- CU = Customer
*					- CO = Contact
*----------------------------------------------------------------------------------------------------------------------------
* nderungen..:
*----------------------------------------------------------------------------------------------------------------------------
* BugFixes....:
*----------------------------------------------------------------------------------------------------------------------------

FUNCTION TransferDatafromJSON( toIDSourceObject AS OBJECT, toSourceObject AS OBJECT, tcTargetAlias AS STRING, tcMode AS STRING )
********************************************************************************************************************************

	LOCAL lcCommand AS STRING, lcField AS STRING, lcDataType AS STRING, lcValue AS STRING, lcFunction AS STRING
	LOCAL lcErrorMessage AS STRING, lcTargetField AS STRING, lcSecondElement AS STRING, lcType AS STRING
	LOCAL lnLoop AS INTEGER, lnMaxLoop AS INTEGER
	LOCAL luValue AS VARIANT, luSecondElement AS VARIANT
	LOCAL loException AS EXCEPTION

	*-- Ist das Array schon deklariert?
	IF VARTYPE( gcCustomerFields ) <> 'C'

		*-- Nein, dann dies durchfhren
		CreateWebShopArrays()
	ENDIF

	*-- Je nach Modus verzweigen
	DO CASE

		*-- Adresse
		CASE tcMode = 'CU'
			lnMaxLoop = ALEN( gcCustomerFields, 1 )

		*-- Ansprechpartner
		CASE tcMode = 'CO'
			lnMaxLoop = ALEN( gcContactFields, 1 )
	ENDCASE

	*-- Schleife ber alle Felder des Arrays
	FOR lnLoop = 1 TO lnMaxLoop

		*-- Parameter setzen
		lcFunction = ''
		luSecondElement = .F.

		*-- Das Feld, eine eventuelle Konvertierfunktion
		*-- bzw. ein Ziel-Feldname holen
		*-- Je nach Modus verzweigen
		DO CASE

			*-- Adresse
			CASE tcMode = 'CU'
				lcField = ALLTRIM( gcCustomerFields[lnLoop, 1] )
				lcTargetField = lcField
				luSecondElement = gcCustomerFields[lnLoop, 2]

			*-- Ansprechpartner
			CASE tcMode = 'CO'
				lcField = ALLTRIM( gcContactFields[lnLoop, 1] )
				lcTargetField = lcField
				luSecondElement = gcContactFields[lnLoop, 2]
		ENDCASE

		*-- Gibt es einen Inhalt im zweiten Array-Element?
		IF VARTYPE( luSecondElement ) == 'C'

			*-- Ja, dann die Art und den Inhalt bestimmen
			lcType = LEFT( luSecondElement, 3 )
			lcSecondElement = SUBSTR( luSecondElement, 4 )

			*-- Je nach Art verzweigen
			DO CASE

				*-- Wir haben eine Funktion
				CASE lcType == 'FU:'
					lcFunction = lcSecondElement

				*-- Wir haben einen Ziel-Feldnamen
				CASE lcType == 'FN:'
					lcTargetField = lcSecondElement
			ENDCASE
		ENDIF

		*-- Beim ID-Feld
		IF lnLoop = 1

			*-- Den Wert bestimmen, den Befehl aufbauen
			*-- und ausfhren
			lcValue = Num2C( toIDSourceObject.&lcField )
			lcCommand = 'REPLACE ' + tcTargetAlias + '.ID_WS WITH ' + lcValue
			EXECSCRIPT( lcCommand )

		*-- Andernfalls kommen die weiteren Felder
		ELSE

			*-- Das Feld prfen, ob es im Quell-Objekt berhaupt
			*-- vorhanden ist und auch in der Zieltabelle.
			IF PEMSTATUS( toSourceObject, lcField, 5 ) .AND. LEN( FIELD( lcTargetField ) ) > 0

				*-- Den Wert und damit den Datentyp bestimmen
				luValue = toSourceObject.&lcField
				lcDataType = VARTYPE( luValue )

				*-- Je nach Datentyp verzweigen
				DO CASE

					*-- Wir haben eine Zeichenkette
					CASE lcDataType == 'C'
						lcValue = '"' + luValue + '"'

					*-- Wir haben ein Datums- oder ein Zeitfeld
					CASE lcDataType $ 'DT'

						*-- Wir mssen den Wert zuerst in eine Zeichenkette umwandeln
						*-- und dann mit den geschweiften Klammern einschlieen
						lcValue = '{' + TRIM( CAST( luValue AS CHARACTER( 200 ) ) ) + '}'

					*-- Wir haben ein logisches Feld
					CASE lcDataType == 'L'

						*-- Wir mssen den Wert zuerst in eine Zeichenkette umwandeln
						*-- und dann mit den Hochkommas einschlieen
						lcValue = '"' + TRIM( CAST( luValue AS CHARACTER( 1 ) ) ) + '"'

					*-- Wir haben ein numerisches oder ein Whrungsfeld
					CASE lcDataType $ 'NY'

						*-- Haben wir eine Funktion?
						IF LEN( lcFunction ) > 0
							lcValue = '"' + &lcFunction( luValue ) + '"'
						ELSE
							lcValue = Num2C( luValue )
						ENDIF
				ENDCASE

				*-- Den Befehl aufbauen
				lcCommand = 'REPLACE ' + tcTargetAlias + '.' + lcTargetField + ' WITH ' + lcValue

				*-- TRY-/Catch-Block
				TRY

					*-- Den Befehl ausfhren
					EXECSCRIPT( lcCommand )

				*-- Den Fehler einfangen
				CATCH TO loException

					*-- Fehler anzeigen
					*lcErrorMessage = 'Fehler in Methode frmKundenVerwaltung.Paket' + CHR( 13 ) + ;
						'Error: ' + STR( loException.ERRORNO ) + CHR( 13 ) + ;
						'LineNo: ' + STR( loException.LINENO )  + CHR( 13 ) + ;
						'Message: ' + loException.MESSAGE + CHR( 13 ) + ;
						'Procedure: ' + loException.PROCEDURE + CHR( 13 ) + ;
						'Details: ' + loException.DETAILS + CHR( 13 ) + ;
						'StackLevel: ' + STR( loException.STACKLEVEL ) + CHR( 13 ) + ;
						'LineContents: ' + loException.LINECONTENTS

					*-- MessageBox anzeigen und Fehler schreiben
					*MESSAGEBOX( lcErrorMessage, 0 + 48, 'Fehler' )
					*WriteError( lcErrorMessage )
				ENDTRY
			ENDIF
		ENDIF
	ENDFOR

ENDFUNC


************************************************************************
* wwapi::CreateGUID
********************
***    Author: Rick Strahl, West Wind Technologies
***            http://www.west-wind.com/
***  Modified: 01/26/98
***  Function: Creates a globally unique identifier using Win32
***            COM services. The value is guaranteed to be unique
***    Format: {9F47F480-9641-11D1-A3D0-00600889F23B}
***    Return: GUID as a string or "" if the function failed
*************************************************************************

FUNCTION CreateBKIGUID( llRaw AS Boolean )
******************************************
	
	LOCAL lcStruc_GUID, lcGUID, lnSize, lcUUID

	DECLARE INTEGER CoCreateGuid IN Ole32.DLL STRING @lcGUIDStruc
	DECLARE INTEGER StringFromGUID2 IN Ole32.DLL STRING cGUIDStruc, STRING @cGUID, LONG nSize

	*** Simulate GUID strcuture with a string
	lcStruc_GUID = REPLICATE( ' ', 16 )
	lcGUID = REPLICATE( ' ', 80 )
	lnSize = LEN( lcGUID ) / 2
	IF CoCreateGuid( @lcStruc_GUID ) # 0
		RETURN ''
	ENDIF

	IF llRaw
		RETURN lcStruc_GUID
	ENDIF

	*** Now convert the structure to the GUID string
	IF StringFromGUID2( lcStruc_GUID, @lcGUID, lnSize ) = 0
		RETURN ""
	ENDIF

	*** String is UniCode so we must convert to ANSI
	lcUUID = SUBSTR( STRCONV( LEFT( lcGUID, 76 ), 6 ), 2, 36 )

	RETURN( lcUUID )

ENDFUNC


***< Neuer Datensatz-ID >***
FUNCTION NewID
**************

	LPARAMETERS lcTABLE
	LOCAL lnCURR_SEL, llUSED, lnCURR_REPR, lnKEY

	* Aktuelles work area sichern, ID_TAB ffnen (wenn ntig), und die
	* gewnschte Tabelle suchen. Falls nicht vorhanden, einen Satz anlegen.

	lnCURR_SEL = SELECT()
	llUSED     = USED('ID_TAB')
	IF llUSED
		SELECT ID_TAB
		SET ORDER TO PRIMARY
	ELSE
		SELECT 0
		USE ..\dat_kund.bki\ID_TAB ORDER PRIMARY
	ENDIF llUSED
	SEEK lcTABLE+" "
	IF NOT FOUND()
		INSERT INTO ID_TAB VALUES (lcTABLE, 0)
	ENDIF NOT FOUND()

	* Increment the next available ID.

	lnCURR_REPR = SET('REPROCESS')
	SET REPROCESS TO 10 SECONDS
	IF RLOCK()
		REPLACE ID WITH ID + 1
		lnKEY = ID
		UNLOCK
	ELSE
		lnKEY = 0	&& Force it to a numeric type before setting it to .NULL.
		lnKEY = .NULL.
	ENDIF RLOCK()

	* Cleanup und return.

	SET REPROCESS TO lnCURR_REPR
	IF NOT llUSED
		USE
	ENDIF NOT llUSED
	SELECT (lnCURR_SEL)
	RETURN lnKEY
ENDFUNC